#if!defined (_dmatrix_h_)

#define _dmatrix_h_

#include<fstream>
#include<iostream>
#include<stdlib.h>
using namespace std;


/*****************************************************************
class dmatrix
      description of variables:
           no_of_rows - this is the no of rows 
           no_of_columns - this is the number of columns
           matrix_ptr - a double pointer to double array with the matrix elements. 
        
****************************************************************/

class dmatrix {
 protected:        
    int no_of_rows, no_of_columns; 
    double ** matrix_ptr; 
 public:

/****************************************************************************
****************************************************************************
****************************************************************************

                    CONSTRUCTORS AND DESTRUCTORS. 

****************************************************************************
****************************************************************************
****************************************************************************/







/****************************************************************************

Implicit constructor for the dmatrix class. 

input variables = none

output variables  = none

return = none

What it does:
     set the 'no_of_rows' and 'no_of_columns' to zero and the 'matrix_ptr' is set to NULL

how to call:
     dmatrix(); 
   
****************************************************************************/
    dmatrix();

/****************************************************************************

Implicit constructor for the dmatrix class. 

input variables = two
        int height - 'no_of_rows' is set to this value
        int width  - 'no_of_columns' is set to this value
output variables  = none

return = none

What it does:
     set the 'no_of_rows' to the first input variable
     and 'no_of_columns' to the second input variable  and the 'matrix_ptr'
     allocates an double array of size [no_of_rows][no_of_columns]
     the indexing of the matrix_ptr is matrix_ptr[0..no_of_rows-1][0..no_of_columns-1]
     all the 'matrix_ptr' elements are set to zero.
how to call:
     dmatrix(height, width); 
   
****************************************************************************/

    dmatrix(int, int) ;


/****************************************************************************

Implicit destructor for the dmatrix class. 

input variables = none
        
output variables  = none

return = none

What it does:
     distroys the dmatrix class, by setting the 'no_of_columns' and
     'no_of_rows' to zero and 'matrix_ptr' to NULL.

how to call:
     ~dmatrix(); 
   
****************************************************************************/
    ~dmatrix(); 


/****************************************************************************
****************************************************************************
****************************************************************************

                              METHODS

****************************************************************************
****************************************************************************
****************************************************************************/

/****************************************************************************


input variables = two
        int height - 'no_of_rows' is set to this value
        int width  - 'no_of_columns' is set to this value
output variables  = none

return = void

What it does:
     set the 'no_of_rows' to the first input variable
     and 'no_of_columns' to the second input variable  and the 'matrix_ptr'
     allocates an double array of size [no_of_rows][no_of_columns]
     the indexing of the matrix_ptr is matrix_ptr[0..no_of_rows-1][0..no_of_columns-1]
     all the 'matrix_ptr' elements are set to zero. 
how to call:
     allocate(height, width); 
   
****************************************************************************/
   
    void allocate(int , int );  


/****************************************************************************


input variables = none

output variables  = none

return = void

What it does:
     set to zero the  'no_of_rows' and 'no_of_columns' and the 'matrix_ptr'
     is set to NULL. 
how to call:
     clear(); 
   
****************************************************************************/

    void clear(void); 

/****************************************************************************


input variables = none

output variables  = none

return = an integer that represent the 'no_of_columns' variable. 

What it does:
     returns the value of the 'no_of_columns' variable
how to call:
     get_no_of_columns(); 
   
****************************************************************************/

    int  get_no_of_columns(); 

/****************************************************************************


input variables = none

output variables  = none

return = an integer that represent the 'no_of_rows' variable. 

What it does:
     returns the value of the 'no_of_rows' variable
how to call:
     get_no_of_rows(); 
   
****************************************************************************/

    int  get_no_of_rows(); 

/****************************************************************************


input variables = none

output variables  = none

return = an double pointer to a double variable that represents the 'matrix_ptr' variable. 

What it does:
     returns the 'matrix_ptr' pointer. 
how to call:
     get_matrix_ptr(); 
   
****************************************************************************/

    double ** get_matrix_ptr(); 

/****************************************************************************

input variables = three
        int i - is the corresponding row in the matrix table that will be modified. 
        int j - is the corresponding column in the matrix table that will be modified
        double x - is the new value of the matrix_ptr[i][j] element. 

output variables  = none

return = void

What it does:
      set the 'matrix_ptr[i][j]' to x. 
how to call:
     get_matrix_ptr(i,j, x); 
   
****************************************************************************/

    void set_matrix_element(int , int , double ); 


/****************************************************************************
*****************************************************************************
*****************************************************************************

                                OPERATORS

*****************************************************************************
*****************************************************************************
*****************************************************************************/

/****************************************************************************

input variables = one
        int i - is the corresponding row in the matrix table. 

output variables  = none

return = a pointer to a double that represents the address of the 
         'i-th' row in the 'matrix_ptr' array.

What it does:
      returns a pointer to the i-th row in the 'matrix_ptr' variable. 
how to call:
      dmatrix::operator[](int)
   
****************************************************************************/
    
    double * operator [] (int);

/****************************************************************************

input variables = one
        dmatrix &  mmatrix - is a new dmatrix variable that in copied in the current one. 

output variables  = none

return = a reference to the new dmatrix variable. Now the 'matrix_ptr' contains
         the mmatrix.matrix_ptr elements.

What it does:
      copy the mmatrix variable in the present one. 
how to call:
      dmatrix::operator =(dmatrix &)
   
****************************************************************************/


    dmatrix & operator = (dmatrix & ); 

/****************************************************************************

input variables = one
        dmatrix &  mmatrix - is a new dmatrix variable that multiplies to the right the current one. 

output variables  = none

return = a reference to the new dmatrix variable, that is the product of the current one
         with the mmatrix. The product is done in the form of matrix multiplication  
         for the 'matrix_ptr' elements of the current and 'mmatrix' variable. 

What it does:
      multiplies the current variable with the 'mmatrix'  in the form of matrix product.
how to call:
      dmatrix::operator *(dmatrix &)
   
****************************************************************************/

    dmatrix & operator *= (const dmatrix &); 

/****************************************************************************
*****************************************************************************
*****************************************************************************

                                FRIEND FUNCTIONS.

*****************************************************************************
*****************************************************************************
*****************************************************************************/


/****************************************************************************

input variables = two
        dmatrix &  lmatrix - left matrix in the product
        dmatrix &  rmatrix - right matrix in the product

output variables  = none

return = a dmatrix variable that is the matrix product of the lmatrice.matrix_ptr matrix
        with the rmatrix.matrix_ptr

What it does:
      multiplies two dmatrix varibles lmatrix*rmatrix and returns an new variable. 
how to call:
      new_dmatrix = lmatrix*rmatrix.
   
****************************************************************************/


    friend dmatrix operator * (const dmatrix & lmatrix, dmatrix & rmatrix); 
}; 

#endif


