#if!defined (_in_out_h_)

#define _in_out_h_

#include "state.h"
#include "symmetry.h"
#include "quantum_operator.h"
#include "llvector.h"
#include "dmnrg.h"

#include<fstream>
#include<iostream>
using namespace std;
#include <stddef.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


typedef char name[256];
extern name *couplings_name;

extern char model[256];
extern double lambda ; // the value of lambda, will be readed as an input parameter.
extern double sqrt_lambda; // the value of the square root of lambda 
extern int max_state_no;
extern int iteration_no;
extern int max_iteration_no;
extern int symmetry_no;
extern int block_state_no;
extern int coupling_no;
extern int local_coupling_no; 
extern int local_state_no;
extern int spectral_operator_no; 
extern int static_operator_no;
extern int hopping_operator_no;
extern int interval_no;
extern double degeneracy_threshhold; 
extern int spectral_function_no;
extern double magnetic_field;
extern double temperature;
extern int * spectral_function_index[2]; 
extern int * self_energy_index[4]; 
extern int self_energy_no; 
extern double * couplings ;
extern double * couplings_rescaled ;
extern double *local_couplings;
extern double i_delta; 
extern long mesh_points_no; 
extern long refined_mesh_neighbors; 
extern long refined_mesh_points_no; 
extern double allowed_error; 
extern double bare_temperature;
extern const int buffer_size;
extern char buffer[5000];
extern char word[5000];
extern int word_offset;

extern quantum_operator *block_spectral_operators;
extern quantum_operator *block_static_operators; 
extern int block_sector_no;
//extern llvector block_sectors_dimension;
//extern llvector * block_sectors_representation_index;
extern sector_properties *block_sectors_properties_array; 

extern int local_sector_no;
//extern llvector local_sectors_dimension;
//extern llvector * local_sectors_representation_index;
extern sector_properties  *local_sectors_properties_array;

extern symmetry * symmetry_types; 
extern int trial_state_no; // After that the symmetries have been read from the parameter file 

extern state * local_states_array_copy; 
extern state * local_states_array;

extern state * block_states_array_copy; 
extern state * block_states_array; 


extern quantum_operator * block_hamiltonian_terms; 
extern quantum_operator block_hamiltonian; 

extern quantum_operator *local_hamiltonian_terms;
extern quantum_operator local_hamiltonian;

extern quantum_operator *local_hopping_hamiltonians; 

extern quantum_operator* block_hopping_operators;
extern quantum_operator* block_spectral_operators; 
extern quantum_operator* local_hopping_operators; 
extern quantum_operator* block_static_operators;

extern quantum_operator* new_block_hopping_operators;
extern quantum_operator* new_block_spectral_operators; 
extern quantum_operator* new_block_static_operators;

extern ofstream spectrum_odd; // file to store the energy spectrum of the odd N-s 
extern ofstream spectrum_even; // file to store the energy spectrum of the even N-s 
extern ofstream sp_plus; // file to store the peak values of the spectral function for positive energies 
extern ofstream sp_minus; // file to store the peak values of the spectral function for negative energies 
extern ofstream energy_spectrum_odd; // file to store the energy spectrum of the odd N-s 
extern ofstream energy_spectrum_even; // file to store the energy spectrum of the even N-s 

extern char broadening_method[100]; 
extern double broadening_energy_minim;
extern double broadening_energy_maxim;
extern double broadening_parameter;
extern int broadening_grid_mesh;
extern double quantum_temperature;

extern int spectral_function_nrg_even; 
extern int spectral_function_nrg_odd;
extern int spectral_function_dmnrg; 
extern int static_average_dmnrg;
extern int green_function_dmnrg;
extern int green_function_nrg_even;
extern int green_function_nrg_odd;
extern long green_function_grid_mesh;
extern char log_file_name[400]; 
extern char working_folder_name[400]; 
extern int m_star;


extern int dmnrg_flag; 
extern int text_swap_files_flag; 
extern int binary_swap_files_flag; 
extern int hoppings_on_site_energies_flag;

extern int self_consistancy_index; 

extern char mapping_folder_name[400]; 
extern char dos_mapping_folder_name[400]; 
extern char on_site_energies_file_name[400]; 
extern char hopping_couplings_file_name[400]; 

extern double beta_tilde; 
extern char swap_folder_name[400]; 
extern char delta_peaks_folder_name[400];
extern char spectral_functions_folder_name[400];
extern char transformations_folder_name[400];
extern char spectral_operators_folder_name[400];
extern char static_operators_folder_name[400];
extern char states_folder_name[400];
extern char sectors_folder_name[400];


void remove_density_of_states_files();
void remove_binary_files(); 


void create_working_folder(char * model,
			   double lambda, 
			   int max_state_no,
			   int iteration_no,
			   double temperature,		   
			   int symmetry_no,
			   symmetry *symmetry_types,
			   int coupling_no, 
			   double *couplings,
			   char* name
			   );


void create_swap_folder(char *working_folder_name, 
			char *swap_folder_name); 

void create_delta_peaks_folder(char *swap_folder_name, 
			       char * delta_peaks_folder_name);

void create_spectral_functions_folder(char *swap_folder_name, 
				      char * spectral_functions_folder_name);
void create_transformations_folder(char *swap_folder_name, 
				   char * transformations_folder_name);
void create_spectral_operators_folder(char *swap_folder_name, 
				      char * spectral_operators_folder_name);
void create_states_folder(char *swap_folder_name, 
			  char * states_folder_name);

void create_sectors_folder(char *swap_folder_name, 
			   char * sectors_folder_name);

void create_static_operators_folder(char *swap_folder_name, 
				    char * static_operators_folder_name);

void create_mapping_folder( char *mapping_folder_name); 

void create_dos_mapping_folder(char *dos_mapping_folder_name); 

  
void transformation_swap_file_name(char *transformations_folder_name, 
				   int iteration_index, 
				   char *transformation_file_name
				   ); 

void transformation_binary_swap_file_name(char *transformations_folder_name, 
					  int iteration_index, 
					  char *transformation_file_name
					  ); 


void delta_peaks_positive_swap_file_name(char *delta_peaks_folder_name,
					 int iteration_index,
					 char *delta_peaks_negative_file_name,
					 quantum_operator & spectral_operator_A,
					 quantum_operator & spectral_operator_B);




void delta_peaks_negative_swap_file_name(char *delta_peaks_folder_name,
					 int iteration_index,
					 char *delta_peaks_negative_file_name,
					 quantum_operator & spectral_operator_A,
					 quantum_operator & spectral_operator_B);


void delta_peaks_odd_swap_file_name(char *delta_peaks_folder_name,
				    char *delta_peaks_odd_file_name,
				    quantum_operator & spectral_operator_A,
				    quantum_operator & spectral_operator_B);

void delta_peaks_even_swap_file_name(char *delta_peaks_folder_name,
				    char *delta_peaks_even_file_name,
				    quantum_operator & spectral_operator_A,
				     quantum_operator & spectral_operator_B);

void delta_peaks_dmnrg_swap_file_name(char *deltap_peaks_folder_name,
				      char *delta_peaks_dmnrg_file_name,
				      quantum_operator & spectral_operator_A,
				      quantum_operator & spectral_operator_B);


void delta_peaks_dmnrg_swap_file_name(char *delta_peaks_folder_name,
				      int iteration_index,
				      char *delta_peaks_dmnrg_file_name,
				      quantum_operator & spectral_operator_A,
				      quantum_operator & spectral_operator_B);

void spectral_function_dmnrg_swap_file_name(char *spectral_functions_folder_name,
					    int iteration_index,
					    double broadening_parameter, 
					    char *spectral_function_dmnrg_file_name,
					    quantum_operator & spectral_operator_A,
					    quantum_operator & spectral_operator_B);


void delta_peaks_positive_dmnrg_swap_file_name(char *delta_peaks_folder_name,
					       int iteration_index,
					       char *delta_peaks_positive_dmnrg_file_name,
					       quantum_operator & spectral_operator_A,
					       quantum_operator & spectral_operator_B); 

void delta_peaks_negative_dmnrg_swap_file_name(char *delta_peaks_folder_name,
					       int iteration_index,
					       char *delta_peaks_negative_dmnrg_file_name,
					       quantum_operator & spectral_operator_A,
					       quantum_operator & spectral_operator_B); 


void spectral_function_odd_swap_file_name(char *spectral_functions_folder_name,
					  double b,
					  char *spectral_function_file_name,
					  quantum_operator & spectral_operator_A, 
					  quantum_operator & spectral_operator_B
					  );


void spectral_function_even_swap_file_name(char *working_folder_name,
					  double b,
					  char *spectral_function_file_name,
					  quantum_operator & spectral_operator_A, 
					  quantum_operator & spectral_operator_B
					  );

void spectral_function_dmnrg_swap_file_name(char *working_folder_name,
					    double b,
					    char *spectral_function_file_name,
					    quantum_operator & spectral_operator_A, 
					    quantum_operator & spectral_operator_B
					    ); 

void static_average_spectral_operators_swap_file_name(char *static_operators_folder_name,
						      char *static_average_spectral_operators_file_name,
						      quantum_operator & spectral_operator_A, 
						      quantum_operator & spectral_operator_B
						      );

void static_average_static_operator_swap_file_name(char *static_operators_folder_name,
						   char *static_average_static_operators_file_name,
						   quantum_operator & static_operator_A
						   );


void static_average_spectral_operators_nrg_file_name(char *working_folder_name,
						     char *static_average_spectral_operators_file_name,
						     quantum_operator & spectral_operator_A, 
						     quantum_operator & spectral_operator_B
						     );



void states_swap_file_name(char *swap_folder_name, 
			   int iteration_index,
			   char *states_file_name
			   ); 


void sectors_swap_file_name(char *swap_folder_name, 
			   int iteration_index,
			   char *sectors_file_name
			    );

void spectral_operator_swap_file_name(char *swap_folder_name,
				      int iteration_index, 
				      char *spectral_operator_file_name,
				      quantum_operator & spectral_operator); 

void spectral_operator_binary_swap_file_name(char *swap_folder_name,
					     int iteration_index, 
					     char *spectral_operator_file_name,
					     quantum_operator & spectral_operator); 


void static_operator_swap_file_name(char *swap_folder_name,
				      int iteration_index, 
				      char *static_operator_file_name,
				      quantum_operator & static_operator); 

void static_operator_binary_swap_file_name(char *swap_folder_name,
					     int iteration_index, 
					     char *static_operator_file_name,
					     quantum_operator & static_operator); 

void m_star_swap_file_name(char *swap_folder_name,
			   char *m_star_file_name); 


void log_swap_file_name(char *working_folder_name,
			char *log_file_name); 


void input_swap_file_name(char *working_folder_name,
			  char *input_file_name);

void density_of_states_negative_mapping_file_name(char *dos_mapping_folder_name,
						  int iteration_index,
						  int self_consistancy_index, 
						  char *density_of_states_file_name
						  ); 

void density_of_states_positive_mapping_file_name(char *dos_mapping_folder_name,
						  int iteration_index,
						  int self_consistancy_index, 
						  char *density_of_states_file_name
						  ); 


void gamma_negative_mapping_file_name(char *dos_mapping_folder_name,
				      int self_consistancy_index, 
				      char *gamma_file_name
				      );


void gamma_positive_mapping_file_name(char *dos_mapping_folder_name,
				      int self_consistancy_index, 
				      char *gamma_file_name
				      );


void zeta_negative_mapping_file_name(char *dos_mapping_folder_name,
				      int self_consistancy_index, 
				      char *zeta_file_name
				     );

void zeta_positive_mapping_file_name(char *dos_mapping_folder_name,
				     int self_consistancy_index, 
				     char *zeta_file_name
				     );



void on_site_energies_mapping_file_name(char *dos_mapping_folder_name,
					int self_consistancy_index, 
					char *on_site_energies_file_name
					); 

void hopping_couplings_mapping_file_name(char *dos_mapping_folder_name,
					 int self_consistancy_index, 
					 char *hopping_couplings_file_name
					 ); 



void error_message(const char *text);
bool get_word(char * string, char * word, int & word_offset); 
void read_parameters_section();
void read_flags_section(); 
void read_symmetries_section();
void read_block_states_section();
void read_local_states_section();
void read_local_states_signs_section();
void read_block_hamiltonian_section();
void read_hopping_operators_section();
void read_spectral_operators_section();
void read_static_operators_section();
void read_local_hopping_operators_section();
void read_spectral_function_section(); 
void read_spectral_function_broadening_section(); 
void copy_input_file();
void copy_hoppings_on_site_energies_files();


void green_function_dmnrg_swap_file_name(char *working_folder_name,
					 double b,
					 char *green_function_file_name,
					 quantum_operator & spectral_operator_A, 
					 quantum_operator & spectral_operator_B
					 );



void green_function_even_swap_file_name(char *working_folder_name,
					 double b,
					 char *green_function_file_name,
					 quantum_operator & spectral_operator_A, 
					 quantum_operator & spectral_operator_B
					 );

void green_function_odd_swap_file_name(char *working_folder_name,
				       double b,
				       char *green_function_file_name,
				       quantum_operator & spectral_operator_A, 
				       quantum_operator & spectral_operator_B
				       );


void read_hopping_couplings_and_on_site_energies(double ** hopping_couplings_array, 
						 double ** on_site_energies_array); 

void read_local_hamiltonian_section();

#endif

