#if!defined (_state_h_)

#define _state_h_
#include <assert.h>
#include<fstream>
#include<iostream>
using namespace std;

#include "llvector.h"
#include "sector.h"


/*********************************************************************
class states

    variables definition
        sector  - tells the sector the state belong to 
        sector_index      - index of the state within the sector, values between 0 and sector_size -1
        representation_index - these are the quantum numbers describing the states. 
        parent_state_index - current state index of the parent state 
        local_state_index - local state index of the parent stae. 
        sign -is the sign of the state.
********************************************************************/

extern double epsilon; 

class state {
 protected:        

  int sector;
  int sector_index;
  llvector representation_index; 
  int parent_state_index; 
  int local_state_index;
  int sgn;
  char elimination_tag;  
 public:

  state();
  ~state(); 
  void activate(int ); 
  int get_parent_state_index(); 
  void set_parent_state_index(int);
  int get_sgn();
  void set_sgn(int); 
  int get_local_state_index(void); 
  void set_local_state_index(int); 
  int get_sector(); 
  void set_sector(int); 
  int get_sector_index(); 
  void set_sector_index(int ); 
  int get_representation_index(int); 
  void set_representation_index(int ,int); 
  int get_representation_index_size(); 
  void set_representation_index(llvector & ) ; 
  llvector & get_representation_index();  
  void set_elimination_tag(char ); 
  char get_elimination_tag(); 

  //operators.   
  state & operator = (state & ); 
  int operator < (state & ); 
  int operator == (state &); 
  ostream & print(ostream & ); 


  friend  ostream & operator << (ostream & , state & ); 

  friend void sort_states(unsigned long n, state ra[]);
  //for a set of states, returns the number of sectors.

  friend int check_states_order(unsigned long n, state ra[],  state rb[]);

  //for a set of states, returns the number of sectors.

 

};



#endif
