


/****************************************************************************

Last Update : 2007/09/27 Pascu Moca 

No documentation update necessary


*****************************************************************************/
#include "../include/clebsch_gordan.h"
#include<math.h>

/*----------------------------------------------------------------------------*/

double clebsch_gordan_spin_1_2(	int S_old, int S_z_old,
				int s_added, int s_z_added,
				int S_new, int S_z_new) {

/******************************************************************************
input variables = six
     S_old     : total spin of the block_spin (block state)
     S_z_old   : z-component of the block_spin (block state)
     S_added   : total spin of the single particle  (local state)
     S_z_added : z-component of the single particle (local state)
     S_new     : total spin of the new block_spin (new state)
     S_z_new   : z-component of the new block_spin (new state)

output variables  = none

return = The Clebsch Gordan coefficient (double).

What it does:
     Calculates the Clebsch Gordan coefficient corresponding to new block spin
     when a spin-1/2 particle is added to the block spin:
     < S_old, S_z_old; s_added, s_z_added | S_new, S_z_new >
     Note that in the code the twice of the s_z component is used! 

how to call:
******************************************************************************/

    double S_old_red_1 = S_old - 1;
    double S_z_old_red_1 = S_z_old - 1;
    double S_old_inc_1 = S_old + 1;
    double S_z_old_inc_1 = S_z_old + 1;
    double S_old_half = S_old / 2.;
    double S_z_old_half = S_z_old / 2.;

    if((S_new == S_old_red_1) && (S_z_new == S_z_old_red_1) && (s_z_added == - 1))
	return double (sqrt((S_old_half + S_z_old_half) / S_old_inc_1));

    if((S_new == S_old_red_1) && (S_z_new == S_z_old_inc_1) && (s_z_added == 1))
	return double (- sqrt((S_old_half - S_z_old_half)/ S_old_inc_1));

    if((S_new == S_old_inc_1) && (S_z_new == S_z_old_inc_1) && (s_z_added == 1))
	return double (sqrt((1.0 + S_old_half + S_z_old_half) / S_old_inc_1));

    if((S_new == S_old_inc_1) && (S_z_new == S_z_old_red_1) && (s_z_added == - 1))
	return double (sqrt((1.0 + S_old_half - S_z_old_half) / S_old_inc_1));

    return 0.0;
}

/*----------------------------------------------------------------------------*/

//	< S_old, S_z_old; s_added, s_z_added | S_new, S_z_new >
//	for spin = 1
double clebsch_gordan_spin_1(	int S_old, int S_z_old,
				int s_added, int s_z_added,
				int S_new, int S_z_new) {

/******************************************************************************
input variables = six
     S_old     : total spin of the block_spin (block state)
     S_z_old   : z-component of the block_spin (block state)
     S_added   : total spin of the single particle  (local state)
     S_z_added : z-component of the single particle (local state)
     S_new     : total spin of the new block_spin (new state)
     S_z_new   : z-component of the new block_spin (new state)

output variables  = none

return = The Clebsch Gordan coefficient (double).

What it does:
     Calculates the Clebsch Gordan coefficient corresponding to new block spin
     when a spin-1 particle is added to the block spin:
     < S_old, S_z_old; s_added, s_z_added | S_new, S_z_new >
     Note that in the code the twice of the s_z component is used! 

how to call:
******************************************************************************/

    double S_old_red_2 = S_old - 2;
    double S_z_old_red_2 = S_z_old - 2;
    double S_old_inc_2 = S_old + 2;
    double S_z_old_inc_2 = S_z_old + 2;
    double S_new_half = S_new / 2.;
    double S_z_new_half = S_z_new / 2.;

    if(S_new == S_old_red_2)
    {
	if((S_z_new == S_z_old_red_2) && (s_z_added == - 2))
	    return double (sqrt((1. + S_new_half + S_z_new_half) * (2. + S_new_half + S_z_new_half)
				/ (2. * (1. + S_new_half) * (3. + S_new))));

	if((S_z_new == S_z_old_inc_2) && (s_z_added == 2))
	    return double (sqrt((1. + S_new_half - S_z_new_half) * (2. + S_new_half - S_z_new_half)
				/ (2. * (1. + S_new_half) * (3. + S_new))));

	if ((S_z_new == S_z_old) && (s_z_added == 0))
	    return double (- sqrt((1. + S_new_half - S_z_new_half) * (1. + S_new_half + S_z_new_half)
				  / ((1. + S_new_half) * (3. + S_new))));
    }

    if(S_new == S_old)
    {
	if((S_z_new == S_z_old_red_2) && (s_z_added == - 2))
	    return double (sqrt((S_new_half - S_z_new_half) * (1. + S_new_half + S_z_new_half)
				/ (2. * S_new_half * (1. + S_new_half))));

	if((S_z_new == S_z_old_inc_2) && (s_z_added == 2))
	    return double (- sqrt((1. + S_new_half - S_z_new_half) * (S_new_half + S_z_new_half)
				  / (2. * S_new_half * (1. + S_new_half))));

	if ((S_z_new == S_z_old) && (s_z_added == 0))
	    return double ( S_z_new_half / sqrt(S_new_half * (1. + S_new_half)));
    }

    if(S_new == S_old_inc_2) 
    {
	if ((S_z_new == S_z_old_red_2) && (s_z_added == -2))
	    return double (sqrt((- 1. + S_new_half - S_z_new_half) * (S_new_half - S_z_new_half)
				/ (2. * S_new_half * (- 1. + S_new))));
	
	if ((S_z_new == S_z_old_inc_2) && (s_z_added == 2))
	    return double (sqrt((- 1. + S_new_half + S_z_new_half) * (S_new_half + S_z_new_half)
				/ (2. * S_new_half * (-1. + S_new)) ));
	
	if ((S_z_new == S_z_old) && (s_z_added == 0))
	    return double (sqrt((S_new_half - S_z_new_half) * (S_new_half + S_z_new_half)
				/ (S_new_half * (-1. + S_new))));
    }

    return 0.0;
}

/*----------------------------------------------------------------------------*/

double clebsch_gordan( int S_old, int S_z_old,
                       int s_added, int s_z_added,
                       int S_new, int S_z_new) {
/******************************************************************************
input variables = six
     S_old     : total spin of the block_spin (block state)
     S_z_old   : z-component of the block_spin (block state)
     S_added   : total spin of the single particle  (local state)
     S_z_added : z-component of the single particle (local state)
     S_new     : total spin of the new block_spin (new state)
     S_z_new   : z-component of the new block_spin (new state)

output variables  = none

return = The Clebsch Gordan coefficient (double).

What it does:
     Calculates the Clebsch Gordan coefficient corresponding to new block spin
     when a spin-0, 1/2, or 1 particle is added to the block spin:
     < S_old, S_z_old; s_added, s_z_added | S_new, S_z_new >
     First some consistency checks are performed. In case of error 0.0 is 
     returned, otherwise, the appropriate function is called based on the spin 
     of the particle added to the block.  
     Note that in the code the twice of the s_z component is used! 

how to call:
******************************************************************************/



    if((S_new < 0) || (S_old < 0) || (s_added < 0))
	return 0.0;

    if((S_z_old + s_z_added) != S_z_new)
	return 0.0;

    if((fabs(S_old - s_added) > S_new) || (S_old + s_added < S_new))
	return 0.0;

    if ((fabs(S_z_old) > S_old) || (fabs(S_z_new) > S_new) || (fabs(s_z_added) > s_added))
	return 0.0;

    if (((S_old + S_z_old) % 2) || ((S_new + S_z_new) % 2) || ((s_added + s_z_added) % 2))
	return 0.0;

//spin = 0
    if(s_added == 0 && s_z_added == 0) 
    {
	if((S_new == S_old) && (S_z_new == S_z_old))
	    return 1.0; 
    }

//spin = 1/2
    if(s_added == 1 && (s_z_added == 1 || s_z_added == - 1))
	return clebsch_gordan_spin_1_2(S_old, S_z_old, s_added, s_z_added, S_new, S_z_new);

//spin = 1
    if (s_added == 2 && (s_z_added == - 2 || s_z_added == 0 || s_z_added == 2))
	return clebsch_gordan_spin_1(S_old, S_z_old, s_added, s_z_added, S_new, S_z_new);

    //for larger spin values we use the routine in the gls_library
    

    return      pow(-1.0, 0.5*(S_old-s_added+S_z_new))*sqrt(S_new+1.0)*
      gsl_sf_coupling_3j(S_old, s_added, S_new,
			 S_z_old,s_z_added, -S_z_new);

     
}

/*----------------------------------------------------------------------------*/

double clebsch_gordan_charge(int Q_old, int q_add, int Q_new) {
/******************************************************************************
input variables = three
     Q_old   : total charge of the block_spin (block state)
     q_add   : total charge of the single particle (local state)
     Q_new   : total charge of the new block spin (new state)

output variables  = none

return = 1.0 for proper charge values and 0.0 otherwise (double).

What it does:
     Calculates the charge Clebsch Gordan coefficient corresponding to new 
     block spin when a single particle is added to the block spin.
     Note that the charge is simply additive.
how to call:
******************************************************************************/


    if ( (Q_old + q_add) == Q_new ) return 1.0;
    else return 0.0;
    
}

/*----------------------------------------------------------------------------*/
