/********************************************************************************
Modification dates: 
2007/10/24
2008/02/04
2008/08/22
Documentation update necessary


*********************************************************************************/
#include "../include/construction.h"
#include "../include/in_out.h"
#include "../include/clebsch_gordan.h"

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void set_original_sectors(int state_no, 
			  state *states_array,
			  int * &block_original_sector_array, 
			  int * &block_original_sector_index_array){

/*******************************************************************************
input variables:  four
    state_no     - is the number of states (either local, current of new)
    states_array - represent an array with the states that has to be
                   organized into sectors.
    block_original_sector_array - array with the sectors
    block_original_sector_index_array- array with the sectors index. 

output variables: none

return: integer
    The routine sets the original sectors for the block_states

What it does:
    This routine return the number of sectors. 
*******************************************************************************/


  if (block_original_sector_array){
    delete [] block_original_sector_array;
  }
    
  block_original_sector_array = new int [block_state_no]; 
  
  
  if (block_original_sector_index_array){
    delete [] block_original_sector_index_array;
  }
  
  block_original_sector_index_array = new int [block_state_no]; 
  




  for (int state_index = 0; state_index<state_no; state_index++){
    block_original_sector_array[state_index] = states_array[state_index].get_sector();
    block_original_sector_index_array[state_index]= states_array[state_index].get_sector_index();
  }



}//set original sectors


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int get_sectors_no(int state_no, 
		   state* states_array){

/*******************************************************************************
input variables:  two
    state_no     - is the number of states (either local, current of new)
    states_array - represent an array with the states that has to be
                   organized into sectors.

output variables: one

return: integer
    The routine returns the number of sectors as an integer.

What it does:
    This routine return the number of sectors. 
*******************************************************************************/

  llvector temporal_representation_index; 

  int sector=0; 
  int sector_index=0; 
  int size = states_array[0].get_representation_index().get_size();

  
  //check if all the states_array has the same representation index form. 
  for(int i=0; i<state_no; i++){
    assert(size==states_array[i].get_representation_index().get_size());
  }//i 

  temporal_representation_index.allocate(size); 
  temporal_representation_index = states_array[0].get_representation_index();
  
  for(int i = 0; i < state_no; i++){

    if(temporal_representation_index == states_array[i].get_representation_index()){
      states_array[i].set_sector(sector);
      states_array[i].set_sector_index(sector_index);
      sector_index++;
    }//if

    else{
      temporal_representation_index = states_array[i].get_representation_index();
      sector++;
      sector_index = 0;
      states_array[i].set_sector(sector);
      states_array[i].set_sector_index(sector_index);
      sector_index++;
    }//else
  }//for i
  sector++;
  return (sector);   

}//get_sectors_no

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void set_sectors_representation_index(int symmetry_no,
				      symmetry *symmetry_types,

				      int state_no, 
				      state * states_array, 

				      int sector_no,
				      sector_properties * &sectors_properties_array){

/*******************************************************************************
input variables:  six
     symmetry_no    - the number of symmetries in the input.dat file. 
     symmetry_types - the types of the symmetries that are U(1), SU(2) etc. 
     state_no       - the number of states
     states_array   - array with the states that has to be organized into 
                      sectors. 
     sector_no      - the number of the sectors. (This has to be compute with 
                      the get_sector_no (....) routine, see above.)

output variables: one
     sectors_properties_array - array of a structure with the properties of 
                                the sectors. The components of the 
                                sector_properties structure are:
     dimension                - represents the dimension of each sector
     first_state              - the first state in each sector
     representation_index     - the quantum numbers of each sector. (these are
                                the same with the quantum 
                                numbers of the states in the given sector)
     degeneracy               - the degeneracy of a given sector. All the 
                                states in that sector have  the same dimension. 

return: void

What it does:
     This routine arrange the states into sectors and sets the properties of the
     sectors. 

How to call:
*******************************************************************************/

  if (sectors_properties_array){
    delete [] sectors_properties_array; 
  }
  sectors_properties_array =NULL;

  sectors_properties_array = new sector_properties [sector_no]; 

  llvector temporal_representation_index;
  int sector_index = 0;
  int size = states_array[0].get_representation_index().get_size();
  
  //check if all the states_array has the same representation index form. 
  for(int i=0; i<state_no; i++){
    assert(size==states_array[i].get_representation_index().get_size());
  }//i 
  temporal_representation_index.allocate(size);
  
  for(int i = 0; i < sector_no; i++){
    sectors_properties_array[i].representation_index.allocate(size);
  }//for i

  // passing over the representation indices 
  
  temporal_representation_index = states_array[0].get_representation_index();
  sectors_properties_array[0].representation_index = temporal_representation_index;
  sectors_properties_array[0].dimension = 0;
  sectors_properties_array[0].first_state = 0;
  sectors_properties_array[0].degeneracy =1; 

  for(int i = 0; i < state_no; i++){

    if(temporal_representation_index == states_array[i].get_representation_index()){
      sectors_properties_array[sector_index].dimension++;
    }//if

    else{
      sector_index++;
      temporal_representation_index = states_array[i].get_representation_index();
      sectors_properties_array[sector_index].representation_index = temporal_representation_index;
      sectors_properties_array[sector_index].first_state = i;
      sectors_properties_array[sector_index].dimension = 1;
      sectors_properties_array[sector_index].degeneracy =1;
    }//else

  }//for i

  //now we compute the degeneracy of each sector based on its representation index. 

  for(int sector_index = 0; sector_index < sector_no; sector_index++){

    int sector_degeneracy=1; 
    
    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

      if (symmetry_types[symmetry_index].get_abelian_tag()==0){
	sector_degeneracy *= (sectors_properties_array[sector_index].representation_index[symmetry_index] + 1);
      }//if

    }//for symmetry_index
    sectors_properties_array[sector_index].degeneracy = sector_degeneracy; 
  }//sector_index

  /*
  for (int sector_index=0; sector_index<sector_no; sector_index++){
    cout<<sector_index<<"\t"
	<<sectors_properties_array[sector_index].representation_index
	<<"\t"<<sectors_properties_array[sector_index].first_state
	<<"\t"<<sectors_properties_array[sector_index].dimension
	<<"\t"<<sectors_properties_array[sector_index].degeneracy<<endl;
  }
  */

  return; 

}//set_sectors_representation_index


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


int  get_local_degeneracy(int symmetry_no, 
			  symmetry* symmetry_types, 

			  int sector_no, 
			  sector_properties* sectors_properties_array){

/*******************************************************************************
input variables: four
     symmetry_no              - the number of symmetries in the input.dat file. 
     symmetry_types           - the types of the symmetries that are U(1), 
                                SU(2) etc. 
     sector_no                - the number of sectors. This has to be computed 
                                first 
     sectors_properties_array - array with the states that has to be organized 
                                into sectors. 

output variables: none

return: integer
     The routine returns the local degeneracy of a added site to the NRG chain. 

What it does:
     This routine returns the degeneracy of a site that is added to the NRG chain. 

How to call:
*******************************************************************************/

  int degeneracy = 0; 

  for(int sector_index = 0; sector_index < sector_no; sector_index++){
    
    int sector_degeneracy (1);
    
    degeneracy += sectors_properties_array[sector_index].dimension * sectors_properties_array[sector_index].degeneracy;
    
  }//for sector_index

  return degeneracy;

}//get_local_degeneracy


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void create_new_block_states_with_list(int symmetry_no, 
				       symmetry * symmetry_types, 
				       
				       int block_state_no, 
				       state * block_states_array, 

				       int local_state_no,
				       state * local_states_array, 

				       int & new_block_state_no, 
				       state *& new_block_states_array){

/*******************************************************************************
input variables:  seven
     symmetry_no          - the number of symmetries in the input.dat file. 
     symmetry_types       - the types of the symmetries that are U(1), SU(2) 
                            etc. 
     block_states_no    - the number of the current states that 
                            describes the system in the previous iteration. 
     block_states_array - array with the current states. 
     local_state_no       - is the number of the local states of the extra site 
                            that is added to the chain. 
     local_states_array   - the states that describe  the local site. 

output variables: two
     new_block_state_no         - the number of the states that are obtained based 
                            on the current and local states. 
     new_block_states_array     - an array with the new states obtained. 

return: void

What it does:
     This routine creates new states when one extra site is added to the 
     NRG chain, based on the current states and local states. 

How to call:
*******************************************************************************/
  
  new_block_state_no = 0; 
  int trial_index [symmetry_no]; 
  state trial_state;  
  trial_state.activate(symmetry_no);
  linked_state_list * new_block_state_list = NULL;

  new_block_state_list = new linked_state_list;


  int index =0;

  // We allocate space for the new base using a linked list
  //     since we do not know yet the number of new states_array.
  //     We will destroy the linked list as soon
  //    as we will construct the array of the new states_array. 

  //   We run through the possible trial states_array
  //     to check whether they are part of the new base.
  //     The new base is built from the current states_array and the local states_array. 


  for(int block_state_index = 0; block_state_index < block_state_no; block_state_index++){
    for(int local_state_index = 0; local_state_index < local_state_no; local_state_index++){
      for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++) {
	trial_index[symmetry_index]=symmetry_types[symmetry_index].get_min_representation_index();
	trial_state.set_representation_index(symmetry_index, symmetry_types[symmetry_index].get_min_representation_index());
      }//symmetry_index


      if(grand_selection_rule(symmetry_no, 
			      symmetry_types, 
			      block_states_array[block_state_index].get_representation_index(), 

			      local_states_array[local_state_index].get_representation_index(), 
			      trial_state.get_representation_index())){

	trial_state.set_parent_state_index(block_state_index);
	trial_state.set_local_state_index(local_state_index);
	new_block_state_list->add_state(trial_state);
	new_block_state_no++;

      }//if grand_selection

      for(int tries = 0; tries < (trial_state_no - 1); tries++) {
	index = 0;
	++trial_index[index];

	while(trial_index[index] > symmetry_types[index].get_max_representation_index() && (index < symmetry_no)){
	  trial_index[index] = symmetry_types[index].get_min_representation_index();
	  if(++index < symmetry_no){
	    ++trial_index[index];
	  }//if
	}//while

	for(int j = 0; j < symmetry_no; j++){
	  trial_state.set_representation_index(j, trial_index[j]);
	}//for j 

	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				block_states_array[block_state_index].get_representation_index(), 
				local_states_array[local_state_index].get_representation_index(), 
				trial_state.get_representation_index())){
	  trial_state.set_parent_state_index(block_state_index);
	  trial_state.set_local_state_index(local_state_index);
	  new_block_state_list->add_state(trial_state);
	  new_block_state_no++;
	}//if 
      }//tries
    }//local_state_index
  }//block_state_index

  
  new_block_states_array = new state[new_block_state_no]; 

  
  new_block_states_array[0] = new_block_state_list->get_state(new_block_state_list->first_ptr);
  state_item * next_state = new_block_state_list->get_next_ptr(new_block_state_list->first_ptr);

  for(int i = 1; i < new_block_state_no; i++){
      new_block_states_array[i] = new_block_state_list->get_state(next_state);
      next_state = new_block_state_list->get_next_ptr(next_state);
  }//i

  delete new_block_state_list; 
  new_block_state_list= NULL;

  return; 
 
  
}//create_new_block_states_with list.


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void create_new_block_hamiltonian(int iteration_index, 

				  int symmetry_no, 
				  symmetry * symmetry_types, 
				  
				  int new_block_sector_no, 
				  sector_properties * new_block_sectors_properties_array, 
				  
				  quantum_operator & new_block_hamiltonian,
				  
				  state * new_block_states_array,
				  state * block_states_array, 
				  state * local_states_array, 
				  
				  quantum_operator &  block_hamiltonian, 
				  
				  int hopping_operator_no,
				  double **  hopping_couplings_array, 
				  double **  on_site_energies_array,

				  quantum_operator* block_hopping_operators, 
				  quantum_operator* local_hopping_operators, 				  
				  quantum_operator* local_hopping_hamiltonians,

				  sector_properties  *local_sectors_properties_array){
  
/*******************************************************************************
input variables:  sixteen
     iteration_index                - represents the block_iteration number 
                                      in the NRG procedure.

     symmetry_no                    - the number of symmetries in the input.dat 
                                      file. 
     symmetry_types                 - the types of the symmetries that are U(1),
                                      SU(2) etc. 

     new_block_sector_no                  - the number of sector in which the 
                                      new_block_states_array are organized. 
     new_block_sectors_properties_array   - an array with the properties of the 
                                      new_block_sectors. 

     new_block_states_array          - array with the new states. 
     block_states_array             - array with the current states
     local_states_array             - array with the local states. 
     block_states_array             - array with the current states. 
     local_state_no                 - is the number of the local states of the 
                                      extra site that is added to the chain. 
     local_states_array             - the states that describe  the local site. 

     block_hamiltonian              - the Hamiltonian of the systems in the 
                                      current representation before one extra 
                                      site is added.

     hopping_operator_no            - the number of the hopping operators. this
                                      number is readed from the input.dat file. 
     hopping_coupling_array         - array with values of the couplings
     on_site_energies_array         - array with the values of the on-site energies. 


     block_hopping_operators      - array with the current hopping operators
     local_hopping_operators        - array with the local hopping operators.

     local_sectors_properties_array - array with properties of the local 
                                      sectors. 

output variables: one
     new_block_hamiltonian - operator that descibed the new Hamiltonian with the 
     extra site added. 

return: void

What it does:
     This routine creates a new Hamiltonian matrix when one extra site is 
     added to the NRG chain.  

How to call:
*******************************************************************************/

  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(block_hamiltonian.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the old Hamiltonian.\nError in routine create_new_block_hamiltonian\nExiting...\n");
    }//if
  }//for symmetry_index

  new_block_hamiltonian.activate(symmetry_no); 
  new_block_hamiltonian.set_name("new_block_Hamiltonian"); 
  new_block_hamiltonian.allocate_sectors_matrix(new_block_sector_no); 

  //we set the representation indices of the new_block_Hamiltonian to be zero's. 

  for(int i = 0; i < symmetry_no; i++){
    new_block_hamiltonian.set_representation_index(i,0); 
  }//i

  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
    if(new_block_hamiltonian.get_representation_index(symmetry_index) != 0){
      error_message("checking the quantum numbers of the new Hamiltonian.\nError in routine create_new_block_hamiltonian\nExiting...\n");
    }//if
  }//for symmetry_index


  //  double xi = (1. - pow(lambda, - iteration_index - 1.)) / sqrt(1. - pow(lambda, - 2. * iteration_index - 1.)) / sqrt(1. - pow(lambda, - 2. * iteration_index - 3.));


  for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
    for(int sector_column = 0; sector_column <new_block_sector_no; sector_column++){

      new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
									      new_block_sectors_properties_array[sector_column].dimension);

      if(grand_selection_rule(symmetry_no, 
			      symmetry_types, 
			      new_block_sectors_properties_array[sector_column].representation_index,
			      new_block_hamiltonian.get_representation_index(), 
			      new_block_sectors_properties_array[sector_row].representation_index)){


	new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].allocate_matrix();

	int height = new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].get_height();
	int width = new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].get_width();

	//now we work within the sector to set up the matrix elements. 

	for(int i=0; i<height; i++){
	  for(int j=0; j<width; j++){

	    
	    double new_block_hamiltonian_matrix_element =0;

	    //   If in the sector (sector_row, sector_column), 
	    //   the element (i,j) is in the diagonal
	    //   then we will import the appropriate matrixelement from the diagonal
	    //   of the old Hamiltonian (stored in block_sectors), namely, the one
	    //   that belongs to the parent state of the i^th (j^th) new_block_state. Then we
	    //   multiply it by sqrt_lambda and place it in the diagonal of new_block_sectors. 
	    
	    
	    int state_row = new_block_sectors_properties_array[sector_row].first_state + i;
	    int state_column = new_block_sectors_properties_array[sector_column].first_state + j;

	    int parent_row = new_block_states_array[state_row].get_parent_state_index();
	    int parent_column = new_block_states_array[state_column].get_parent_state_index();

	    int parent_sector_row = block_states_array[parent_row].get_sector();
	    int parent_sector_column = block_states_array[parent_column].get_sector();

	    int parent_sector_index_row = block_states_array[parent_row].get_sector_index();
	    int parent_sector_index_column = block_states_array[parent_column].get_sector_index();
        
	    if(state_row == state_column){
	      
	      // We already know from the check of the grand_selection_rule(...) that
	      //sector_row = sector_column 
	      
	      new_block_hamiltonian_matrix_element += (block_hamiltonian.get_sector_matrix()[parent_sector_row][parent_sector_column]
						       .get_matrix()[parent_sector_index_row][parent_sector_index_column])*sqrt_lambda;
	    }//if (state_row==state_column)
	   

	    int local_row = new_block_states_array[state_row].get_local_state_index();
	    int local_column = new_block_states_array[state_column].get_local_state_index();

	    int local_sector_row = local_states_array[local_row].get_sector();
	    int local_sector_column = local_states_array[local_column].get_sector();

	    int local_sector_index_row = local_states_array[local_row].get_sector_index();
	    int local_sector_index_column = local_states_array[local_column].get_sector_index();


	    //we add here the on-site energy terms. 
	    //we will import the appropiate matrix element from the local_hopping_hamiltonian 
	    //the one that belong to the local_state of the i'th (j'th) new block_state. 

	    
	    if((parent_sector_row==parent_sector_column)
	       &&(parent_sector_index_row==parent_sector_index_column)
	       &&(local_sector_row==local_sector_column)){

	      //this corresponds to the on-site hopping term that multiplies with the calculated epsilon_n from the density of states. 
	      for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

		  /* Notice that the hopping term in iteration N in the rescaled Hamiltonian 
		     is xi_{N-1} by Wilson's convention !!!!!!!!!!!! */ 
		new_block_hamiltonian_matrix_element += on_site_energies_array[iteration_index][hopping_index]*
		  local_hopping_hamiltonians[hopping_index].get_sector_matrix()[local_sector_row][local_sector_column].get_matrix()[local_sector_index_row][local_sector_index_column];       
		
	      }//for hopping_index; 
	      

	      //this corresponds to the aditional on-site Hamiltonian terms such as the superconducting correlations etc. 
	      if(local_coupling_no>0){
		new_block_hamiltonian_matrix_element += local_hamiltonian.get_sector_matrix()[local_sector_row][local_sector_column].get_matrix()[local_sector_index_row][local_sector_index_column];
	      }//if local_coupling_no;
	    }//if
	    


	    
	    for(int hopping_index = 0; hopping_index < hopping_operator_no; hopping_index++){

	      double sum = 0;
	      // We check the selection rules twice: 

	      //at first, we check if they allow a transition between the parent states
              //(stored in block_states) if the added state is f_N;
	      //at second, we check if they allow a transition between the local states
	      //(stored in local_states) if the added state is f_N+1_dagger. 

	      if((grand_selection_rule(symmetry_no, 
				       symmetry_types, 
				       block_states_array[parent_row].get_representation_index(),
				       block_hopping_operators[hopping_index].get_representation_index(), 
				       block_states_array[parent_column].get_representation_index())) 
		 &&	
		 
		 (grand_selection_rule(symmetry_no, 
				       symmetry_types, 
				       local_states_array[local_column].get_representation_index(),
				       local_hopping_operators[hopping_index].get_representation_index(), 
				       local_states_array[local_row].get_representation_index()))){
		
		// We assign to the initial and final Q, S, R their actual value and choose Q_z to be Q. 
		
		int Q[symmetry_no], Q_z[symmetry_no];
		int Si[symmetry_no], Si_z[symmetry_no], Ri[symmetry_no], Ri_z[symmetry_no];
		int Sf[symmetry_no], Sf_z[symmetry_no], Rf[symmetry_no], Rf_z[symmetry_no];
		int mu[symmetry_no], mu_z[symmetry_no]; 

		//Since mu_i and mu_j should be equal in case of SU(2) and
		//charge_SU(2) symmetries we introduce mu[] to reduce the
		//number of variables. 
     
		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		  
		  Q[symmetry_index] = new_block_states_array[state_column].get_representation_index(symmetry_index);
		  Q_z[symmetry_index] = Q[symmetry_index];

		  Ri[symmetry_index] = local_states_array[local_column].get_representation_index(symmetry_index);
		  Si[symmetry_index] = block_states_array[parent_column].get_representation_index(symmetry_index);

		  Rf[symmetry_index] = local_states_array[local_row].get_representation_index(symmetry_index);
		  Sf[symmetry_index] = block_states_array[parent_row].get_representation_index(symmetry_index);

		  mu[symmetry_index] = local_hopping_operators[hopping_index].get_representation_index(symmetry_index);
		  mu_z[symmetry_index] = mu[symmetry_index];
		}//for symmetry_index




		// We calculate the number of possible terms coming from the hopping operators &
		//the number of possible initial/final states_array. 


		int possible_hopping_term_no = 1;

		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++) {
		  if(symmetry_types[symmetry_index].get_abelian_tag()==0){
		    possible_hopping_term_no *= mu[symmetry_index] + 1;
		  }//if
		}//symmetry_ind;

		for(int term_index = 0; term_index < possible_hopping_term_no; term_index++){

		  // define the following variables to determine the number of states_array to the given mu_z set 

		  int max_Si_z[symmetry_no], min_Si_z[symmetry_no];
		  int state_no_from_Si[symmetry_no];
		  int state_no(1);


		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){		    
		    state_no_from_Si[symmetry_index] = 1;
		  }//int symmetry_index;

		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

		    if(symmetry_types[symmetry_index].get_abelian_tag()==0){
		      max_Si_z[symmetry_index] = min(Si[symmetry_index], 
						     min(Q_z[symmetry_index] + Ri[symmetry_index], 
							 min(Sf[symmetry_index] + mu_z[symmetry_index],
							     Q_z[symmetry_index] + mu_z[symmetry_index] + Rf[symmetry_index])));

		      min_Si_z[symmetry_index] = max(- Si[symmetry_index], 
						     max(Q_z[symmetry_index] - Ri[symmetry_index], 
							 max(- Sf[symmetry_index] + mu_z[symmetry_index],
							     Q_z[symmetry_index] + mu_z[symmetry_index] - Rf[symmetry_index])));
		      
		      state_no_from_Si[symmetry_index] *= (int)((max_Si_z[symmetry_index] - min_Si_z[symmetry_index]) / 2) + 1;
		    }//if 

		  }//for symmetry_index

		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    if(state_no_from_Si[symmetry_index] < 1){
		      state_no = 0;
		    }//if
		    state_no *= state_no_from_Si[symmetry_index];
		  }//for symmetry_index



		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    if(symmetry_types[symmetry_index].get_abelian_tag()==0){
		      Si_z[symmetry_index] = max_Si_z[symmetry_index];
		      Ri_z[symmetry_index] = Q_z[symmetry_index] - Si_z[symmetry_index];
		      Sf_z[symmetry_index] = Si_z[symmetry_index] - mu_z[symmetry_index];
		      Rf_z[symmetry_index] = Q_z[symmetry_index] + mu_z[symmetry_index] - Si_z[symmetry_index];
		    }//if
		  }//for symmetry_index

		  for(int state_index = 0; state_index < state_no; state_index++){
		    double CG_4_state_new (1);
		    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		      if(symmetry_types[symmetry_index].get_abelian_tag()==0){

		      CG_4_state_new *= clebsch_gordan(Si[symmetry_index],Si_z[symmetry_index],Ri[symmetry_index],Ri_z[symmetry_index],Q[symmetry_index],Q_z[symmetry_index]);
			CG_4_state_new *= clebsch_gordan(Sf[symmetry_index],Sf_z[symmetry_index],Rf[symmetry_index],Rf_z[symmetry_index],Q[symmetry_index],Q_z[symmetry_index]);
			CG_4_state_new *= clebsch_gordan(Sf[symmetry_index],Sf_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Si[symmetry_index],Si_z[symmetry_index]);
			CG_4_state_new *= clebsch_gordan(Ri[symmetry_index],Ri_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Rf[symmetry_index],Rf_z[symmetry_index]);
		      }//if 

		      // We would correct for the sign of the hopping operator,
		      //but at the available types of symmetries we do not need to. 

		    }//for symmetry_index

		    sum += CG_4_state_new;

		    // running through the possible decompositions of the states_array,
		    //preparing the next possible distribution of the quantum numbers 

		    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		     

		      while(symmetry_types[symmetry_index].get_abelian_tag() &&
			    symmetry_index < (symmetry_no - 1)){
			symmetry_index++;
		      }//while

		      if(Si_z[symmetry_index] > min_Si_z[symmetry_index]){
			Si_z[symmetry_index] -= 2;
			Ri_z[symmetry_index] = Q_z[symmetry_index] - Si_z[symmetry_index];
			Sf_z[symmetry_index] = Si_z[symmetry_index] - mu_z[symmetry_index];
			Rf_z[symmetry_index] = Q_z[symmetry_index] + mu_z[symmetry_index] - Si_z[symmetry_index];
			symmetry_index = symmetry_no;
		      }//if
		      else{
			Si_z[symmetry_index] = max_Si_z[symmetry_index];
			Ri_z[symmetry_index] = Q_z[symmetry_index] - Si_z[symmetry_index];
			Sf_z[symmetry_index] = Si_z[symmetry_index] - mu_z[symmetry_index];
			Rf_z[symmetry_index] = Q_z[symmetry_index] + mu_z[symmetry_index] - Si_z[symmetry_index];
		      }//else
		    }//symmetry_index
		  }//state_index

		  // running through the possible inner quantum numbers of the hopping operator 
		  
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

		    while(symmetry_types[symmetry_index].get_abelian_tag() &&
			  symmetry_index < (symmetry_no - 1)){
		      
		      symmetry_index++;

		    }//while

		    if(mu_z[symmetry_index] > (-mu[symmetry_index])){
		      mu_z[symmetry_index] -= 2;
		      symmetry_index = symmetry_no;
		    }//if

		    else{
		      mu_z[symmetry_index] = mu[symmetry_index];
		    }//else
		    
		  }//int symmetry_index	       
		}//for term_index
		
		sum *= block_hopping_operators[hopping_index].get_sector_matrix()[parent_sector_column][parent_sector_row].get_matrix()[parent_sector_index_column][parent_sector_index_row];
		sum *= local_hopping_operators[hopping_index].get_sector_matrix()[local_sector_row][local_sector_column].get_matrix()[local_sector_index_row][local_sector_index_column];
		



		//  In case we have charge SU(2) symmetries the matrixelements of the hopping operators when the electron
		//  numbers decrease receive alternating signs with respect to the iteration number. It has to be taken into
		//  account in the construction of the Hamiltonian and in that of the hopping operator. So we have to determine
		//  whether at this matrixelement the electron number has decreased or not (in case of charge SU(2) symmetries). 
		
		if((iteration_index % 2) == 0) {
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    if(!strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)")&& 
		       (block_hopping_operators[hopping_index].get_sgn()==-1)) {
		      if(local_sectors_properties_array[local_sector_column].representation_index[symmetry_index] > local_sectors_properties_array[local_sector_row].representation_index[symmetry_index]){
			sum *= -1;
		      }//if local
		    }//if strcmp
		  }// for symmetry_index
		}//if iteration
		
		sum *= hopping_couplings_array[iteration_index-1][hopping_index];
		sum *= -local_states_array[local_column].get_sgn()*block_hopping_operators[hopping_index].get_sgn();               
	      }//if grand_selection_rule
	      new_block_hamiltonian_matrix_element += sum;

	    }//for int hopping_index

	    

	    if((new_block_hamiltonian_matrix_element != 0)) {
	      new_block_hamiltonian.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i,j, new_block_hamiltonian_matrix_element); 

	      // One more term (namely, the one at the transposed position)
	      //comes from the adjoint of (~f_N+1^dagger * f_N). */
	      // There is a problem if the hopping operators have diagonal matrix elements. XXX 
	      new_block_hamiltonian.get_sector_matrix()[sector_column][sector_row].set_matrix_element(j,i,new_block_hamiltonian_matrix_element); 
	    }//if 	  	    
	  }//j
	}//i
      }//if grand_selection_rule
    }//sector_column
  }//sector_row
  return;

}//create_new_block_hamiltonian


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void create_new_block_hopping_operator(int iteration_index,
				       
				       int symmetry_no, 
				       symmetry * symmetry_types, 
				  
				       int new_block_sector_no, 
				       
				       sector_properties * new_block_sectors_properties_array, 
				       
				       quantum_operator & new_block_hopping_operator, 
				       
				       state * new_block_states_array,
				       state * block_states_array, 
				       state * local_states_array, 
				       
				       quantum_operator & block_hopping_operator, 
				       quantum_operator & local_hopping_operator,
				       
				       sector_properties * local_sectors_properties_array
				       ){
  
/*******************************************************************************

input variables:  two
     iteration_index              - represents the block_iteration number in 
                                    the NRG procedure.

     symmetry_no                  - the number of symmetries in the input.dat 
                                    file. 
     symmetry_types               - the types of the symmetries that are U(1),
                                     SU(2) etc. 

     new_block_sector_no                - the number of sector in which the 
                                    new_block_states_array are organized. 
     new_block_sectors_properties_array - an array with the properties of the 
                                    new_block_sectors. 


     new_block_states_array             - array with the new states. 
     block_states_array         - array with the current states
     local_states_array           - array with the local states. 
     block_states_array         - array with the current states. 
     local_state_no               - is the number of the local states of the 
                                    extra site that is added to the chain. 
     local_states_array           - the states that describe  the local site. 
     block_hopping_operator     - the matrix for the hopping operator in the
                                    current basis.
     local_hopping_operator       - the local operator 

     hopping_operator_no          - the number of the hopping operators. this 
                                    number is readed from the input.dat file. 
     block_hopping_operators    - array with the current hopping operators
     local_hopping_operators      - array with the local hopping operators.

     local_sectors_properties_array - array with properties of the local sectors. 

output variables: one
     new_block_hopping_operator         - hopping operator in the new basis  made of 
                                    the new_block_states_array. 

return: void

What it does:
     This routine creates a new hopping operators based on the 
     current one and the local one when one extra site is added to the 
     NRG chain.  

How to call:
*******************************************************************************/
  
  new_block_hopping_operator.activate(symmetry_no); 
  new_block_hopping_operator.set_name(block_hopping_operator.get_name()); 
  new_block_hopping_operator.allocate_sectors_matrix(new_block_sector_no); 
  new_block_hopping_operator.set_sgn(block_hopping_operator.get_sgn());

  for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
    int representation_index = block_hopping_operator.get_representation_index(symmetry_index); 
    new_block_hopping_operator.set_representation_index(symmetry_index, representation_index);
  }//for symmetry_index;


  for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
    for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++){

      new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
									      new_block_sectors_properties_array[sector_column].dimension);

      if(grand_selection_rule(symmetry_no, 
			      symmetry_types, 
			      new_block_sectors_properties_array[sector_column].representation_index,
			      new_block_hopping_operator.get_representation_index(), 
			      new_block_sectors_properties_array[sector_row].representation_index)){


	new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();

	int height = new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	int width = new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].get_width();


	for(int i=0; i<height; i++){
	  for(int j=0; j<width; j++){

	    
	    int state_row = new_block_sectors_properties_array[sector_row].first_state + i;
	    int state_column = new_block_sectors_properties_array[sector_column].first_state + j;

	    int parent_row = new_block_states_array[state_row].get_parent_state_index();
	    int parent_column = new_block_states_array[state_column].get_parent_state_index();
	    
	    int parent_sector_row = block_states_array[parent_row].get_sector();
	    int parent_sector_column = block_states_array[parent_column].get_sector();

	    int parent_sector_index_row = block_states_array[parent_row].get_sector_index();
	    int parent_sector_index_column = block_states_array[parent_column].get_sector_index();
        

	    
	    if(parent_row==parent_column){
	      
	      int local_row = new_block_states_array[state_row].get_local_state_index();
	      int local_column = new_block_states_array[state_column].get_local_state_index();
	      
	      int local_sector_row = local_states_array[local_row].get_sector();
	      int local_sector_column = local_states_array[local_column].get_sector();
	      
	      int local_sector_index_row = local_states_array[local_row].get_sector_index();
	      int local_sector_index_column = local_states_array[local_column].get_sector_index();

	      // We check whether a transition from the initial local state to the final
	      //local state is possible by means of the given hopping operator. 

	      if(grand_selection_rule(symmetry_no, 
				      symmetry_types, 
				      local_states_array[local_column].get_representation_index(),
				      local_hopping_operator.get_representation_index(), 
				      local_states_array[local_row].get_representation_index())){

		// We assign to the initial and final Q, S, R their actual value and choose
		//Qf_z to be Qf, mu_z to be mu and so Qi_z to be Qf_z - mu_z. */


		int Qi[symmetry_no], Qi_z[symmetry_no], Qf[symmetry_no], Qf_z[symmetry_no];
		int S[symmetry_no], S_z[symmetry_no];
		int Ri[symmetry_no], Ri_z[symmetry_no], Rf[symmetry_no], Rf_z[symmetry_no];
		int mu[symmetry_no], mu_z[symmetry_no]; /* to store the Q numbers of f_N+1_dagger */
		int max_S_z[symmetry_no], min_S_z[symmetry_no];
		int state_no_from_S[symmetry_no];
		int all_state_no(1);


		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++) {

		  Qf[symmetry_index] = new_block_states_array[state_row].get_representation_index(symmetry_index);
		  Qi[symmetry_index] = new_block_states_array[state_column].get_representation_index(symmetry_index);
		  mu[symmetry_index] = local_hopping_operator.get_representation_index(symmetry_index);
		  Qf_z[symmetry_index] = Qf[symmetry_index];
		  Qi_z[symmetry_index] = Qi[symmetry_index];
		  mu_z[symmetry_index] = Qf[symmetry_index] - Qi[symmetry_index];
		  
		  // Above, we have fixed quantum numbers so that
		  //Clebsch_Gordan(Qi[symmetry_index],Qi_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index])
		  //is not equal to zero. Therefore we are left with 3 inner quantum numbers.
		  //One of them could be chosen arbitrarily, the remaining 2 are constrained. 


		  Ri[symmetry_index] = local_states_array[local_column].get_representation_index(symmetry_index);
		  Rf[symmetry_index] = local_states_array[local_row].get_representation_index(symmetry_index);
		  S[symmetry_index] = block_states_array[parent_column].get_representation_index(symmetry_index);
		  state_no_from_S[symmetry_index] = 1;

		}//for symmetry_index

		// Determination of the Number of Possible Initial and Final States 

                for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++) {
		  if(symmetry_types[symmetry_index].get_abelian_tag()==0){

		    max_S_z[symmetry_index] = min(S[symmetry_index], min(Qf_z[symmetry_index] + Rf[symmetry_index], Qi_z[symmetry_index] + Ri[symmetry_index]));
		    min_S_z[symmetry_index] = max(- S[symmetry_index], max(Qf_z[symmetry_index] - Rf[symmetry_index], Qi_z[symmetry_index] - Ri[symmetry_index]));
		    state_no_from_S[symmetry_index] *= (int)((max_S_z[symmetry_index] - min_S_z[symmetry_index]) / 2) + 1;		    
		    
		  }//if
		}//for
		
		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		  if(state_no_from_S[symmetry_index] < 1)
		    all_state_no = 0;
		  all_state_no *= state_no_from_S[symmetry_index];
		}//for symmetry_index
		

		// initializing the quantum numbers 

		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		  if(symmetry_types[symmetry_index].get_abelian_tag()==0){
		    S_z[symmetry_index] = max_S_z[symmetry_index];
		    Ri_z[symmetry_index] = Qi_z[symmetry_index] - S_z[symmetry_index];
		    Rf_z[symmetry_index] = Qf_z[symmetry_index] - S_z[symmetry_index];
		  }//if
		}//for symmetry_index



		double hopping_matrix_element = 0;

		for(int state_index = 0; state_index < all_state_no; state_index++){
		  double CG_4_state = 1;
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

		    if(symmetry_types[symmetry_index].get_abelian_tag()==0){
		      CG_4_state *= clebsch_gordan(Ri[symmetry_index],Ri_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Rf[symmetry_index],Rf_z[symmetry_index]);
		      CG_4_state *= clebsch_gordan(S[symmetry_index],S_z[symmetry_index],Rf[symmetry_index],Rf_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index]);
		      CG_4_state *= clebsch_gordan(S[symmetry_index],S_z[symmetry_index],Ri[symmetry_index],Ri_z[symmetry_index],Qi[symmetry_index],Qi_z[symmetry_index]);
		    }//if		    
		  }//for symmetry_inded
		  hopping_matrix_element += CG_4_state;

		  /* running through the decompositions of the states */

		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++) {
		      while(symmetry_types[symmetry_index].get_abelian_tag() &&
			    symmetry_index < (symmetry_no - 1))
			symmetry_index++;

		      if(S_z[symmetry_index] > min_S_z[symmetry_index])
			{
			  S_z[symmetry_index] -= 2;
			  Ri_z[symmetry_index] = Qi_z[symmetry_index] - S_z[symmetry_index];
			  Rf_z[symmetry_index] = Qf_z[symmetry_index] - S_z[symmetry_index];
			  symmetry_index = symmetry_no;
			}
		      else
			{
			  S_z[symmetry_index] = max_S_z[symmetry_index];
			  Ri_z[symmetry_index] = Qi_z[symmetry_index] - S_z[symmetry_index];
			  Rf_z[symmetry_index] = Qf_z[symmetry_index] - S_z[symmetry_index];
			}
		    }



		}//state_index


		for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		  if(symmetry_types[symmetry_index].get_abelian_tag() ==0)
		    hopping_matrix_element /= clebsch_gordan(Qi[symmetry_index],Qi_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index]);
		}//for symmetry index

		hopping_matrix_element *= local_hopping_operator.get_sector_matrix()[local_sector_row][local_sector_column].get_matrix()[local_sector_index_row][local_sector_index_column];

		// In case we have charge SU(2) symmetries the matrixelements of the hopping operators when the electron
		//   numbers decrease receive alternating signs with respect to the iteration number. It has to be taken into
		//   account in the construction of the Hamiltonian and in that of the hopping operator. So we have to determine
		//   whether at this matrixelement the electron number has decreased or not (in case we use charge SU(2) symmetries). 

		if((iteration_index % 2) == 0){
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    if(!strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)")
		       && (block_hopping_operator.get_sgn()==-1)) {
		      if(local_sectors_properties_array[local_sector_column].representation_index[symmetry_index] > 
			 local_sectors_properties_array[local_sector_row].representation_index[symmetry_index])

			hopping_matrix_element *= -1;
		    }//if
		  }//for symmetry
		}//if iteration.

		/* The next multiplication is needed only if we calculate 1 channel Kondo with a charge SU(2)
		   and a spin SU(2) symmetry. */

		if((local_sector_row == 0) && (local_sector_column == 1) && (local_sector_index_row == 0) && (local_sector_column == 0)
		  && ((iteration_index % 2) == 0)
		   && (!strcmp(symmetry_types[0].get_name(),"charge_SU(2)")) && (!strcmp(symmetry_types[1].get_name(),"SU(2)")))
		  {
		    hopping_matrix_element *= -1;
		  }

		new_block_hopping_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, hopping_matrix_element);             	       

	      }//if grand_selection
	    }//if parent_row==parent_column	    
	  }//for j
	}//for i
      }//if grand_selection_rule
    }//sector_column
  }//sector_row

  return; 
}//create_new_block_hopping_operator


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void create_new_block_spectral_operator(int symmetry_no, 
					symmetry * symmetry_types, 
					
					int new_block_sector_no, 
					sector_properties * new_block_sectors_properties_array,
					
					quantum_operator & new_block_spectral_operator, 
					
					state * new_block_states_array,
					state * block_states_array, 
					state * local_states_array, 
					
					quantum_operator & block_spectral_operator){
  
/*******************************************************************************
input variables:  eight
     symmetry_no                  - the number of symmetries in the input.dat 
                                    file. 
     symmetry_types               - the types of the symmetries that are U(1),
                                    SU(2) etc. 

     new_block_sector_no                - the number of sector in which the 
                                    new_block_states_array are organized. 
     new_block_sectors_properties_array - an array with the properties of the 
                                    new_block_sectors. 

     new_block_states_array             - array with the new states. 
     block_states_array         - array with the current states
     local_states_array           - array with the local states. 

     block_spectral_operators   - array with the current spectral operator

output variables: one
     new_block_spectral_operator        - spectral operator in the new basis  made of
                                    the new_block_states_array. 

return: void

What it does:
     This routine creates a new spectral operators. The input matrices for 
     these operators are in the input.dat file. The spectral functions 
     are computed for all these operators.

How to call:
*******************************************************************************/


  new_block_spectral_operator.activate(symmetry_no); 
  new_block_spectral_operator.set_name(block_spectral_operator.get_name());
  new_block_spectral_operator.allocate_sectors_matrix(new_block_sector_no); 
  new_block_spectral_operator.set_sgn(block_spectral_operator.get_sgn());

  for(int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
    int representation_index = block_spectral_operator.get_representation_index(symmetry_index);
    new_block_spectral_operator.set_representation_index(symmetry_index, representation_index);    
  }//for symmetry_index;


    for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++){

	new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_size(new_block_sectors_properties_array[sector_row].dimension,
										     new_block_sectors_properties_array[sector_column].dimension);


	if(grand_selection_rule(symmetry_no, 
				symmetry_types, 
				new_block_sectors_properties_array[sector_column].representation_index,
				new_block_spectral_operator.get_representation_index(), 
				new_block_sectors_properties_array[sector_row].representation_index)){

	  

	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].activate(symmetry_no);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(0, new_block_sectors_properties_array[sector_row].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_representation_index(1, new_block_sectors_properties_array[sector_column].representation_index);
	  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].allocate_matrix();
	  
	  int height = new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_height();
	  int width = new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].get_width();

	  for(int i=0; i<height; i++){
	    for(int j=0; j<width; j++){
	      
	      int state_row = new_block_sectors_properties_array[sector_row].first_state + i;
	      int state_column = new_block_sectors_properties_array[sector_column].first_state + j;
	      
	      int local_row = new_block_states_array[state_row].get_local_state_index();
	      int local_column = new_block_states_array[state_column].get_local_state_index();
	      
	      int local_sector_row = local_states_array[local_row].get_sector();
	      int local_sector_column = local_states_array[local_column].get_sector();
	      
	      int local_sector_index_row = local_states_array[local_row].get_sector_index();
	      int local_sector_index_column = local_states_array[local_column].get_sector_index();
	      
	      if(local_row==local_column){

		
		int parent_row = new_block_states_array[state_row].get_parent_state_index();
		int parent_column = new_block_states_array[state_column].get_parent_state_index();
		
		int parent_sector_row = block_states_array[parent_row].get_sector();
		int parent_sector_column = block_states_array[parent_column].get_sector();
		
		int parent_sector_index_row = block_states_array[parent_row].get_sector_index();
		int parent_sector_index_column = block_states_array[parent_column].get_sector_index();
	      
		
		if(grand_selection_rule(symmetry_no, 
					symmetry_types, 
					block_states_array[parent_column].get_representation_index(),
					block_spectral_operator.get_representation_index(), 
					block_states_array[parent_row].get_representation_index())){
		  
		  
		  
		  // We assign to the initial and final Q, S, R their actual value and choose
		  //   Qf_z to be Qf, mu_z to be mu and so Qi_z to be Qf_z - mu_z. 

		  int Qi[symmetry_no], Qi_z[symmetry_no], Qf[symmetry_no], Qf_z[symmetry_no];
		  int Si[symmetry_no], Si_z[symmetry_no], Sf[symmetry_no], Sf_z[symmetry_no];
		  int R[symmetry_no], R_z[symmetry_no];
		  int max_Sf_z[symmetry_no], min_Sf_z[symmetry_no], max_Si_z[symmetry_no], min_Si_z[symmetry_no];
		  int max_R_z[symmetry_no], min_R_z[symmetry_no];
		  int mu[symmetry_no], mu_z[symmetry_no]; // to store the Q numbers of the local operator 
		  
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    
		    Qf[symmetry_index] = new_block_states_array[state_row].get_representation_index(symmetry_index);
		    Qf_z[symmetry_index] = Qf[symmetry_index];
		    Qi[symmetry_index] = new_block_states_array[state_column].get_representation_index(symmetry_index);
		    Qi_z[symmetry_index] = Qi[symmetry_index];
		    mu[symmetry_index] = block_spectral_operator.get_representation_index(symmetry_index);
		    mu_z[symmetry_index] = Qf_z[symmetry_index] - Qi_z[symmetry_index];
		    
		    // Above, we have fixed quantum numbers so that
		    //Clebsch_Gordan(Qi[symmetry_index],Qi_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index])
		    //is not equal to zero. Therefore we are left with 3 inner quantum numbers.
		    //One of them could be chosen arbitrarily, the remaining 2 are constrained. 
		    
		    Si[symmetry_index] = block_states_array[parent_column].get_representation_index(symmetry_index);
		    Sf[symmetry_index] = block_states_array[parent_row].get_representation_index(symmetry_index);
		    R[symmetry_index] = local_states_array[local_row].get_representation_index(symmetry_index);
		    
		  }//for symmetry_index
		  
		  
		  int state_no = 1;
		  int state_no_from_Si[symmetry_no];
		  
		  // Determination of the Maximal & Minimal Inner Quantum Number Values of the Initial and Final Local States 
		  // Determination of the Number of Possible Initial and Final States 
		  
		  
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    state_no_from_Si[symmetry_index] = 1;
		    
		    if(symmetry_types[symmetry_index].get_abelian_tag() == 0){

		      max_Si_z[symmetry_index] = min(min(Si[symmetry_index], Sf[symmetry_index] - mu_z[symmetry_index]), Qi_z[symmetry_index] + R[symmetry_index]);
		      min_Si_z[symmetry_index] = max(max(-Si[symmetry_index], - Sf[symmetry_index] - mu_z[symmetry_index]), Qi_z[symmetry_index] - R[symmetry_index]);
		      state_no_from_Si[symmetry_index] *= (int)((max_Si_z[symmetry_index] - min_Si_z[symmetry_index])/2) + 1;
		      Si_z[symmetry_index] = max_Si_z[symmetry_index];
		      Sf_z[symmetry_index] = Si_z[symmetry_index] + mu_z[symmetry_index];
		      R_z[symmetry_index] = Qi_z[symmetry_index] - Si_z[symmetry_index];

		      //cout << state_no_from_Si[symmetry_index] << endl;
		    }//if
		  }//for symmetry_type
		  
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		    if(state_no_from_Si[symmetry_index] < 1)
		      state_no = 0;
		    state_no *= state_no_from_Si[symmetry_index];
		  }//ofr symmetry
		  
		  double matrix_element = 0; // to store a matrix element for the spectral operator. 
		  
		  
		  for(int state_index = 0; state_index < state_no; state_index++){
		    double CG_4_state = 1;
		    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		      if( symmetry_types[symmetry_index].get_abelian_tag() ==0 ){
			CG_4_state *= clebsch_gordan(Si[symmetry_index],Si_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Sf[symmetry_index],Sf_z[symmetry_index]);
			CG_4_state *= clebsch_gordan(Sf[symmetry_index],Sf_z[symmetry_index],R[symmetry_index],R_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index]);
			CG_4_state *= clebsch_gordan(Si[symmetry_index],Si_z[symmetry_index],R[symmetry_index],R_z[symmetry_index],Qi[symmetry_index],Qi_z[symmetry_index]);
		      }//if
		    }//symmetry_index
		    matrix_element += CG_4_state;
		    
		    // Here we run through the possible quantum number values of the states. 
		    
		    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
		      while(symmetry_types[symmetry_index].get_abelian_tag() &&
			    symmetry_index < (symmetry_no - 1))
			symmetry_index++;

		      if(Si_z[symmetry_index] > min_Si_z[symmetry_index]){
			Si_z[symmetry_index] -= 2;
			Sf_z[symmetry_index] = Si_z[symmetry_index] + mu_z[symmetry_index];
			R_z[symmetry_index] = Qi_z[symmetry_index] - Si_z[symmetry_index];
			symmetry_index = symmetry_no;
		      }//if
		      else{
			Si_z[symmetry_index] = max_Si_z[symmetry_index];
			Sf_z[symmetry_index] = Si_z[symmetry_index] + mu_z[symmetry_index];
			R_z[symmetry_index] = Qi_z[symmetry_index] - Si_z[symmetry_index];
		      }//else
		    }//symmetry_index
		  }//state_index
		  
		  for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

		    if(symmetry_types[symmetry_index].get_abelian_tag()==0)
		      matrix_element /= clebsch_gordan(Qi[symmetry_index],Qi_z[symmetry_index],mu[symmetry_index],mu_z[symmetry_index],Qf[symmetry_index],Qf_z[symmetry_index]);
		  }//for symmetry_index
		  
		  matrix_element *= block_spectral_operator.get_sector_matrix()[parent_sector_row][parent_sector_column].get_matrix()[parent_sector_index_row][parent_sector_index_column];
		  

		  
		  if(block_spectral_operator.get_sgn()==-1){
		    // If we are to calculate fermionic operator than the following multiplication should be performed.	  
		    matrix_element *= local_states_array[local_column].get_sgn();
		  }//i
		  		  				  
		  new_block_spectral_operator.get_sector_matrix()[sector_row][sector_column].set_matrix_element(i, j, matrix_element);             	       
		  
		  
		}//if grand_selection_rule
	      }//if local_row==local_column
	      
	    }//j
	  }//i	  
	  
	}//if grand_selection_rule
      }//sector_column
    }//sector_row
    return; 
}//create_new_block_spectral_operator


void set_hopping_couplings_and_on_site_energies_array(int iteration_no, 
						      int hopping_operator_no,
						      double ** &  hopping_couplings_array, 
						      double ** &  on_site_energies_array){
  
/*******************************************************************************
input variables:  two
   iteration_no     - is the number of iterations
   hopping_operator_no - is the number of hopping operators from the input file. 

output variables: two
    double ** hopping_couplings_array - array with the couplings at each site for each local operator.
    double ** on_site_energies_array  - array with the on-site energies at each site. 

return: none
    

What it does:
    This routine allocates memory and computes/reads the hoppings and the on-site energies at each local site. 
    It computes the "dimensionless rescaled" hoppings xi_N, and stores them in the array hopping_couplings_array. 
    The dimensionful hopping is defined in terms of this as :
          t_{N-1} ( f^dagger_{N-1} f_N + h.c.) = D * xi_{N-1} * 2.0/(1.0+1.0/lambda)*lambda^(N-1)/2
    The on-site energy of iteration N correspods to terms: 
          epsilon_N *  f^dagger_N f_N = D * 2.0/(1.0+1.0/lambda)*lambda^(N-1)/2 * e_N *   f^dagger_N f_N,
     with epsilon_N/e_N the dimensionful/rescaled on-site energies. 


*******************************************************************************/

  //hopping_couplings_array =[-1, iteration_no]; 
  //on_site_energies_array =[0, iteration_no]; 


  //this is the hopping_couplings_array; 

  int no_of_rows;
  int no_of_columns; 

  no_of_rows  = iteration_no+1; 
  no_of_columns = hopping_operator_no;

  hopping_couplings_array = new double * [no_of_rows];
  if (!hopping_couplings_array){
    cout << "allocation failure 1 in set_hopping_couplings_and_on_site_energies\nEXITING..." << endl; 
    exit(1);
  }//if
  
  hopping_couplings_array[0]= new double[no_of_rows * no_of_columns];
  if (!hopping_couplings_array[0]){
    cout << "allocation failure 2 in set_hopping_couplings_and_on_site_energies\nEXITING..." << endl; 
    exit(1);
  }//if
	
  /* Now I set the pointers of the other rows */
  for(int i = 1; i < no_of_rows; i++){
    hopping_couplings_array[i] = hopping_couplings_array[i-1] + no_of_columns;
  }//for i 
  for(int i = 0; i < no_of_rows; i++){
    for(int j = 0; j < no_of_columns; j++){
      hopping_couplings_array[i][j] = 0;
    }//j
  }//i

  //this is the on_site_energies_array; 

  
  no_of_rows  = iteration_no+1; 
  no_of_columns = hopping_operator_no;

  on_site_energies_array = new double * [no_of_rows];
  if (!on_site_energies_array){
    cout << "allocation failure 1 in set_hopping_couplings_and_on_site_energies\nEXITING..." << endl; 
    exit(1);
  }//if
  
  on_site_energies_array[0]= new double[no_of_rows * no_of_columns];
  if (!on_site_energies_array[0]){
    cout << "allocation failure 2 in set_hopping_couplings_and_on_site_energies\nEXITING..." << endl; 
    exit(1);
  }//if
	
  /* Now I set the pointers of the other rows */
  for(int i = 1; i < no_of_rows; i++){
    on_site_energies_array[i] = on_site_energies_array[i-1] + no_of_columns;
  }//for i 
  for(int i = 0; i < no_of_rows; i++){
    for(int j = 0; j < no_of_columns; j++){
      on_site_energies_array[i][j] = 0;
    }//j
  }//i


  if (hoppings_on_site_energies_flag){
    //cout<<"reading_couplings"<<endl;

      /* read in the dimensionful hopping of the unrescaled Hamiltonian. These hoppings are supposed 
	 decay exponentially for large N...
      */
    read_hopping_couplings_and_on_site_energies(hopping_couplings_array,on_site_energies_array);


    /* rescale the hoppings and compute the rescaled hoppings and on-site energies xi_n and e_n */ 
    for(int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
      for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
	hopping_couplings_array [iteration_index][hopping_index]  = 2.0/(1.0+1.0/lambda)*pow(lambda, iteration_index/2.0)
	  *hopping_couplings_array[iteration_index][hopping_index]; 
	on_site_energies_array [iteration_index][hopping_index] = 2.0/(1.0+1.0/lambda)*pow(lambda, iteration_index/2.0)
	  *on_site_energies_array[iteration_index][hopping_index]; 
      }//for hopping_index
    }//for iteration_index

  }//if

  else{
      /* For a flat density we use Wilson's expression */
    for(int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
      for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
	hopping_couplings_array[iteration_index][hopping_index] = (1. - pow(lambda, - iteration_index - 1.)) 
	  / sqrt(1. - pow(lambda, - 2. * iteration_index - 1.))
	  / sqrt(1. - pow(lambda, - 2. * iteration_index - 3.));
	on_site_energies_array[iteration_index][hopping_index]= 0.0; 


    }//for iteration_index
    }
    
  }//else
  return; 
}//set hopping_couplings_and_on_site_energies_array


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/









