/****************************************************************************

Last Update : 2007/09/25 Pascu Moca 

No documentation update necessary


*****************************************************************************/


#include "../include/dmatrix.h"
#include "../include/in_out.h"
#include "../include/llvector.h"
#include "../include/sector.h"
#include "../include/state.h"
#include "../include/state_list.h"
#include "../include/symmetry.h"
#include "../include/variables.h"
#include "../include/quantum_operator.h"
#include "../include/construction.h"
#include "../include/diagonalization.h"
#include "../include/transformation.h"
#include "../include/cut_off.h"
#include "../include/spectral_function.h"
#include "../include/dmnrg.h"
#include "../include/real_axis_spectral_function.h"
#include "../include/hilbert_transform.h"
#include "../include/green_function.h"



enum CHOICE {Generate_spectrum_filter='f',Generate_spectrum_no_filter='e',Representation_index ='g',Representation_index_each_iteration = 'r',Quit='q'};

char do_menu();
bool readln(int & index);
 
int main(){

  char file_name_odd[254]; 
  char file_name_even[254];
  char swap_string_odd[254];
  char swap_string_even[254];
  char swap_string_temp[254];

  read_parameters_section();
  read_flags_section();
  read_symmetries_section();
  read_block_states_section();
  read_local_states_section();
  read_local_states_signs_section();

  
  block_sector_no = get_sectors_no(block_state_no, block_states_array);
  local_sector_no = get_sectors_no(local_state_no, local_states_array); 

  set_sectors_representation_index(symmetry_no,
				   symmetry_types,
				   block_state_no, 
				   block_states_array, 
				   block_sector_no, 
				   block_sectors_properties_array); 

  set_sectors_representation_index(symmetry_no,
				   symmetry_types,
				   local_state_no, 
				   local_states_array, 
				   local_sector_no, 
				   local_sectors_properties_array); 

  local_degeneracy = get_local_degeneracy(symmetry_no, 
					  symmetry_types, 
					  local_sector_no, 
					  local_sectors_properties_array); 

  read_block_hamiltonian_section();  
  read_hopping_operators_section();




  if (spectral_operator_no>0)  read_spectral_operators_section();
  if(static_operator_no>0)  read_static_operators_section();
  if (local_coupling_no>0) read_local_hamiltonian_section();

  read_local_hopping_operators_section();
  read_spectral_function_section();
  read_spectral_function_broadening_section();


  create_working_folder(model,
			lambda, 
			max_state_no,
			iteration_no,
			temperature,		   
			symmetry_no,
			symmetry_types,
			coupling_no, 
			couplings,
			working_folder_name			
			); 

  

  create_swap_folder(working_folder_name, 
		     swap_folder_name); 

  create_delta_peaks_folder(swap_folder_name, 
			    delta_peaks_folder_name); 
  create_spectral_functions_folder(swap_folder_name, 
				   spectral_functions_folder_name); 
  create_transformations_folder(swap_folder_name, 
				transformations_folder_name); 
  create_spectral_operators_folder(swap_folder_name, 
				   spectral_operators_folder_name); 
  create_static_operators_folder(swap_folder_name, 
				 static_operators_folder_name); 
  create_states_folder(swap_folder_name, 
		       states_folder_name); 
  create_sectors_folder(swap_folder_name, 
			sectors_folder_name); 
  create_mapping_folder(mapping_folder_name); 
 
  create_dos_mapping_folder(dos_mapping_folder_name);



  llvector current_representation_index; 
  llvector ground_state_representation_index;
  if(!current_representation_index.get_ptr()){
    current_representation_index.allocate(symmetry_no);    
  } //if
  
  if (!ground_state_representation_index.get_ptr()){
    ground_state_representation_index.allocate(symmetry_no);
  }//if




   int choice;
   int fQuit=false;
   int energy_levels_no;
   ofstream energy_levels_file_odd;
   ofstream energy_levels_file_even; 

   while(!fQuit)
     {
       choice=do_menu();
       
       if(choice != Generate_spectrum_filter && choice !=Generate_spectrum_no_filter && choice != Representation_index && choice != Representation_index_each_iteration  && choice != Quit){
	 cout<<"\n\tInvalid choice. Please, try again.\n\n";
	 continue;
       }
       
       switch(choice)
	 {




	 case Representation_index_each_iteration:
	   

	   m_star=1;

	   
	  	   
	   cout<<"\tIteration number\tRepresentation_index for ";
	   for(int symmetry_index = 0; symmetry_index<symmetry_no; symmetry_index++){
	     cout<<" "<<symmetry_types[symmetry_index].get_name()<<"\t";
	   }
	   cout<<"\n\n"<<endl;
	   


	   for (int iteration_index = m_star; iteration_index<=iteration_no; iteration_index++){

	     read_new_block_states(m_star,
				   iteration_index, 
				   
				   symmetry_no,
				   
				   new_block_state_no,
				   new_block_states_array,
				   new_block_energies_array,
				   sorted_new_block_energies_array,
				   new_block_energies_index
				   );
	     ground_state_representation_index = new_block_states_array[new_block_energies_index[0]].get_representation_index();
	     cout<<"\t\t "<<iteration_index<<"\t\t\t"<<ground_state_representation_index<<endl;
	     
	   }//for iteration_index



	   break; 

	   
	 case Representation_index:

	   
	   m_star=1;

       

	   read_new_block_states(m_star,
				 iteration_no, 
				 
				 symmetry_no,
				 
				 new_block_state_no,
				 new_block_states_array,
				 new_block_energies_array,
				 sorted_new_block_energies_array,
				 new_block_energies_index
				 );
	 

	   ground_state_representation_index = new_block_states_array[new_block_energies_index[0]].get_representation_index();

	   if(div(iteration_no,2).rem==0){

	     cout<<"\n\tGround state representation indices for even iterations:"<<ground_state_representation_index<<"\n"<<endl;

	   }
	   else{

	     cout<<"\n\tGround state representation indices for odd iterations:"<<ground_state_representation_index<<"\n"<<endl;
	   }
	   

	   	   
	   read_new_block_states(m_star,
				 iteration_no-1, 
				 
				 symmetry_no,
				 
				 new_block_state_no,
				 new_block_states_array,
				 new_block_energies_array,
				 sorted_new_block_energies_array,
				 new_block_energies_index
				 );
	 

	   ground_state_representation_index = new_block_states_array[new_block_energies_index[0]].get_representation_index();

	   if(div(iteration_no-1,2).rem==1){

	     cout<<"\n\tGround state representation indices for odd iterations: "<<ground_state_representation_index<<"\n"<<endl;
	   }
	     
	   else{
	     cout<<"\n\tGround state representation indices for even iterations:"<<ground_state_representation_index<<"\n"<<endl;
	   }




	   break; 

	 case Generate_spectrum_filter:
	   printf("Enter the number of energy levels:");
	     readln(energy_levels_no);	    
	 
	   for(int i =0; i<symmetry_no; i++){
	     printf ("Enter the  representation index for the %s symmetry number %d:", symmetry_types[i].get_name(), i);
	     readln(current_representation_index[i] );  
	   }//i

 


	   strcpy(file_name_odd, working_folder_name);
	   
	   sprintf(swap_string_odd, "//energy_levels_odd_"); 
	   
	   for (int i=0; i<symmetry_no; i++){
	     strcat(swap_string_odd, symmetry_types[i].get_name()); 
	     sprintf(swap_string_temp, "_%d_",current_representation_index[i]);
	     strcat(swap_string_odd, swap_string_temp);
	   }//i

	   sprintf (swap_string_temp, "no_%d.dat", energy_levels_no);
	   strcat(swap_string_odd, swap_string_temp);
	   strcat(file_name_odd, swap_string_odd);
	   
	  energy_levels_file_odd.open(file_name_odd, ios::out);
	   

	  energy_levels_file_odd.precision(double_precision);

	   
	   strcpy(file_name_even, working_folder_name);
	   
	   sprintf(swap_string_even, "//energy_levels_even_"); 
	   
	   for (int i=0; i<symmetry_no; i++){
	     strcat(swap_string_even, symmetry_types[i].get_name()); 
	     sprintf(swap_string_temp, "_%d_",current_representation_index[i]);
	     strcat(swap_string_even, swap_string_temp);
	   }//i
	   
	   sprintf (swap_string_temp, "no_%d.dat",energy_levels_no );
	   strcat(swap_string_even, swap_string_temp);
	   strcat(file_name_even, swap_string_even);
	   
	   
	   energy_levels_file_even.open(file_name_even, ios::out); 
	   
	   energy_levels_file_even.precision(double_precision);
	   
	   //we set m_star to zero so all iterations will be readed in
	   
	   m_star=1;
	   
	   
	   
	   for(int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){
	     
	     read_new_block_states(m_star,
				   iteration_index, 
				   
				   symmetry_no,
				   
				   new_block_state_no,
				   new_block_states_array,
				   new_block_energies_array,
				   sorted_new_block_energies_array,
				   new_block_energies_index
				   );
	     
	     
	     //even iterations
	     if(div(iteration_index,2).rem==0){
	       
	       
	       int energy_index=0; 
	       int saved_level = 0; 
	       
	       do {       
		 if(current_representation_index ==  new_block_states_array[new_block_energies_index[energy_index]].get_representation_index()){
		   
		   saved_level ++;
		   if (saved_level ==1){
		     energy_levels_file_even<<iteration_index<<"\t";
		   }
		   energy_levels_file_even<<sorted_new_block_energies_array[energy_index]- sorted_new_block_energies_array[0]<<"\t";
		   
		 }//if      
		 energy_index++;
	       }while((saved_level<energy_levels_no) && (energy_index<new_block_state_no));      
	       energy_levels_file_even<<endl;
	     }//if div
	     
	     
	     if(div(iteration_index,2).rem==1){
	       
	  	       
	       int energy_index=0; 
	       int saved_level = 0; 
	       
	       do {       
		 if(current_representation_index ==  new_block_states_array[new_block_energies_index[energy_index]].get_representation_index()){
		   saved_level ++;
		   if (saved_level ==1){
		     energy_levels_file_odd<<iteration_index<<"\t";
		   }
		   
		   energy_levels_file_odd<<sorted_new_block_energies_array[energy_index]- sorted_new_block_energies_array[0]<<"\t";
		   
		 }//if      
		 energy_index++;
	       }while((saved_level<energy_levels_no) && (energy_index<new_block_state_no));      
	       energy_levels_file_odd<<endl;
	     }//if div
	     
	   }//iteration_index;
	   
	   energy_levels_file_odd.close();
	   energy_levels_file_even.close();

	 
	   break;


	 case Generate_spectrum_no_filter:


	   printf("Enter the number of energy levels:");
	   readln(energy_levels_no);	    
	

	   strcpy(file_name_odd, working_folder_name);
	   strcpy(file_name_even, working_folder_name); 
	   sprintf(swap_string_odd, "//energy_levels_odd_no_%d.dat", energy_levels_no); 
	   strcat(file_name_odd, swap_string_odd);
	   sprintf(swap_string_even, "//energy_levels_even_no_%d.dat", energy_levels_no); 
	   strcat(file_name_even, swap_string_even);
	   

	   energy_levels_file_even.open(file_name_even, ios::out); 
	   energy_levels_file_odd.open(file_name_odd, ios::out);
	   
	   energy_levels_file_even.precision(double_precision); 
	   energy_levels_file_odd.precision(double_precision);

	   //we set m_star to zero so all iterations will be readed in
	   m_star=1;

	   if (energy_levels_no> max_state_no){
	     energy_levels_no=max_state_no;
	   }//if 

	   for(int iteration_index=m_star; iteration_index<=iteration_no; iteration_index++){
    
	     
	     read_new_block_states(m_star,
				   iteration_index, 
				   
				   symmetry_no,
				   
				   new_block_state_no,
				   new_block_states_array,
				   new_block_energies_array,
				   sorted_new_block_energies_array,
				   new_block_energies_index
				   );
	     
	     	     
	   
	     //even iterations
	     if(div(iteration_index,2).rem==0){
	       energy_levels_file_even<<iteration_index<<"\t";
	       for (int energy_level=0; energy_level<energy_levels_no; energy_level++){
		 energy_levels_file_even<<sorted_new_block_energies_array[energy_level]-sorted_new_block_energies_array[0]<<"\t";
	       }//for energy_level
	       energy_levels_file_even<<endl;
	     }//if div
	     
	     if(div(iteration_index,2).rem==1){
	       energy_levels_file_odd<<iteration_index<<"\t";
	       for (int energy_level=0; energy_level<energy_levels_no; energy_level++){
		 energy_levels_file_odd<<sorted_new_block_energies_array[energy_level]-sorted_new_block_energies_array[0]<<"\t";
	       }//for energy_level
	       energy_levels_file_odd<<endl;
	     }//if div
	     	     	    
  }//iteration_index;
    
  energy_levels_file_even.close();
  energy_levels_file_odd.close();








	   break; 


	 case Quit:
	   fQuit=true;
	   cout<<"\n\tEXITING...\n\n";
	   break;
	 default:
	   cout<<"\tError in choice!\n";
	   fQuit=true;
	   break;
	 }//end switch
     }//end while
   

   current_representation_index.clear(); 
   ground_state_representation_index.clear();
  return 0; 
}



char  do_menu()
{
    char buffer[256];
    cout<<"\n\n\t\t\t\t *** MENU *** \n\n\n";
    cout<<"\t(g)\tGet the representation indices for the ground state from the last two iterations\n";
    cout<<"\t(r)\tGet the representation indices for the ground state at each iteration\n";
    cout<<"\t(f)\tGenerate energy spectrum filtered by quantum numbers \n";
    cout<<"\t(e)\tGenerate energy spectrum without filter \n";
    cout<<"\t(q)\tQuit\n\n\t";
    cin.getline(buffer,255,'\n');
    return buffer[0];
}
bool readln(int & index)
{
    char buffer[256];
    cin.getline(buffer,255,'\n');
    index=atoi(buffer);
    return true;
}
