#include "../include/dmatrix.h"
#include "../include/in_out.h"
#include "../include/llvector.h"
#include "../include/sector.h"
#include "../include/state.h"
#include "../include/state_list.h"
#include "../include/symmetry.h"
#include "../include/variables.h"
#include "../include/quantum_operator.h"
#include "../include/construction.h"
#include "../include/diagonalization.h"
#include "../include/transformation.h"
#include "../include/cut_off.h"
#include "../include/spectral_function.h"
#include "../include/dmnrg.h"
#include "../include/hilbert_transform.h"
#include "../include/memory.h"



int main(){


  read_parameters_section();
  read_flags_section();
  read_symmetries_section();
  read_block_states_section();
  read_local_states_section();
  read_local_states_signs_section();

  
  block_sector_no = get_sectors_no(block_state_no, block_states_array);
  local_sector_no = get_sectors_no(local_state_no, local_states_array); 


  set_sectors_representation_index(symmetry_no,
				   symmetry_types,
				   block_state_no, 
				   block_states_array, 
				   block_sector_no, 
				   block_sectors_properties_array); 



  set_original_sectors(block_state_no, 
		       block_states_array,
		       block_original_sector_array, 
		       block_original_sector_index_array);
  
  set_sectors_representation_index(symmetry_no,
				   symmetry_types,
				   local_state_no, 
				   local_states_array, 
				   local_sector_no, 
				   local_sectors_properties_array); 

  local_degeneracy = get_local_degeneracy(symmetry_no, 
					  symmetry_types, 
					  local_sector_no, 
					  local_sectors_properties_array); 


  read_block_hamiltonian_section();
  

  read_hopping_operators_section();


  if (spectral_operator_no>0)  read_spectral_operators_section();
  if(static_operator_no>0)  read_static_operators_section();
  if (local_coupling_no>0) read_local_hamiltonian_section();
  read_local_hopping_operators_section();

  read_spectral_function_section();

  create_working_folder(model,
			lambda, 
			max_state_no,
			iteration_no,
			temperature,		   
			symmetry_no,
			symmetry_types,
			coupling_no, 
			couplings,		       
			working_folder_name
			); 
  

  create_swap_folder(working_folder_name, 
		     swap_folder_name); 

  create_delta_peaks_folder(swap_folder_name, 
			    delta_peaks_folder_name); 
  create_spectral_functions_folder(swap_folder_name, 
				   spectral_functions_folder_name); 
  create_transformations_folder(swap_folder_name, 
				transformations_folder_name); 
  create_spectral_operators_folder(swap_folder_name, 
				   spectral_operators_folder_name); 
  create_static_operators_folder(swap_folder_name, 
				 static_operators_folder_name); 
  create_states_folder(swap_folder_name, 
		       states_folder_name); 
  create_sectors_folder(swap_folder_name, 
			sectors_folder_name); 
  create_mapping_folder(mapping_folder_name); 
 
  create_dos_mapping_folder(dos_mapping_folder_name);

  
  set_hopping_couplings_and_on_site_energies_array( iteration_no, 
						    hopping_operator_no,
						    hopping_couplings_array, 
						    on_site_energies_array); 

   if (hoppings_on_site_energies_flag){
     copy_hoppings_on_site_energies_files();
   }

   copy_input_file();

  block_energies_array = new double[block_state_no]; 
  transformation_sectors = new sector[block_sector_no]; 
  block_energies_index = new unsigned long [block_state_no]; 
  sorted_block_energies_array = new double [block_state_no]; 

 

  block_hamiltonian_diagonalization(block_sector_no, 
			      block_sectors_properties_array, 
			      block_hamiltonian, 
			      transformation_sectors, 
			      block_energies_array); 

  
  generate_energy_index_table(block_state_no, block_energies_array, block_energies_index); 
  sort_energies(block_state_no, block_energies_array, sorted_block_energies_array); 
  

  
  //check_orthogonality(block_sector_no, transformation_sectors); 
  
  operator_transformation(block_sector_no,
			  block_hamiltonian, 
			  transformation_sectors); 

 


  substract_ground_state_energy(block_sector_no, 
				block_sectors_properties_array, 
				block_hamiltonian, 
				block_energies_array,
				sorted_block_energies_array);
  


  for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
    
   operator_transformation(block_sector_no,
			    block_hopping_operators[hopping_index], 
			    transformation_sectors); 
  }//for hopping_index
  
  
  for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
    operator_transformation(block_sector_no,
			    block_spectral_operators[spectral_index], 
			    transformation_sectors); 
  }//for spectral_index; 
  
  
  for(int static_index=0; static_index<static_operator_no; static_index++){
    operator_transformation(block_sector_no,
			    block_static_operators[static_index], 
			    transformation_sectors); 
  }//for static_index
  
  
 

  
  int iteration_index= 0; 
  

    if(dmnrg_flag!=0){


      write_transformation_sectors_binary( m_star,
					   
					   iteration_index, 
					   
					   block_state_no,
					   block_sector_no, 
					   block_sectors_properties_array,
					   
					   transformation_sectors
					   ); 

      if ( text_swap_files_flag!=0){
	write_transformation_sectors( m_star,
				      
				      iteration_index, 
				      
				      block_state_no,
				      block_sector_no, 
				      block_sectors_properties_array,
				      
				      transformation_sectors
				      ); 
      }
    }//if dmnrg_flag
    
    


    write_states(m_star,
		 iteration_index, 
		 block_state_no,
		 block_states_array,
		 block_energies_array,
		 block_state_no, 
		 block_states_array,
		 block_original_sector_array, 
		 block_original_sector_index_array
		 );
    

    write_sectors(m_star,
		  iteration_index, 
		  block_state_no,
		  block_sector_no,
		  block_sectors_properties_array,
		  block_sector_no, 
		  block_sectors_properties_array
		  ); 



    
    if(dmnrg_flag!=0){

      for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
	
	write_spectral_operator_binary( m_star,
					iteration_index, 
					block_state_no,
					block_sector_no,
					block_spectral_operators[spectral_index]
					); 

	if ( text_swap_files_flag!=0){
	  write_spectral_operator( m_star,
				   iteration_index, 
				   block_state_no,
				   block_sector_no,
				   block_spectral_operators[spectral_index]
				   ); 
	}//if
      }//spectral_index
    }//if dmnrg_flag
      

    if(dmnrg_flag!=0){

      for(int static_index=0; static_index<static_operator_no; static_index++){
	write_static_operator_binary( m_star,
				      iteration_index, 
				      block_state_no,
				      block_sector_no,
				      block_static_operators[static_index]
				      ); 
      	if ( text_swap_files_flag!=0){
	  	write_static_operator( m_star,
				       iteration_index, 
				       block_state_no,
				       block_sector_no,
				       block_static_operators[static_index]
				       ); 
	}//if 
      }//static_index
    }//if dmnrg_flag





  for(int i=0; i<spectral_function_no; i++){

    generate_spectral_function_peaks_nrg(iteration_index,
					 
					 symmetry_no, 
					 symmetry_types,
					 
					 block_state_no,
					 block_states_array,
					 
					 block_sector_no,
					 block_sectors_properties_array, 
					 
					 block_energies_array, 
					 sorted_block_energies_array,
					 block_energies_index,
					 
					 block_spectral_operators[spectral_function_index[0][i]],
					 block_spectral_operators[spectral_function_index[1][i]]
					 ); 
    
  }//i 
    
   

    delete [] block_energies_index ; 
    block_energies_index = NULL;

    delete [] sorted_block_energies_array ; 
    sorted_block_energies_array=NULL; 

    delete [] block_energies_array; 
    block_energies_array = NULL; 

    
    for(int i =0; i<block_sector_no; i++){
      transformation_sectors[i].clear();
    }
    delete [] transformation_sectors; 
    transformation_sectors=NULL; 
        
 
  //here the self-consistency will start.


    partition_function = 0.0; 
    
    for( int iteration_index=1; iteration_index<=iteration_no; iteration_index++){
                  


      create_new_block_states_with_list( symmetry_no, 
					 symmetry_types, 
					 block_state_no, 
					 block_states_array, 
					 local_state_no,
					 local_states_array, 
					 new_block_state_no, 
					 new_block_states_array); 
           
      sort_states((unsigned long)(new_block_state_no), new_block_states_array-1); 
      
      new_block_sector_no = get_sectors_no(new_block_state_no, new_block_states_array);
      
      
      set_sectors_representation_index(symmetry_no,
				       symmetry_types,
				       new_block_state_no, 
				       new_block_states_array, 
				       
				       new_block_sector_no,
				       new_block_sectors_properties_array
				       ); 


 
      
      
      create_new_block_hamiltonian(iteration_index, 
				   
				   symmetry_no, 
				   symmetry_types, 
				   
				   new_block_sector_no, 
				   new_block_sectors_properties_array,
				   
				   new_block_hamiltonian,
			 
				   new_block_states_array,
				   block_states_array, 
				   local_states_array, 
				   
				   block_hamiltonian, 
				   
				   hopping_operator_no,			   
				   hopping_couplings_array, 
				   on_site_energies_array,

				   block_hopping_operators, 
				   local_hopping_operators, 
				   local_hopping_hamiltonians, 

				   local_sectors_properties_array); 
    
    

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      
      create_new_block_hopping_operator( iteration_index,
				   
				   symmetry_no, 
				   symmetry_types, 
				   
				   new_block_sector_no, 
				   new_block_sectors_properties_array, 
				   
				   new_block_hopping_operators[hopping_index], 
				   
				   new_block_states_array,
				   block_states_array, 
				   local_states_array, 
				   
				   block_hopping_operators[hopping_index], 
				   local_hopping_operators[hopping_index],
				  
				   local_sectors_properties_array
				   ); 
      
    }//for hopping_index

    for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
      
      create_new_block_spectral_operator(symmetry_no, 
					 symmetry_types, 
					 
					 new_block_sector_no, 
					 
					 new_block_sectors_properties_array, 
					 
					 new_block_spectral_operators[spectral_index], 
					 
					 new_block_states_array,
					 block_states_array, 
					 local_states_array, 
					 
					 block_spectral_operators[spectral_index]);


    }//spectral_index

    

  for(int static_index=0; static_index<static_operator_no; static_index++){
    create_new_block_spectral_operator(symmetry_no, 
				       symmetry_types, 
				       
				       new_block_sector_no, 
				       
				       new_block_sectors_properties_array, 				       
				       new_block_static_operators[static_index], 
				       
				       new_block_states_array,
				       block_states_array, 
				       local_states_array, 
				       
				       block_static_operators[static_index]);

  }//for static_index




    new_block_energies_array = new double[new_block_state_no]; 
    new_block_energies_index = new unsigned long [new_block_state_no]; 
    sorted_new_block_energies_array = new double [new_block_state_no]; 
    transformation_sectors = new sector[new_block_sector_no]; 


    block_hamiltonian_diagonalization(new_block_sector_no, 
				      new_block_sectors_properties_array, 
				      new_block_hamiltonian, 
				      transformation_sectors, 
				      new_block_energies_array); 

    //check_orthogonality(new_block_sector_no, transformation_sectors); 



    generate_energy_index_table(new_block_state_no, new_block_energies_array, new_block_energies_index); 
    sort_energies(new_block_state_no, new_block_energies_array, sorted_new_block_energies_array); 

    int kept_state_no=0; 

    set_states_elimination_tags(iteration_index,
				max_state_no, 
				new_block_state_no, 
				new_block_states_array,
				new_block_energies_array, 
				sorted_new_block_energies_array, 
				kept_state_no); 


    if (temperature> 0){

      partition_function += get_discarded_partition_function (iteration_index,
							      symmetry_no, 
							      symmetry_types, 
							      
							      new_block_sector_no, 
							      new_block_sectors_properties_array, 
							      
							      new_block_state_no, 
							      new_block_states_array, 
							      
							      new_block_energies_array,
							      sorted_new_block_energies_array,
							      new_block_energies_index,
							      
							      temperature);

    }//if temperature>0


  

    //now we save the transformation_sectors into files
    //next we transform the Block_Hamiltonian to the diagonal basis

    
    operator_transformation(new_block_sector_no,
			    new_block_hamiltonian, 
			    transformation_sectors); 

    
    substract_ground_state_energy( new_block_sector_no, 
				   new_block_sectors_properties_array, 
				   new_block_hamiltonian, 
				   new_block_energies_array,
				   sorted_new_block_energies_array);
    

    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      
      operator_transformation(new_block_sector_no,
			      new_block_hopping_operators[hopping_index], 
			      transformation_sectors); 
      

    }//hopping_index



    
    for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
      operator_transformation(new_block_sector_no,
			      new_block_spectral_operators[spectral_index], 
			      transformation_sectors); 
    }//spectral_index
    

    for(int static_index=0; static_index<static_operator_no; static_index++){
      operator_transformation(new_block_sector_no,
			      new_block_static_operators[static_index], 
			      transformation_sectors); 
    }


    if(dmnrg_flag!=0){
      write_transformation_sectors_binary( m_star,
					   
					   iteration_index, 
					   
					   new_block_state_no,
					   new_block_sector_no, 
					   new_block_sectors_properties_array,
					   
					   transformation_sectors
					   ); 
    
      if(text_swap_files_flag!=0){
        write_transformation_sectors( m_star,
				      
				      iteration_index, 
				      
				      new_block_state_no,
				      new_block_sector_no, 
				      new_block_sectors_properties_array,
				      
				      transformation_sectors
				      ); 
      }//if
    }//if dmnrg_flag!=0


    write_states(m_star,
		 iteration_index, 
		 new_block_state_no,
		 new_block_states_array,
		 new_block_energies_array,
		 block_state_no, 
		 block_states_array,
		 block_original_sector_array, 
		 block_original_sector_index_array
		 );
    
    write_sectors(m_star,
		  iteration_index, 
		  new_block_state_no,
		  new_block_sector_no,
		  new_block_sectors_properties_array,
		  block_sector_no, 
		  block_sectors_properties_array
		  ); 


    if(dmnrg_flag!=0){
    
      for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
      
	write_spectral_operator_binary( m_star,
					iteration_index, 
					new_block_state_no,
					new_block_sector_no,
					new_block_spectral_operators[spectral_index]
					); 
      if(text_swap_files_flag!=0){
	write_spectral_operator( m_star,
				 iteration_index, 
				 new_block_state_no,
				 new_block_sector_no,
				 new_block_spectral_operators[spectral_index]
				 ); 
	
      }//
      
      }//spectral_index
      
      for(int static_index=0; static_index<static_operator_no; static_index++){
	
	write_static_operator_binary( m_star,
				      iteration_index, 
				      new_block_state_no,
				      new_block_sector_no,
				      new_block_static_operators[static_index]
				      ); 
        if(text_swap_files_flag!=0){
	  write_static_operator( m_star,
				 iteration_index, 
				 new_block_state_no,
				 new_block_sector_no,
				 new_block_static_operators[static_index]
				 ); 	  
	}//if 
	
      }//static_index
    }//if dmnrg_flag!=0

    //if we are at the last iteration then we construct the density matrix. 

    if (iteration_index==iteration_no){

      construct_density_matrix(symmetry_no,
			       symmetry_types,
			       
			       new_block_sector_no,
			       new_block_sectors_properties_array,
			       
			       new_block_state_no,
			       new_block_states_array,
			   
			       new_block_energies_array,
			       sorted_new_block_energies_array,
			       new_block_energies_index,
			       
			       
			       new_block_hamiltonian,
			       new_block_density_matrix,

			       temperature,
			       partition_function
			       ); 

      double trace =calculate_trace(new_block_sector_no,
				    new_block_sectors_properties_array,		       
				    new_block_density_matrix); 

            
    }//if iteration_index. 

    
    
    for(int i=0; i<spectral_function_no; i++){
      generate_spectral_function_peaks_nrg(iteration_index,
					   
					   symmetry_no, 
					   symmetry_types,
					   
					   new_block_state_no,
					   new_block_states_array,
					   
					   new_block_sector_no,
					   new_block_sectors_properties_array, 
					   
					   
					   new_block_energies_array, 
					   sorted_new_block_energies_array,
					   new_block_energies_index,
					   
					   new_block_spectral_operators[spectral_function_index[0][i]],
					   new_block_spectral_operators[spectral_function_index[1][i]]
					 ); 


    }//i 
    








        
    if (iteration_index<iteration_no){  

      transfer_states_and_sectors(symmetry_no, 
				  symmetry_types,
				  
				  max_state_no, 
				  kept_state_no, 

				  new_block_state_no, 
				  new_block_states_array, 
				  new_block_sector_no,
				  new_block_sectors_properties_array, 
				  
				  block_state_no, 
				  block_states_array,
				  block_sector_no,
				  block_sectors_properties_array,
				  
				  new_block_energies_array, 
				  sorted_new_block_energies_array);


      transfer_quantum_operator( symmetry_no, 
				 symmetry_types,
				 
				 max_state_no, 
				 new_block_state_no, 
				 
				 new_block_states_array, 
				 new_block_sector_no,
				 new_block_sectors_properties_array, 
				 
				 block_state_no, 
				 block_states_array,
				 block_sector_no,
				 block_sectors_properties_array,
				 
				 block_hamiltonian,
				 new_block_hamiltonian, 
				 
				 new_block_energies_array, 
				 sorted_new_block_energies_array); 
      
      
      


      for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

	transfer_quantum_operator( symmetry_no, 
				   symmetry_types,
				   
				   max_state_no, 
				   new_block_state_no, 
				   
				   new_block_states_array, 
				   new_block_sector_no,
				   new_block_sectors_properties_array, 
				   
				   block_state_no, 
				   block_states_array,
				   block_sector_no,
				   block_sectors_properties_array,
				   
				   
				   block_hopping_operators[hopping_index],
				   new_block_hopping_operators[hopping_index], 
				   
				   new_block_energies_array, 
				   sorted_new_block_energies_array); 
	
      }//hopping_index


      for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
	
	transfer_quantum_operator( symmetry_no, 
				   symmetry_types,
				   
				   max_state_no, 
				   new_block_state_no, 
				   
				   new_block_states_array, 
				   new_block_sector_no,
				   new_block_sectors_properties_array, 
				   
				   block_state_no, 
				   block_states_array,
				   block_sector_no,
				   block_sectors_properties_array,
				   
				   
				   block_spectral_operators[spectral_index],
				   new_block_spectral_operators[spectral_index], 
				   
				   new_block_energies_array, 
				   sorted_new_block_energies_array); 
      }//spectral_index


      for(int static_index=0; static_index<static_operator_no; static_index++){
	
	transfer_quantum_operator( symmetry_no, 
				   symmetry_types,
				   
				   max_state_no, 
				   new_block_state_no, 
				   
				   new_block_states_array, 
				   new_block_sector_no,
				   new_block_sectors_properties_array, 
				   
				   block_state_no, 
				   block_states_array,
				   block_sector_no,
				   block_sectors_properties_array,
				   
				   
				   block_static_operators[static_index],
				   new_block_static_operators[static_index], 
				   
				   new_block_energies_array, 
				   sorted_new_block_energies_array); 
      }//spectral_index










    } //if iteration_index<iteration_no


    get_m_star(iteration_index,
	       max_state_no, 
	       new_block_state_no,
	       &m_star);

    
    


    clear_new_block_states_and_sectors(new_block_states_array,
				 new_block_sector_no,
				 new_block_sectors_properties_array,
				 new_block_energies_array,
				 sorted_new_block_energies_array,
				 new_block_energies_index,
				 transformation_sectors); 

    
    
  }//for iteration_index  

    
    write_m_star( m_star,
		  m_star_file_name); 

  


    /* if the dmnrg_flag is off then we stop the code here. 
       otherwise we do the backwards sweep. During this backwards sweep we read in data 
       from the files we save on the disc in course of the forward sweep. 
    */




    if(dmnrg_flag!=0){


      /* we are at the last iteration, we read in the last spectrum of the Hamiltonian
	 and construct the corresponding sectors */
      

      read_new_block_states( m_star,
			     iteration_no, 
			     
			     symmetry_no,
			     
			     new_block_state_no,
			     new_block_states_array,
			     new_block_energies_array,
			     sorted_new_block_energies_array,
			     new_block_energies_index);
      
      new_block_sector_no = get_sectors_no(new_block_state_no, new_block_states_array);
      
      
      
      set_sectors_representation_index(symmetry_no,
				       symmetry_types,
				       new_block_state_no, 
				       new_block_states_array, 
				       new_block_sector_no, 
				       new_block_sectors_properties_array); 
      
      
      /* we read in the unitary transformation matrices */
	read_transformation_sectors_binary (m_star,
					    iteration_no, 
					    new_block_state_no,
					    new_block_sector_no, 
					    new_block_sectors_properties_array,
					    transformation_sectors);
	
	
	check_orthogonality(new_block_sector_no, transformation_sectors);
	
	/* read in the matrix elements for spectral operators */  
	for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
	  
	  read_spectral_operator_binary(m_star, 
					iteration_no, 
					new_block_state_no, 
					new_block_sector_no, 
					new_block_sectors_properties_array, 
					new_block_spectral_operators[spectral_index]);
	  
	}//spectral_index; 
	
	/* read in the matrix elements for static operators */  
	for(int static_index=0; static_index<static_operator_no; static_index++){
	  
	  read_static_operator_binary(m_star, 
				      iteration_no, 
				      new_block_state_no, 
				      new_block_sector_no, 
				      new_block_sectors_properties_array, 
				      new_block_static_operators[static_index]);
	  
	}//static_index; 
	
	
	/* Now we generate delta peaks from the last iteration and dump it to a file  */ 
	for(int i=0; i<spectral_function_no; i++){
	  
	  generate_spectral_function_peaks_dmnrg(iteration_no,
						 
						 symmetry_no, 
						 symmetry_types,
						 
						 new_block_state_no,
						 new_block_states_array,
						 
						 new_block_sector_no,
						 new_block_sectors_properties_array, 
						 
						 new_block_energies_array, 
						 sorted_new_block_energies_array,
						 new_block_energies_index,
						 
						 new_block_spectral_operators[spectral_function_index[0][i]],
						 new_block_spectral_operators[spectral_function_index[1][i]],
						 new_block_density_matrix
						 );
	  
	  /* Now we generate the contribution to static averages  for pairs of
	     spectral operators from the last iteration and dump it to a file  */ 
	  generate_static_average_dmnrg(iteration_no,
					
					symmetry_no, 
					symmetry_types,
					
					new_block_state_no,
					new_block_states_array,
					
					new_block_sector_no,
					new_block_sectors_properties_array, 
					
					new_block_spectral_operators[spectral_function_index[0][i]],
					new_block_spectral_operators[spectral_function_index[1][i]],
					new_block_density_matrix
					);

	  
	  
	}//i

	/* Now we generate the contribution to static averages  for 
	       static operators from the last iteration and dump it to a file  */ 
	for(int i=0; i<static_operator_no; i++){

	    generate_static_average_dmnrg( iteration_no,
				    
					   symmetry_no, 
					   symmetry_types,
					   
					   new_block_state_no,
					   new_block_states_array,
					   
					   new_block_sector_no,
					   new_block_sectors_properties_array, 
					   
					   new_block_static_operators[i],
					   new_block_density_matrix
		);
	}//i


	/* rotate density matrix to the non-diagonal basis */

	operator_transformation(new_block_sector_no,
				new_block_density_matrix, 
				transformation_sectors); 
	
	double trace =calculate_trace(new_block_sector_no,
				      new_block_sectors_properties_array,		       
				      new_block_density_matrix); 
	
    


	/* read all kept block states of the last iteration (discarded states are not read here) */
	read_block_states(m_star,
			  iteration_no, 
			  
			  symmetry_no,

			  block_state_no, 
			  block_states_array,
			  block_original_sector_array, 
			  block_original_sector_index_array
			  );
        
    block_sector_no = get_sectors_no(block_state_no, block_states_array);
    
  
    set_sectors_representation_index(symmetry_no,
				     symmetry_types,
				     block_state_no, 
				     block_states_array, 
				     block_sector_no, 
				     block_sectors_properties_array); 
    

    construct_reduced_density_matrix(symmetry_no,
				     symmetry_types,

				     new_block_sector_no,
				     new_block_sectors_properties_array,
				     new_block_state_no,
				     new_block_states_array,
				      
				     block_sector_no,
				     block_sectors_properties_array,
				     block_state_no,
				     block_states_array,

				     new_block_density_matrix,
				     block_density_matrix,

				      temperature);
    
    new_block_density_matrix.clear();
    

     for (int iteration_index=iteration_no-1; iteration_index>=m_star; iteration_index--){
      

      //we first read the new_block_states_array from the previous iteration 

      read_new_block_states( m_star,
			     iteration_index, 
			     
			     symmetry_no,
			     
			     new_block_state_no,
			     new_block_states_array,
			     new_block_energies_array,
			     sorted_new_block_energies_array,
			     new_block_energies_index
	  );

       int free_new_block_sector_no = new_block_sector_no; 
      
       new_block_sector_no = get_sectors_no(new_block_state_no, new_block_states_array);
    

       set_sectors_representation_index(symmetry_no,
					symmetry_types,
					new_block_state_no, 
					new_block_states_array, 
					new_block_sector_no, 
					new_block_sectors_properties_array); 
      
       /* Now we transfer the kept-kept density matrix constructed in the last step to the full density matrix 
	  of this iteration by including the conttribution of the discarded-discarded channel by hand... */ 

       transfer_reduced_density_matrix(iteration_index, 
				       symmetry_no,
				       symmetry_types,
				       
				       block_sector_no,
				       block_sectors_properties_array,
				       block_state_no,
				       block_states_array,
				       
				       block_original_sector_array,
				       block_original_sector_index_array,
				      
				       new_block_sector_no,
				       new_block_sectors_properties_array,
				       new_block_state_no,
				       new_block_states_array,
				      
				       new_block_energies_array,
				       sorted_new_block_energies_array,
				       new_block_energies_index,
				     
				       block_density_matrix,				     
				       new_block_density_matrix,
				      
				       temperature,
				       partition_function); 
      
      block_density_matrix.clear();


      double trace =calculate_trace(new_block_sector_no,
				  new_block_sectors_properties_array,		       
				  new_block_density_matrix); 
      


      
      if(transformation_sectors){
	for(int i =0; i<free_new_block_sector_no; i++){
	  transformation_sectors[i].clear();
	}//for    
	delete [] transformation_sectors;
      }//if
      
      transformation_sectors = NULL;

   
      read_transformation_sectors_binary( m_star,
					  iteration_index, 
					  new_block_state_no,
					  new_block_sector_no, 
					  new_block_sectors_properties_array,
					  transformation_sectors
					  );
      


      for(int spectral_index=0; spectral_index<spectral_operator_no; spectral_index++){
	
	read_spectral_operator_binary(m_star, 
				      iteration_index, 
				      new_block_state_no, 
				      new_block_sector_no, 
				      new_block_sectors_properties_array, 
				      new_block_spectral_operators[spectral_index]);

      }//spectral_index; 



      for(int static_index=0; static_index<static_operator_no; static_index++){
	
	read_static_operator_binary(m_star, 
				    iteration_index, 
				    new_block_state_no, 
				    new_block_sector_no, 
				    new_block_sectors_properties_array, 
				    new_block_static_operators[static_index]);

      }//spectral_index; 



     
    for(int i=0; i<spectral_function_no; i++){

      generate_spectral_function_peaks_dmnrg(iteration_index,
					     
					     symmetry_no, 
					     symmetry_types,
					     
					     new_block_state_no,
					     new_block_states_array,
					    
					     new_block_sector_no,
					     new_block_sectors_properties_array, 
					     
					     new_block_energies_array, 
					     sorted_new_block_energies_array,
					     new_block_energies_index,
					    
					     new_block_spectral_operators[spectral_function_index[0][i]],
					     new_block_spectral_operators[spectral_function_index[1][i]],
					     new_block_density_matrix
					     );


      generate_static_average_dmnrg(iteration_index,
				    
				    symmetry_no, 
				    symmetry_types,
				    
				    new_block_state_no,
				    new_block_states_array,
				    
				    new_block_sector_no,
				    new_block_sectors_properties_array, 
					     
				        
				    new_block_spectral_operators[spectral_function_index[0][i]],
				    new_block_spectral_operators[spectral_function_index[1][i]],
				    new_block_density_matrix
				    );

    }//i


    for(int i=0; i<static_operator_no; i++){

      generate_static_average_dmnrg( iteration_index,
				    
				    symmetry_no, 
				    symmetry_types,
				    
				    new_block_state_no,
				    new_block_states_array,
				   
				    new_block_sector_no,
				    new_block_sectors_properties_array, 
				   				   
				     new_block_static_operators[i],
				     new_block_density_matrix
				    );
    }//i


      
    operator_transformation(new_block_sector_no,
			    new_block_density_matrix, 
			    transformation_sectors); 


      


    read_block_states(m_star,
		      iteration_index, 
		      
		      symmetry_no,
		      
		      block_state_no, 
		      block_states_array,
		      block_original_sector_array, 
		      block_original_sector_index_array
		      );
    


        
      block_sector_no = get_sectors_no(block_state_no, block_states_array);
    
  
      set_sectors_representation_index(symmetry_no,
				       symmetry_types,
				       block_state_no, 
				       block_states_array, 
				       block_sector_no, 
				       block_sectors_properties_array); 
    
      construct_reduced_density_matrix(symmetry_no,
				       symmetry_types,
				       
				       new_block_sector_no,
				       new_block_sectors_properties_array,
				       new_block_state_no,
				       new_block_states_array,
				      
				       block_sector_no,
				       block_sectors_properties_array,
				       block_state_no,
				       block_states_array,
				       
				       new_block_density_matrix,
				       block_density_matrix,
				       
				       temperature);

      new_block_density_matrix.clear();
   
      trace =calculate_trace(block_sector_no,
			     block_sectors_properties_array,		       
			     block_density_matrix); 
      
      
      //return 1;  

     }//iteration_index


      if(transformation_sectors){
	for(int i =0; i<new_block_sector_no; i++){
	  transformation_sectors[i].clear();
	}//for    
	delete [] transformation_sectors;
      }//if
      
      transformation_sectors = NULL;

     block_density_matrix.clear();

    }//if dmnrg_flag!=0        


     if (!binary_swap_files_flag){
       remove_binary_files();
     }//if 

     clear_all_memory();
         
     //here we clear all the variables that we have in the variable.h file

  return 1;
  
}


























