/****************************************************************************

Last Update : 2007/09/25 Pascu Moca 
              2008/06/26 Pascu Moca

Documentation update necessary

*****************************************************************************/

#include "../include/mapping.h"

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void read_density_of_states(){
  /*****************************************************************************
input variables:  none
 
output variables: none

return: void

What it does:
     This routine reads the density of states from the file 
     ./dos_mapping/dos.dat. This data will later be used 
     to generate the hoppings and the on-site energies.
How to call:

  ******************************************************************************/

  long line_no=0; 


  ifstream dos_file("dos_mapping//dos.dat", ios::in);

  if (!dos_file.good()){
    error_message("cannot open the dos_mapping/dos.dat file"); 
  }//if

 char line[512];
 double enrg;
 while(!(dos_file.eof())){
   dos_file.getline(buffer, buffer_size, '\n');
   word_offset =0;
   if(!get_word(buffer, word, word_offset)) break;
   line_no++; 
  }//while

  dos_energy_no = line_no; 
  dos_file.close();
  dos_file.clear();
  
  //now we allocate the necessary arrays  

  dos_energy_array = new double [dos_energy_no]; 
  if(!dos_energy_array){
    error_message("alocation failure for dos_energy_array"); 
  }//if

  for(int i=0; i<dos_energy_no; i++){
    dos_energy_array[i] = 0.0; 
  }//for i 

  dos_hopping_operator_array = new double *[hopping_operator_no]; 
  if(!dos_hopping_operator_array){
    error_message("allocation failure for dos_hopping_operator_array");
  }//if
  
  dos_hopping_operator_array[0]= new double[ dos_energy_no*hopping_operator_no];
  
  if(!dos_hopping_operator_array[0]){
    error_message("allocation failure for dos_hopping_operator_array[0]");
  }//if
  

  for(int i=1; i<hopping_operator_no; i++){
    dos_hopping_operator_array[i] = dos_hopping_operator_array[i-1]+dos_energy_no; 
  }//for i 


  for(int i=0; i<hopping_operator_no; i++){
    for(int j=0; j<dos_energy_no; j++){
      dos_hopping_operator_array[i][j] = 0.0; 
    }//j
  }//for i


  dos_file.open("dos_mapping//dos.dat", ios::in);
  
  if (!dos_file.good()){
    error_message("cannot open the dos_mapping/dos.dat file"); 
  }//if


  for(int energy_index=0; energy_index<dos_energy_no; energy_index++){
    dos_file.getline(buffer, buffer_size, '\n'); 
    word_offset = 0; 
    

    if(!get_word(buffer, word, word_offset)){
      error_message("reading dos_energy_array in routine reading_density_of_states()"); 
    }//if
    
    dos_energy_array[energy_index]=atof(word);

    for(int hopping_operator_index=0; hopping_operator_index<hopping_operator_no; hopping_operator_index++){

      if(!get_word(buffer, word, word_offset)){
	error_message("reading  dos_hopping_operator_array in routine reading_density_of_states()"); 
      }//if
    
      dos_hopping_operator_array[hopping_operator_index][energy_index]= atof(word); 

    }//for hopping_coupling_index
    
  }//energy_index


  if(fabs(dos_energy_array[0]+1.0)>allowed_error){
    error_message("the smallest energy in dos.dat has to be -1.0"); 
  }//if 

  if(fabs(dos_energy_array[dos_energy_no-1]-1.0)>allowed_error){
    error_message("the largest energy in dos.dat has to be 1.0"); 
  }//if 

  dos_file.close();
  dos_file.clear();
  
  return;
}//read_density_of_states();










void generate_initial_density_of_states(int dos_points_no_per_interval,
					int self_consistance_index, 
					int hopping_operator_no){

/*******************************************************************************
input variables:  three
     int no_of_points - is the initial number of points used for each 
                        discretized interval
     int precision    - how many digits should we keep 
     model            - string that represent the model density of states used. 

output variables: none

return: void

What it does:
     This routine generates an initial density of states for the
     model into consideration. This is next used to compute the 
     hoppings as well as the on-site energies

How to call:
*******************************************************************************/

  //first we set the precision based on the iteration_no;

  int precision = (int)(5.0*log10(powl(lambda, iteration_no))); 

  char string_buffer[10000]; 
  char dos_buffer[10000]; 

  mpf_set_default_prec(10000); 
  

  //density of states for negative energies.

  
  for (int iteration_index=0; iteration_index<=iteration_no; iteration_index++){

    density_of_states_negative_mapping_file_name(dos_mapping_folder_name,
						 iteration_index,
						 0, 
						 density_of_states_file_name
						 ); 
  

    ofstream density_of_states_file(density_of_states_file_name, ios::out); 
    
    if (!density_of_states_file.good()){
      error_message("cannot open density_of_states_file in routine generate_initial_density_of_states");
    }//if

    
    mpf_t energy_min;               mpf_init(energy_min); 
    mpf_t energy_max;               mpf_init(energy_max);
    mpf_t local_lambda;             mpf_init_set_d(local_lambda, lambda);
    mpf_t local_lambda_n;           mpf_init(local_lambda_n);                //variable used to compute lambda to some generic power n
    mpf_t inverse_local_lambda_n;   mpf_init(inverse_local_lambda_n); //variable used to keep 1/lambda^n 
    mpf_t one;                      mpf_init_set_d(one, 1.0);                          //used to keep 1 in the mpf_t format

    mpf_t energy;                   mpf_init(energy); 
    mpf_t energy_gap;               mpf_init(energy_gap);
    mpf_t delta_energy;             mpf_init(delta_energy);
    mpf_t density_of_states;        mpf_init(density_of_states); 
    mpf_t energy_gap_over_2;        mpf_init(energy_gap_over_2);
    mpf_t shifted_energy;           mpf_init(shifted_energy);

    mpf_pow_ui(local_lambda_n, local_lambda, ((unsigned long int) (iteration_index))); 
    mpf_div(inverse_local_lambda_n, one, local_lambda_n); 


    //energy_min is set to -1/lambda^n; 
    mpf_neg(energy_min, inverse_local_lambda_n); 


    mpf_pow_ui(local_lambda_n, local_lambda, ((unsigned long int) (iteration_index+1))); 
    mpf_div(inverse_local_lambda_n, one, local_lambda_n); 

    //energy_max is set to -1/lambda^(n+1); 
    mpf_neg(energy_max, inverse_local_lambda_n); 
    
    //delta energy is set to energy_max-energy_min;
    mpf_sub(delta_energy, energy_max, energy_min); 

    //energy_gap ==delta_energy/dos_points_no_per_interval; 

    mpf_div_ui(energy_gap, delta_energy, ((unsigned long int)(dos_points_no_per_interval))); 
    mpf_div_ui(energy_gap_over_2, energy_gap, ((unsigned long int)(2.0))); 

    density_of_states_file<<"dos_points_no_per_interval = "<<dos_points_no_per_interval<<endl;
    gmp_sprintf(string_buffer, "energy_gap = %.*Ff", precision, energy_gap); 
    density_of_states_file<<string_buffer<<endl;

    //now we generate dos_points_no_per_interval density points into each interval [-1/L^n, -1/L^(n+1)];

    for(int local_index=0; local_index<dos_points_no_per_interval; local_index++){

      
      mpf_mul_ui(delta_energy, energy_gap, ((unsigned long int)(local_index)));

      //this is the energy for which we compute the density of states.

      mpf_add(energy, energy_min, delta_energy);
      mpf_add(shifted_energy, energy, energy_gap_over_2); 

      gmp_sprintf(string_buffer, "%.*Ff\t", precision, shifted_energy); 


      for (int i=0; i<hopping_operator_no; i++){

	mpf_set_d(density_of_states, M_PI* density_of_states_function(hopping_operator_no,i, mpf_get_d(shifted_energy))); 
	gmp_sprintf(dos_buffer, "%.*Ff\t", precision, density_of_states); 
	strcat(string_buffer, dos_buffer); 
      }//int i 

      density_of_states_file<<string_buffer<<endl;
     
    }//local_index

    mpf_clear(energy_min); 
    mpf_clear(energy_max); 
    mpf_clear(local_lambda);  
    mpf_clear(local_lambda_n); 
    mpf_clear(inverse_local_lambda_n); 
    mpf_clear(one); 
    mpf_clear(energy); 
    mpf_clear(delta_energy); 
    mpf_clear(energy_gap); 
    mpf_clear(density_of_states); 
    mpf_clear(energy_gap_over_2); 
    mpf_clear(shifted_energy);

    density_of_states_file.close();

  }//for iteration_index





  //density of states for positive energies.

  
  for (int iteration_index=iteration_no; iteration_index>=0; iteration_index--){


    density_of_states_positive_mapping_file_name(dos_mapping_folder_name,
						 iteration_index,
						 0, 
						 density_of_states_file_name
						 ); 
  

    ofstream density_of_states_file(density_of_states_file_name, ios::out); 
    
    if (!density_of_states_file.good()){
      error_message("cannot open density_of_states_file in routine generate_initial_density_of_states");
    }//if
    
    mpf_t energy_min;               mpf_init(energy_min); 
    mpf_t energy_max;               mpf_init(energy_max);
    mpf_t local_lambda;             mpf_init_set_d(local_lambda, lambda);
    mpf_t local_lambda_n;           mpf_init(local_lambda_n);                //variable used to compute lambda to some generic power n
    mpf_t inverse_local_lambda_n;   mpf_init(inverse_local_lambda_n); //variable used to keep 1/lambda^n 
    mpf_t one;                      mpf_init_set_d(one, 1.0);                          //used to keep 1 in the mpf_t format
    mpf_t energy;                   mpf_init(energy); 
    mpf_t energy_gap;               mpf_init(energy_gap);
    mpf_t delta_energy;             mpf_init(delta_energy);
    mpf_t density_of_states;        mpf_init(density_of_states); 
    mpf_t energy_gap_over_2;        mpf_init(energy_gap_over_2);
    mpf_t shifted_energy;           mpf_init(shifted_energy);


    mpf_pow_ui(local_lambda_n, local_lambda, ((unsigned long int) (iteration_index+1))); 
    mpf_div(inverse_local_lambda_n, one, local_lambda_n); 

    //energy_min is set to 1/lambda^(n+1); 
    mpf_set(energy_min, inverse_local_lambda_n); 


    mpf_pow_ui(local_lambda_n, local_lambda, ((unsigned long int) (iteration_index))); 
    mpf_div(inverse_local_lambda_n, one, local_lambda_n); 

    //energy_max is set to 1/lambda^(n); 
    mpf_set(energy_max, inverse_local_lambda_n); 
    
    //delta energy is set to energy_max-energy_min;
    mpf_sub(delta_energy, energy_max, energy_min); 

    //energy_gap ==delta_energy/dos_points_no_per_interval; 

    mpf_div_ui(energy_gap, delta_energy, ((unsigned long int)(dos_points_no_per_interval))); 
    mpf_div_ui(energy_gap_over_2, energy_gap, ((unsigned long int)(2.0))); 

    gmp_sprintf(string_buffer, "energy_gap = %.*Ff", precision, energy_gap); 

    density_of_states_file<<"dos_points_no_per_interval = "<<dos_points_no_per_interval<<endl;
    density_of_states_file<<string_buffer<<endl;
    //now we generate dos_points_no_per_interval density points into each interval [-1/L^n, -1/L^(n+1)];


    for(int local_index=1; local_index<=dos_points_no_per_interval; local_index++){
      
      mpf_mul_ui(delta_energy, energy_gap, ((unsigned long int)(local_index)));

      //this is the energy for which we compute the density of states.
      mpf_add(energy, energy_min, delta_energy);
      mpf_sub(shifted_energy, energy, energy_gap_over_2); 
      
      
      gmp_sprintf(string_buffer, "%.*Ff\t", precision, shifted_energy); 

      for (int i=0; i<hopping_operator_no; i++){

	mpf_set_d(density_of_states, M_PI*density_of_states_function(hopping_operator_no,i, mpf_get_d(shifted_energy))); 
	gmp_sprintf(dos_buffer, "%.*Ff\t", precision, density_of_states); 
	strcat(string_buffer, dos_buffer); 
      }//int i 

      density_of_states_file<<string_buffer<<endl;
      
    }//local_index

    mpf_clear(energy_min); 
    mpf_clear(energy_max); 
    mpf_clear(local_lambda);  
    mpf_clear(local_lambda_n); 
    mpf_clear(inverse_local_lambda_n); 
    mpf_clear(one); 
    mpf_clear(energy); 
    mpf_clear(delta_energy); 
    mpf_clear(energy_gap); 
    mpf_clear(density_of_states); 
    mpf_clear(energy_gap_over_2); 
    mpf_clear(shifted_energy);

    density_of_states_file.close();

  }//for iteration_index




  return; 

}//generate_initial_density_of_states


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_gamma_and_zeta(int dos_points_no_per_interval,
			     int self_consistancy_index, 
			     int hopping_operator_no){
/*******************************************************************************
input variables:  three
     int dos_points_no_per_interval - ??? 
     int self_consistancy_index     - ??? 
     int hopping_operator_no        - ???

output variables: none

return: void

What it does:
     ????
How to call:
*******************************************************************************/  

  const int buffer_size = 10000;
  char buffer[buffer_size + 1];
  char word[buffer_size + 1];
  int word_offset;
  char string_buffer[10000];

  int precision = (int)(5.0*log10(powl(lambda, iteration_no))); 

  mpf_set_default_prec(10000); 


  //this is the negative part of the spectrum

  gamma_negative_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     gamma_file_name
				     );

  zeta_negative_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     zeta_file_name
				     );

  ofstream gamma_file_minus(gamma_file_name, ios::out); 
  ofstream zeta_file_minus(zeta_file_name, ios::out); 

  if (!gamma_file_minus.good()){
    error_message("cannot open gamma_file_minus in routine generate_gamma");
  }//if


  gamma_file_minus<<"iteration_no = "<<iteration_no<<endl;
  gamma_file_minus<<"hopping_operator_no = "<<hopping_operator_no<<endl;

  if (!zeta_file_minus.good()){
    error_message("cannot open zeta_file_minus in routine generate_gamma");
  }//if


  zeta_file_minus<<"iteration_no = "<<iteration_no<<endl;
  zeta_file_minus<<"hopping_operator_no = "<<hopping_operator_no<<endl;


  for (int iteration_index=0; iteration_index<=iteration_no; iteration_index++){


    mpf_t energy_gap; mpf_init(energy_gap);
    mpf_t energy; mpf_init(energy);     

    mpf_t *density_of_states; 
    density_of_states = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(density_of_states[hopping_index]);
    }//hopping_index; 

    mpf_t *gamma_minus; 
    gamma_minus = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(gamma_minus[hopping_index]);
    }//hopping_index; 

    mpf_t *zeta_minus; 
    zeta_minus = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(zeta_minus[hopping_index]);
    }//hopping_index; 


    mpf_t *sum_gamma; 
    sum_gamma = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(sum_gamma[hopping_index]);
    }//hopping_index; 

    mpf_t *sum_zeta; 
    sum_zeta = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(sum_zeta[hopping_index]);
    }//hopping_index; 

    mpf_t *product_zeta; 
    product_zeta = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(product_zeta[hopping_index]);
    }//hopping_index; 

   
    //set the intial values of the integrals to zero. 
    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_set_d(sum_gamma[hopping_index], 0.0);
      mpf_set_d(sum_zeta[hopping_index], 0.0); 
    }//hopping_index; 
    


    density_of_states_negative_mapping_file_name(dos_mapping_folder_name,
						 iteration_index,
						 self_consistancy_index, 
						 density_of_states_file_name
						 ); 
  

    ifstream density_of_states_file(density_of_states_file_name, ios::in); 

    if (!density_of_states_file.good()){
      error_message("cannot open density_of_states_file in routine generate_gamma");
    }//if

    
    int no_of_points = 0; 
    
    density_of_states_file.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "dos_points_no_per_interval")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("dos_points_no_per_interval in routine generate_gamma\nExiting..."); 
	}//if
	   no_of_points= atoi(word); 	      	
      }//if 
    }//if 
    assert(no_of_points == dos_points_no_per_interval); 
    
    density_of_states_file.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "energy_gap")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("energy_gap in routine generate_gamma\nExiting..."); 
	}//if
	  mpf_set_str(energy_gap,word, 10); 	      	
      }//if 
    }//if 

    //now we read in all the lines.

    for (int index=0; index<dos_points_no_per_interval; index++){

      density_of_states_file.getline(buffer, buffer_size, '\n'); 
            
      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading energy in routine generate_Gamma\nExiting..."); 
      }//if
      mpf_set_str(energy, word, 10); 

      for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading density_of_states in routine generate_Gamma\nExiting..."); 
	}//if

	//calculations for gamma
	mpf_set_str(density_of_states[hopping_index], word, 10); 	
	mpf_add(gamma_minus[hopping_index], sum_gamma[hopping_index], density_of_states[hopping_index]); 
	mpf_set(sum_gamma[hopping_index], gamma_minus[hopping_index]); 
	
	//calculations for zeta

	mpf_mul(product_zeta[hopping_index], energy, density_of_states[hopping_index]); 
	mpf_add(zeta_minus[hopping_index], sum_zeta[hopping_index], product_zeta[hopping_index]); 
	mpf_set(sum_zeta[hopping_index], zeta_minus[hopping_index]); 
      }//hopping_index
    }//index

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      
      //calculations for gamma.

      mpf_mul(gamma_minus[hopping_index], sum_gamma[hopping_index], energy_gap); 


      //calculations for zeta.

      mpf_mul(zeta_minus[hopping_index], sum_zeta[hopping_index], energy_gap); 

      int zero_check; 
      zero_check = mpf_cmp_d(gamma_minus[hopping_index], 0.0); 

      if(zero_check>0){
      mpf_div(sum_zeta[hopping_index], zeta_minus[hopping_index], gamma_minus[hopping_index]); 
      mpf_set(zeta_minus[hopping_index], sum_zeta[hopping_index]); 
      }//if

      if (zero_check<0){
	cout<<"error in routine generate_gamma_and_zeta, negative density of states ;\nExiting";
      }

      if(zero_check==0){
	mpf_set_d(zeta_minus[hopping_index],0.0); 
      }

      //calculations for gamma to extract the square root.
      mpf_sqrt(sum_gamma[hopping_index], gamma_minus[hopping_index]); 
      mpf_set(gamma_minus[hopping_index], sum_gamma[hopping_index]); 

    }//hopping_index

    gamma_file_minus<<iteration_index<<"\t"; 
    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, gamma_minus[hopping_index]);
      gamma_file_minus<<string_buffer<<"\t"; 
    }//hopping_index
    gamma_file_minus<<endl;


    zeta_file_minus<<iteration_index<<"\t"; 
    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, zeta_minus[hopping_index]);
      zeta_file_minus<<string_buffer<<"\t"; 
    }//hopping_index
    zeta_file_minus<<endl;

    
    density_of_states_file.close();
    density_of_states_file.clear();


    mpf_clear(energy_gap); 
    mpf_clear(energy); 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(density_of_states[hopping_index]);
    }//hopping_index; 

    delete [] density_of_states; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(gamma_minus[hopping_index]);
    }//hopping_index; 

    delete [] gamma_minus; 


    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(zeta_minus[hopping_index]);
    }//hopping_index; 

    delete [] zeta_minus; 

    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(sum_gamma[hopping_index]);
    }//hopping_index; 

    delete [] sum_gamma; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(sum_zeta[hopping_index]);
    }//hopping_index; 

    delete [] sum_zeta; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(product_zeta[hopping_index]);
    }//hopping_index; 

    delete [] product_zeta; 

  }//iteration_index


  gamma_file_minus.close();
  gamma_file_minus.clear();

  zeta_file_minus.close();
  zeta_file_minus.clear();






  //this is the positive part of the spectrum

  gamma_positive_mapping_file_name(dos_mapping_folder_name,
				   self_consistancy_index, 
				   gamma_file_name
				   );

  zeta_positive_mapping_file_name(dos_mapping_folder_name,
				  self_consistancy_index, 
				  zeta_file_name
				  );

  ofstream gamma_file_plus(gamma_file_name, ios::out); 
  ofstream zeta_file_plus(zeta_file_name, ios::out); 

  if (!gamma_file_plus.good()){
    error_message("cannot open gamma_file_plus in routine generate_gamma");
  }//if


  gamma_file_plus<<"iteration_no = "<<iteration_no<<endl;
  gamma_file_plus<<"hopping_operator_no = "<<hopping_operator_no<<endl;

  if (!zeta_file_plus.good()){
    error_message("cannot open zeta_file_plus in routine generate_gamma");
  }//if


  zeta_file_plus<<"iteration_no = "<<iteration_no<<endl;
  zeta_file_plus<<"hopping_operator_no = "<<hopping_operator_no<<endl;


  for (int iteration_index=0; iteration_index<=iteration_no; iteration_index++){


    mpf_t energy_gap; mpf_init(energy_gap);
    mpf_t energy; mpf_init(energy);     

    mpf_t *density_of_states; 
    density_of_states = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(density_of_states[hopping_index]);
    }//hopping_index; 

    mpf_t *gamma_plus; 
    gamma_plus = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(gamma_plus[hopping_index]);
    }//hopping_index; 

    mpf_t *zeta_plus; 
    zeta_plus = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(zeta_plus[hopping_index]);
    }//hopping_index; 


    mpf_t *sum_gamma; 
    sum_gamma = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(sum_gamma[hopping_index]);
    }//hopping_index; 

    mpf_t *sum_zeta; 
    sum_zeta = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(sum_zeta[hopping_index]);
    }//hopping_index; 

    mpf_t *product_zeta; 
    product_zeta = new mpf_t [hopping_operator_no]; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_init(product_zeta[hopping_index]);
    }//hopping_index; 

   
    //set the intial values of the integrals to zero. 
    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_set_d(sum_gamma[hopping_index], 0.0);
      mpf_set_d(sum_zeta[hopping_index], 0.0); 
    }//hopping_index; 
    


    density_of_states_positive_mapping_file_name(dos_mapping_folder_name,
						 iteration_index,
						 self_consistancy_index, 
						 density_of_states_file_name
						 ); 
  

    ifstream density_of_states_file(density_of_states_file_name, ios::in); 

    if (!density_of_states_file.good()){
      error_message("cannot open density_of_states_file in routine generate_gamma");
    }//if

    
    int no_of_points = 0; 
    
    density_of_states_file.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "dos_points_no_per_interval")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("dos_points_no_per_interval in routine generate_gamma\nExiting..."); 
	}//if
	   no_of_points= atoi(word); 	      	
      }//if 
    }//if 
    assert(no_of_points == dos_points_no_per_interval); 
    
    density_of_states_file.getline(buffer, buffer_size, '\n');
    word_offset = 0; 

    if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "energy_gap")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("energy_gap in routine generate_gamma\nExiting..."); 
	}//if
	  mpf_set_str(energy_gap,word, 10); 	      	
      }//if 
    }//if 

    //now we read in all the lines.

    for (int index=0; index<dos_points_no_per_interval; index++){

      density_of_states_file.getline(buffer, buffer_size, '\n'); 
            
      word_offset = 0; 
      if(!get_word(buffer, word, word_offset)){
	error_message("reading energy in routine generate_Gamma\nExiting..."); 
      }//if
      mpf_set_str(energy, word, 10); 

      for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
	if(!get_word(buffer, word, word_offset)){
	  error_message("reading density_of_states in routine generate_Gamma\nExiting..."); 
	}//if

	//calculations for gamma
	mpf_set_str(density_of_states[hopping_index], word, 10); 	
	mpf_add(gamma_plus[hopping_index], sum_gamma[hopping_index], density_of_states[hopping_index]); 
	mpf_set(sum_gamma[hopping_index], gamma_plus[hopping_index]); 

	//calculations for zeta

	mpf_mul(product_zeta[hopping_index], energy, density_of_states[hopping_index]); 
	mpf_add(zeta_plus[hopping_index], sum_zeta[hopping_index], product_zeta[hopping_index]); 
	mpf_set(sum_zeta[hopping_index], zeta_plus[hopping_index]); 
      }//hopping_index
    }//i

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      
      //calculations for gamma.

      mpf_mul(gamma_plus[hopping_index], sum_gamma[hopping_index], energy_gap); 


      //calculations for zeta.

      mpf_mul(zeta_plus[hopping_index], sum_zeta[hopping_index], energy_gap); 

      int zero_check; 
      zero_check = mpf_cmp_d(gamma_plus[hopping_index], 0.0); 

      if(zero_check>0){
      mpf_div(sum_zeta[hopping_index], zeta_plus[hopping_index], gamma_plus[hopping_index]); 
      mpf_set(zeta_plus[hopping_index], sum_zeta[hopping_index]); 
      }//if

      if (zero_check<0){
	cout<<"error in routine generate_gamma_and_zeta, negative density of states ;\nExiting";
      }

      if(zero_check==0){
	mpf_set_d(zeta_plus[hopping_index],0.0); 
      }

      //calculations for gamma to extract the square root.
      mpf_sqrt(sum_gamma[hopping_index], gamma_plus[hopping_index]); 
      mpf_set(gamma_plus[hopping_index], sum_gamma[hopping_index]); 

    }//hopping_index

    gamma_file_plus<<iteration_index<<"\t"; 
    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, gamma_plus[hopping_index]);
      gamma_file_plus<<string_buffer<<"\t"; 
    }//hopping_index
    gamma_file_plus<<endl;


    zeta_file_plus<<iteration_index<<"\t"; 
    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, zeta_plus[hopping_index]);
      zeta_file_plus<<string_buffer<<"\t"; 
    }//hopping_index
    zeta_file_plus<<endl;

    
    density_of_states_file.close();
    density_of_states_file.clear();


    mpf_clear(energy_gap); 
    mpf_clear(energy); 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(density_of_states[hopping_index]);
    }//hopping_index; 

    delete [] density_of_states; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(gamma_plus[hopping_index]);
    }//hopping_index; 

    delete [] gamma_plus; 


    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(zeta_plus[hopping_index]);
    }//hopping_index; 

    delete [] zeta_plus; 

    
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(sum_gamma[hopping_index]);
    }//hopping_index; 

    delete [] sum_gamma; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(sum_zeta[hopping_index]);
    }//hopping_index; 

    delete [] sum_zeta; 

    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      mpf_clear(product_zeta[hopping_index]);
    }//hopping_index; 

    delete [] product_zeta; 

  }//iteration_index


  gamma_file_plus.close();
  gamma_file_plus.clear();

  zeta_file_plus.close();
  zeta_file_plus.clear();

  return; 
}//generate_gamma


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_hoppings_and_on_site_energies(int self_consistancy_index,
					    int hopping_operator_no){

/*******************************************************************************
input variables:  two
     int self_consistancy_index - ???
     int hopping_operator_no    - ???

output variables: none

return: void

What it does:
     ???
How to call:
*******************************************************************************/

  const int buffer_size = 1000;
  char buffer[buffer_size + 1];
  char word[buffer_size + 1];
  int word_offset;
  char string_buffer[1000];

  int precision = (int)(5.0*log10(powl(lambda, iteration_no))); 

  mpf_set_default_prec(10000); 


  //generate arrays for gammma's zeta's etc. 



  mpf_t **gamma_plus; 
  
  gamma_plus =new mpf_t *[hopping_operator_no]; 
  
  if (!gamma_plus){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  gamma_plus[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    gamma_plus[i] = gamma_plus[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(gamma_plus[i][j]);
    }//j
  }//i



  mpf_t **gamma_minus; 
  
  gamma_minus =new mpf_t *[hopping_operator_no]; 

  if (!gamma_minus){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  gamma_minus[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    gamma_minus[i] = gamma_minus[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(gamma_minus[i][j]);
    }//j
  }//i



  mpf_t **zeta_plus; 
  
  zeta_plus =new mpf_t *[hopping_operator_no]; 
  
  if (!zeta_plus){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  zeta_plus[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    zeta_plus[i] = zeta_plus[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(zeta_plus[i][j]);
    }//j
  }//i


  mpf_t **zeta_minus; 
  
  zeta_minus =new mpf_t *[hopping_operator_no]; 
  
  if (!zeta_minus){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  zeta_minus[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    zeta_minus[i] = zeta_minus[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(zeta_minus[i][j]);
    }//j
  }//i



  mpf_t **on_site_energy; 
  
  on_site_energy =new mpf_t *[hopping_operator_no]; 
  
  if (!on_site_energy){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  on_site_energy[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    on_site_energy[i] = on_site_energy[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(on_site_energy[i][j]);
    }//j
  }//i


  mpf_t **hopping_coupling; 
  
  hopping_coupling =new mpf_t *[hopping_operator_no]; 
  
  if (!hopping_coupling){
    error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
  }//if
  hopping_coupling[0] = new mpf_t [hopping_operator_no*(iteration_no+1)]; 

  for (int i=1; i< hopping_operator_no; i++){
    hopping_coupling[i] = hopping_coupling[i-1]+(iteration_no+1);
  }
  
  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_init(hopping_coupling[i][j]);
    }//j
  }//i


  mpf_t * hopping_coupling_minus_1;
  
  hopping_coupling_minus_1 = new mpf_t [hopping_operator_no]; 
  
  for(int i=0; i<hopping_operator_no; i++){
    mpf_init(hopping_coupling_minus_1[i]);
  }//i

  

  //this is the negative part of the spectrum

  char gamma_file_name_negative[400]; 
  char gamma_file_name_positive[400]; 
  char zeta_file_name_negative[400]; 
  char zeta_file_name_positive[400]; 

  gamma_negative_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     gamma_file_name_negative
				     );

  zeta_negative_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     zeta_file_name_negative
				     );

  gamma_positive_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     gamma_file_name_positive
				     );

  zeta_positive_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     zeta_file_name_positive
				     );

  ifstream gamma_file_minus(gamma_file_name_negative, ios::in); 
  ifstream zeta_file_minus(zeta_file_name_negative, ios::in);
  ifstream gamma_file_plus(gamma_file_name_positive, ios::in); 
  ifstream zeta_file_plus(zeta_file_name_positive, ios::in);



  if (!gamma_file_minus.good()){
    error_message("cannot open gamma_file_minus in routine generate_gamma");
  }//if

  if (!zeta_file_minus.good()){
    error_message("cannot open zeta_file_minus in routine generate_gamma");
  }//if
  if (!gamma_file_plus.good()){
    error_message("cannot open gamma_file_minus in routine generate_gamma");
  }//if

  if (!zeta_file_plus.good()){
    error_message("cannot open zeta_file_minus in routine generate_gamma");
  }//if

  int local_iteration_no = 0;   
  int local_hopping_operator_no = 0; 


  //now we read in the gamma's and zeta's used to compute the hoppings and the on-site energies.

  //this is the gamma_minus data.

  
  gamma_file_minus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  
  //read the iteration_no line first

  if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "iteration_no")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("iteration_no in routine generate_hoppings_and_on_site_energies"); 
	}//if
	local_iteration_no= atoi(word); 	      	
      }//if 
  }//if 

  assert(local_iteration_no == iteration_no); 

  
  gamma_file_minus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  
  
  if(get_word(buffer, word, word_offset)){
    if(strcmp(word, "hopping_operator_no")==0){
      if(!get_word(buffer, word, word_offset)){
	error_message("hopping_operator_no in routine generate_hoppings_and_on_site_energies"); 
      }//if
      local_hopping_operator_no= atoi(word); 	      	
    }//if 
  }//if 

  assert(local_hopping_operator_no == hopping_operator_no); 


  for (int index=0; index<=iteration_no; index++){
    gamma_file_minus.getline(buffer, buffer_size, '\n');
    
    int local_index=0; 

    word_offset = 0; 
    if(!get_word(buffer, word, word_offset)){
      error_message("reading index in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    local_index = atoi(word); 
    assert(local_index==index); 

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

      if(!get_word(buffer, word, word_offset)){
	error_message("reading gamma_minus in routine generate_hoppings_and_on_site_energies\nExiting..."); 
      }//if

      //calculations for gamma
      mpf_set_str(gamma_minus[hopping_index][index], word, 10); 	      
    }//hopping_index
  }//index


  //this is the gamma_plus data.

  
  gamma_file_plus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  

  //read the iteration_no line first

  if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "iteration_no")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("iteration_no in routine generate_hoppings_and_on_site_energies"); 
	}//if
	local_iteration_no= atoi(word); 	      	
      }//if 
  }//if 

  assert(local_iteration_no == iteration_no); 

  
  gamma_file_plus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  
  
  if(get_word(buffer, word, word_offset)){
    if(strcmp(word, "hopping_operator_no")==0){
      if(!get_word(buffer, word, word_offset)){
	error_message("hopping_operator_no in routine generate_hoppings_and_on_site_energies"); 
      }//if
      local_hopping_operator_no= atoi(word); 	      	
    }//if 
  }//if 

  assert(local_hopping_operator_no == hopping_operator_no); 


  for (int index=0; index<=iteration_no; index++){
    gamma_file_plus.getline(buffer, buffer_size, '\n');
    
    int local_index=0; 

    word_offset = 0; 
    if(!get_word(buffer, word, word_offset)){
      error_message("reading index in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    local_index = atoi(word); 
    assert(local_index==index); 

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

      if(!get_word(buffer, word, word_offset)){
	error_message("reading gamma_plus in routine generate_hoppings_and_on_site_energies\nExiting..."); 
      }//if

      //calculations for gamma
      mpf_set_str(gamma_plus[hopping_index][index], word, 10); 	      
    }//hopping_index
  }//index









  //this is the zeta_minus part




  zeta_file_minus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  

  //read the iteration_no line first

  if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "iteration_no")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("iteration_no in routine generate_hoppings_and_on_site_energies"); 
	}//if
	local_iteration_no= atoi(word); 	      	
      }//if 
  }//if 

  assert(local_iteration_no == iteration_no); 

  
  zeta_file_minus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  
  
  if(get_word(buffer, word, word_offset)){
    if(strcmp(word, "hopping_operator_no")==0){
      if(!get_word(buffer, word, word_offset)){
	error_message("hopping_operator_no in routine generate_hoppings_and_on_site_energies"); 
      }//if
      local_hopping_operator_no= atoi(word); 	      	
    }//if 
  }//if 

  assert(local_hopping_operator_no == hopping_operator_no); 


  for (int index=0; index<=iteration_no; index++){
    zeta_file_minus.getline(buffer, buffer_size, '\n');
    
    int local_index=0; 

    word_offset = 0; 
    if(!get_word(buffer, word, word_offset)){
      error_message("reading index in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    local_index = atoi(word); 
    assert(local_index==index); 

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

      if(!get_word(buffer, word, word_offset)){
	error_message("reading zeta_minus in routine generate_hoppings_and_on_site_energies\nExiting..."); 
      }//if

      //calculations for zeta
      mpf_set_str(zeta_minus[hopping_index][index], word, 10); 	      
    }//hopping_index
  }//index











  //this is the zeta_plus part.




  zeta_file_plus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  

  //read the iteration_no line first

  if(get_word(buffer, word, word_offset)){
      if(strcmp(word, "iteration_no")==0){
	if(!get_word(buffer, word, word_offset)){
	  error_message("iteration_no in routine generate_hoppings_and_on_site_energies"); 
	}//if
	local_iteration_no= atoi(word); 	      	
      }//if 
  }//if 

  assert(local_iteration_no == iteration_no); 

  
  zeta_file_plus.getline(buffer, buffer_size, '\n');
  word_offset = 0; 
  
  
  if(get_word(buffer, word, word_offset)){
    if(strcmp(word, "hopping_operator_no")==0){
      if(!get_word(buffer, word, word_offset)){
	error_message("hopping_operator_no in routine generate_hoppings_and_on_site_energies"); 
      }//if
      local_hopping_operator_no= atoi(word); 	      	
    }//if 
  }//if 

  assert(local_hopping_operator_no == hopping_operator_no); 


  for (int index=0; index<=iteration_no; index++){
    zeta_file_plus.getline(buffer, buffer_size, '\n');
    
    int local_index=0; 

    word_offset = 0; 
    if(!get_word(buffer, word, word_offset)){
      error_message("reading index in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    local_index = atoi(word); 
    assert(local_index==index); 

    for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){

      if(!get_word(buffer, word, word_offset)){
	error_message("reading zeta_plus in routine generate_hoppings_and_on_site_energies\nExiting..."); 
      }//if

      //calculations for zeta
      mpf_set_str(zeta_plus[hopping_index][index], word, 10); 	      
    }//hopping_index
  }//index

  gamma_file_minus.close();
  gamma_file_minus.clear();
  gamma_file_plus.close();
  gamma_file_plus.clear();

  zeta_file_minus.close();
  zeta_file_minus.clear();
  zeta_file_plus.close();
  zeta_file_plus.clear();




  //here starts the tridiagonalization procedure.


  for (int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
    

    //first we allocate arrays for u's and v's as well as for theta_0; 

    
    mpf_t **u; 
  
    u =new mpf_t *[iteration_no+1]; 
    
    if (!u){
      error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    u[0] = new mpf_t [(iteration_no+1)*(iteration_no+1)]; 
    
    for (int i=1; i<=iteration_no; i++){
      u[i] = u[i-1]+(iteration_no+1);
    }
    
    for (int i=0;i<=iteration_no; i++){
      for(int j=0; j<=iteration_no; j++){
	mpf_init(u[i][j]);
      }//j
    }//i
    
    mpf_t **v; 
    
    v =new mpf_t *[iteration_no+1]; 
    
    if (!v){
      error_message("allocation failure in routine generate_hoppings_and_on_site_energies\nExiting..."); 
    }//if
    v[0] = new mpf_t [(iteration_no+1)*(iteration_no+1)]; 
    
    for (int i=1; i<=iteration_no; i++){
      v[i] = v[i-1]+(iteration_no+1);
    }
    
    for (int i=0;i<=iteration_no; i++){
      for(int j=0; j<=iteration_no; j++){
	mpf_init(v[i][j]);
      }//j
    }//i

    
    mpf_t pi; mpf_init_set_d(pi, 3.1415926535897932384626433832795028841971693993751); 
    mpf_t theta_0; mpf_init(theta_0);   
    mpf_t sum; mpf_init(sum); 
    mpf_t sum_copy; mpf_init(sum_copy); 
    mpf_t square; mpf_init(square); 
    mpf_t sqrt_theta_0; mpf_init(sqrt_theta_0); 
    mpf_t product; mpf_init(product); 
    mpf_t square_gamma; mpf_init(square_gamma); 
    mpf_t square_zeta; mpf_init(square_zeta); 
    mpf_t sqrt_pi; mpf_init(sqrt_pi); 

    //first we compute theta_0
    
    for (int i=0; i<=iteration_no; i++){

      mpf_mul(square_gamma, gamma_plus[hopping_index][i], gamma_plus[hopping_index][i]); 
      mpf_add(sum,sum_copy,square_gamma);
      mpf_set(sum_copy, sum);

    }//i

    

    for (int i=0; i<=iteration_no; i++){

      mpf_mul(square_gamma, gamma_minus[hopping_index][i], gamma_minus[hopping_index][i]); 
      mpf_add(sum,sum_copy,square_gamma);
      mpf_set(sum_copy, sum);
    }//i

    mpf_set(theta_0, sum); 
    

    mpf_sqrt(sqrt_theta_0, theta_0); 

    //now we compute hopping_coupling_minus_1; the first coupling in the chain

    mpf_sqrt(sqrt_pi, pi); 

    mpf_div(hopping_coupling_minus_1[hopping_index], sqrt_theta_0, sqrt_pi); 
    

    //now we compute u[0] and v[0]
    
    
    for (int m=0; m<=iteration_no; m++){
      mpf_div(u[0][m], gamma_plus[hopping_index][m], sqrt_theta_0);
      mpf_div(v[0][m], gamma_minus[hopping_index][m], sqrt_theta_0); 
    }//j
    

    //now we compute on_site_energy[0] 
    
    mpf_set_d(sum, 0.0); 
    mpf_set(sum_copy, sum); 

    for (int m=0; m<=iteration_no; m++){
      mpf_mul(square_gamma, gamma_plus[hopping_index][m], gamma_plus[hopping_index][m]); 
      mpf_mul(product, square_gamma, zeta_plus[hopping_index][m]); 
      mpf_add(sum, sum_copy, product); 
      mpf_set(sum_copy, sum); 
    }//m

    for (int m=0; m<=iteration_no; m++){
      mpf_mul(square_gamma, gamma_minus[hopping_index][m], gamma_minus[hopping_index][m]); 
      mpf_mul(product, square_gamma, zeta_minus[hopping_index][m]); 
      mpf_add(sum, sum_copy, product); 
      mpf_set(sum_copy, sum); 
    }//m

    mpf_div(on_site_energy[hopping_index][0], sum, theta_0); 
    

    
    //now we compute hopping_coupling[0]

    
    mpf_set_d(sum, 0); 
    mpf_set(sum_copy, sum); 

    for(int m=0; m<=iteration_no; m++){
      mpf_mul(square_gamma, gamma_plus[hopping_index][m], gamma_plus[hopping_index][m]);
      mpf_mul(square_zeta, zeta_plus[hopping_index][m], zeta_plus[hopping_index][m]);
      mpf_mul(product, square_gamma, square_zeta); 
      mpf_add(sum, sum_copy, product); 
      mpf_set(sum_copy, sum); 
    }//m

    for(int m=0; m<=iteration_no; m++){
      mpf_mul(square_gamma, gamma_minus[hopping_index][m], gamma_minus[hopping_index][m]);
      mpf_mul(square_zeta, zeta_minus[hopping_index][m], zeta_minus[hopping_index][m]);
      mpf_mul(product, square_gamma, square_zeta); 
      mpf_add(sum, sum_copy, product); 
      mpf_set(sum_copy, sum); 
    }//m


    mpf_mul(product,on_site_energy[hopping_index][0], theta_0); 

    mpf_sub(sum, sum_copy, product); 
    mpf_set(sum_copy, sum); 

    mpf_div(sum, sum_copy, theta_0); 
    mpf_set(sum_copy, sum); 
    mpf_sqrt(hopping_coupling[hopping_index][0], sum); 



    //now we compute u[1][m] and v[1][m]
    

    for(int m=0; m<=iteration_no; m++){
      mpf_set_d(sum, 0.0); 
      mpf_set(sum_copy, sum); 

      mpf_sub(sum, zeta_plus[hopping_index][m], on_site_energy[hopping_index][0]); 
      mpf_set(sum_copy, sum); 
      mpf_mul(sum, sum_copy, gamma_plus[hopping_index][m]);
      mpf_set(sum_copy, sum); 
      mpf_div(sum, sum_copy, sqrt_theta_0); 
      mpf_set(sum_copy, sum); 
      mpf_div(sum, sum_copy, hopping_coupling[hopping_index][0]);
      mpf_set(sum_copy, sum); 
      
      mpf_set(u[1][m], sum); 
    }//m


    for(int m=0; m<=iteration_no; m++){
      mpf_set_d(sum, 0.0); 
      mpf_set(sum_copy, sum); 

      mpf_sub(sum, zeta_minus[hopping_index][m], on_site_energy[hopping_index][0]); 
      mpf_set(sum_copy, sum); 
      mpf_mul(sum, sum_copy, gamma_minus[hopping_index][m]);
      mpf_set(sum_copy, sum); 
      mpf_div(sum, sum_copy, sqrt_theta_0); 
      mpf_set(sum_copy, sum); 
      mpf_div(sum, sum_copy, hopping_coupling[hopping_index][0]);
      mpf_set(sum_copy, sum); 
      mpf_set(v[1][m], sum); 
    }//m



    //now we start the recursivity loop for computing all the on-site energies and hopping_couplings.


    for(int n=1; n<=iteration_no; n++){

      //first we compute the on-site energies
      
      mpf_set_d(sum, 0.0); 
      mpf_set(sum_copy, sum); 

      for(int m=0; m<=iteration_no; m++){
       	mpf_mul(square, u[n][m], u[n][m]); 
	mpf_mul(product, square, zeta_plus[hopping_index][m]);
	mpf_add(sum, sum_copy, product); 
	mpf_set(sum_copy, sum); 
      }//m

      for(int m=0; m<=iteration_no; m++){
       	mpf_mul(square, v[n][m], v[n][m]); 
	mpf_mul(product, square, zeta_minus[hopping_index][m]);
	mpf_add(sum, sum_copy, product); 
	mpf_set(sum_copy, sum); 
      }//m
      
      mpf_set(on_site_energy[hopping_index][n], sum); 



      //now we compute the hopping couplings

      mpf_set_d(sum, 0.0); 
      mpf_set(sum_copy, sum); 

      
      for(int m=0; m<=iteration_no; m++){
	mpf_mul(square, u[n][m], u[n][m]);
	mpf_mul(square_zeta, zeta_plus[hopping_index][m], zeta_plus[hopping_index][m]); 
	mpf_mul(product, square, square_zeta); 
	mpf_add(sum, sum_copy, product); 
	mpf_set(sum_copy, sum); 
      }//m

      for(int m=0; m<=iteration_no; m++){
	mpf_mul(square, v[n][m], v[n][m]);
	mpf_mul(square_zeta, zeta_minus[hopping_index][m], zeta_minus[hopping_index][m]); 
	mpf_mul(product, square, square_zeta); 
	mpf_add(sum, sum_copy, product); 
	mpf_set(sum_copy, sum); 
      }//m

      mpf_mul(square, hopping_coupling[hopping_index][n-1], hopping_coupling[hopping_index][n-1]); 
      mpf_sub(sum, sum_copy, square); 
      mpf_set(sum_copy, sum); 

      mpf_mul(square, on_site_energy[hopping_index][n], on_site_energy[hopping_index][n]);
      
      mpf_sub(sum, sum_copy, square); 
      mpf_set(sum_copy, sum); 

      mpf_sqrt(hopping_coupling[hopping_index][n], sum); 

      //now we recalculate the u's and v's

      
      if (n<iteration_no){

	mpf_set_d(sum, 0.0); 
	mpf_set(sum_copy, sum); 


	for(int m=0; m<=iteration_no; m++){
	  mpf_sub(sum, zeta_plus[hopping_index][m], on_site_energy[hopping_index][n]);
	  mpf_set(sum_copy, sum); 
	  mpf_mul(sum, sum_copy, u[n][m]);
	  mpf_set(sum_copy, sum); 
	  
	  mpf_mul(product, hopping_coupling[hopping_index][n-1], u[n-1][m]);
	  mpf_sub(sum, sum_copy, product); 
	  mpf_set(sum_copy, sum); 
	  
	  mpf_div(u[n+1][m], sum, hopping_coupling[hopping_index][n]); 
	}//m

	mpf_set_d(sum, 0.0); 
	mpf_set(sum_copy, sum); 


	for(int m=0; m<=iteration_no; m++){
	  mpf_sub(sum, zeta_minus[hopping_index][m], on_site_energy[hopping_index][n]);
	  mpf_set(sum_copy, sum); 
	  mpf_mul(sum, sum_copy, v[n][m]);
	  mpf_set(sum_copy, sum); 
	  
	  mpf_mul(product, hopping_coupling[hopping_index][n-1], v[n-1][m]);
	  mpf_sub(sum, sum_copy, product); 
	  mpf_set(sum_copy, sum); 
	  
	  mpf_div(v[n+1][m], sum, hopping_coupling[hopping_index][n]); 
	}//m


      }//if n 
    }//n







               
    for (int i=0;i<=iteration_no; i++){
      for(int j=0; j<=iteration_no; j++){
	mpf_clear(u[i][j]);
      }//j
    }//i
    delete [] u[0];
    u[0] = NULL;
    
    delete [] u; 
    u= NULL; 
    
    
    for (int i=0;i<=iteration_no; i++){
      for(int j=0; j<=iteration_no; j++){
	mpf_clear(v[i][j]);
      }//j
    }//i
    delete [] v[0];
    v[0] = NULL;
    
    delete [] v; 
    v= NULL; 
    
    

    mpf_clear(pi); 
    mpf_clear(sqrt_pi);
    mpf_clear(theta_0); 
    mpf_clear(sum); 
    mpf_clear(sum_copy);
    mpf_clear(square); 
    mpf_clear(sqrt_theta_0); 
    mpf_clear(product); 
    mpf_clear(square_gamma); 
    mpf_clear(square_zeta); 
  }//hopping_index
  

  //now we can save the on_site energies and the hoppings into a file.


  on_site_energies_mapping_file_name(dos_mapping_folder_name,
				     self_consistancy_index, 
				     on_site_energies_file_name
				     );
  hopping_couplings_mapping_file_name(dos_mapping_folder_name,
				      self_consistancy_index, 
				      hopping_couplings_file_name
				      );



  ofstream on_site_energies_file(on_site_energies_file_name, ios::out); 
  ofstream hopping_couplings_file(hopping_couplings_file_name, ios::out); 

  if(!on_site_energies_file.good()){
    error_message("cannot open on_site_energies_file in routine compute_hoppings_and_on_site_energies\nExiting..."); 
  }//if

  if(!hopping_couplings_file.good()){
    error_message("cannot open hopping_couplings_file in routine compute_hoppings_and_on_site_energies\nExiting..."); 
  }//if

  on_site_energies_file<<"# lambda = "<<lambda<<endl;
  on_site_energies_file<<"# iteration_no = "<<iteration_no<<endl;
  on_site_energies_file<<"# hopping_operator_no = "<<hopping_operator_no<<endl;

  
  for(int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
    on_site_energies_file<<iteration_index<<"\t"; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, on_site_energy[hopping_index][iteration_index]);
      on_site_energies_file<<string_buffer<<"\t"; 
    }//hopping_index
    on_site_energies_file<<endl; 
  }//iteration_index
  
  hopping_couplings_file<<"# lambda = "<<lambda<<endl;
  hopping_couplings_file<<"# iteration_no = "<<iteration_no<<endl; 
  hopping_couplings_file<<"# hopping_operator_no = "<<hopping_operator_no<<endl;

  /*
  hopping_couplings_file<<-1<<"\t"; 
  for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, hopping_coupling_minus_1[hopping_index]);
      hopping_couplings_file<<string_buffer<<"\t"; 
  }//hopping_index
  hopping_couplings_file<<endl;
  */

  for(int iteration_index=0; iteration_index<=iteration_no; iteration_index++){
    hopping_couplings_file<<iteration_index<<"\t"; 
    for(int hopping_index=0; hopping_index<hopping_operator_no; hopping_index++){
      gmp_sprintf(string_buffer, "%.*Ff", precision, hopping_coupling[hopping_index][iteration_index]);
      hopping_couplings_file<<string_buffer<<"\t"; 
    }//hopping_index
    hopping_couplings_file<<endl; 
  }//iteration_index






  //now we delete all the large numbers used in this routine.

  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_clear(zeta_plus[i][j]);
    }//j
  }//i

  delete [] zeta_plus[0]; 
  zeta_plus[0] = NULL;

  delete[] zeta_plus;
  zeta_plus =NULL;


  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<iteration_no; j++){
      mpf_clear(zeta_minus[i][j]);
    }//j
  }//i

  delete [] zeta_minus[0]; 
  zeta_minus[0] = NULL;

  delete[] zeta_minus;
  zeta_minus =NULL;



  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_clear(gamma_plus[i][j]);
    }//j
  }//i

  delete [] gamma_plus[0]; 
  gamma_plus[0] = NULL;

  delete[] gamma_plus;
  gamma_plus =NULL;


  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_clear(gamma_minus[i][j]);
    }//j
  }//i

  delete [] gamma_minus[0]; 
  gamma_minus[0] = NULL;

  delete[] gamma_minus;
  gamma_minus =NULL;



  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_clear(on_site_energy[i][j]);
    }//j
  }//i

  delete [] on_site_energy[0]; 
  on_site_energy[0] = NULL;

  delete[] on_site_energy;
  on_site_energy =NULL;


  for (int i=0;i<hopping_operator_no; i++){
    for(int j=0; j<=iteration_no; j++){
      mpf_clear(hopping_coupling[i][j]);
    }//j
  }//i

  delete [] hopping_coupling[0]; 
  hopping_coupling[0] = NULL;

  delete[] hopping_coupling;
  hopping_coupling =NULL;

  
  for(int i=0; i<hopping_operator_no; i++){
    mpf_clear(hopping_coupling_minus_1[i]);
  }//i

  delete [] hopping_coupling_minus_1 ; 
  hopping_coupling_minus_1 =NULL; 
  return; 
}//generate_hoppings_and_on_site_energies


