/****************************************************************************

Modified 

2007/11/21 by Pascu Moca 
2008/02/07 by Pascu Moca
2008/04/02 by Pascu Moca
2008/04/03 by Moca Pascu , new procedure added. 
Documentation update necessary


*****************************************************************************/

#include "../include/spectral_function.h"


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


double clebsch_gordan_coefficients_sum(int symmetry_no, 
				       symmetry * symmetry_types, 
				       sector_properties * new_block_sectors_properties_array, 
				       int sector_row, 
				       quantum_operator  & one_spectral_operator,
				       int sector_column){

/*******************************************************************************
input variables:  two
     ???
output variables: one

return: void

What it does:
    This routine computes the log gauss distribution

How to call:
*******************************************************************************/


    int S_new[symmetry_no], S_added[symmetry_no], S_old[symmetry_no];
    int S_z_new[symmetry_no], S_z_added[symmetry_no], S_z_old[symmetry_no];
    int max_S_z_new[symmetry_no], max_S_z_old[symmetry_no];
    int min_S_z_new[symmetry_no], min_S_z_old[symmetry_no];
    int new_block_no = 1;
    int old_no = 1;

    double CG_coefficient_square = 1;
    double CG_coefficient_square_sum = 0;

    for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){

      if((!strcmp(symmetry_types[symmetry_index].get_name(),"SU(2)")) || (!strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)"))){
	S_new[symmetry_index] = new_block_sectors_properties_array[sector_row].representation_index[symmetry_index];
	S_added[symmetry_index] = one_spectral_operator.get_representation_index(symmetry_index);
	S_old[symmetry_index] = new_block_sectors_properties_array[sector_column].representation_index[symmetry_index];

	// calculating the number of possible S_z_old values and setting S_z_old to it maximal possible value 

	max_S_z_old[symmetry_index] = min(S_old[symmetry_index], S_new[symmetry_index] + S_added[symmetry_index]);
	min_S_z_old[symmetry_index] = max(-S_old[symmetry_index], - S_new[symmetry_index] - S_added[symmetry_index]);
	old_no *= (int)((max_S_z_old[symmetry_index] - min_S_z_old[symmetry_index])/2) + 1;
	S_z_old[symmetry_index] = max_S_z_old[symmetry_index];
      }//if ! strcmp
    }// for symmetry_index 


    // summing up the CG coefficients 

    for(int old_index = 0; old_index < old_no; old_index++){

      // calculating that to the given S_z_old how many S_z_new belong (in case of SU(2) or charge_SU(2) symmetry) 

      new_block_no = 1;
      for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	if((!strcmp(symmetry_types[symmetry_index].get_name(),"SU(2)")) || (!strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)"))){
	  max_S_z_new[symmetry_index] = min(S_z_old[symmetry_index] + S_added[symmetry_index], S_new[symmetry_index]);
	  min_S_z_new[symmetry_index] = max(S_z_old[symmetry_index] - S_added[symmetry_index], -S_new[symmetry_index]);
	  new_block_no *= (int)((max_S_z_new[symmetry_index] - min_S_z_new[symmetry_index])/2) + 1;
	  S_z_new[symmetry_index] = max_S_z_new[symmetry_index];
	  S_z_added[symmetry_index] = S_z_new[symmetry_index] - S_z_old[symmetry_index];
	}//if 
      }//for int symmetry_index
      
      for(int new_block_index = 0; new_block_index < new_block_no; new_block_index++){
	CG_coefficient_square = 1;
	for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	  if((!strcmp(symmetry_types[symmetry_index].get_name(),"SU(2)")) || (!strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)"))){
	    ////cout << S_old[symmetry_index] << "\t" << S_z_old[symmetry_index] << "\t" << S_added[symmetry_index] << "\t" << S_z_added[symmetry_index] << "\t"
	    //<< S_new[symmetry_index] << "\t" << S_z_new[symmetry_index] << endl;
	    CG_coefficient_square *= clebsch_gordan(S_old[symmetry_index],S_z_old[symmetry_index],S_added[symmetry_index],S_z_added[symmetry_index],S_new[symmetry_index],S_z_new[symmetry_index]);
	  }//if
	}//symmetry_index

	// raising the product of the CG coeffs (coming from the different symmetries) to the 2nd power 
	    
	CG_coefficient_square *= CG_coefficient_square;
	CG_coefficient_square_sum += CG_coefficient_square;
	//cout << "CG_coefficient_square:" << CG_coefficient_square  << endl;
	    
	// changing one inner quantum number of the new state (and correspondingly one of the added state) 

	for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	  while(strcmp(symmetry_types[symmetry_index].get_name(),"SU(2)") && 
		strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)") && 
		symmetry_index < (symmetry_no - 1))
 	    symmetry_index++;

	  if(S_z_new[symmetry_index] > min_S_z_new[symmetry_index]){
	    S_z_new[symmetry_index] -= 2;		    
	    S_z_added[symmetry_index] = S_z_new[symmetry_index] - S_z_old[symmetry_index];
	    symmetry_index = symmetry_no;
	  }//if S_z_new

	  else{
	    S_z_new[symmetry_index] = max_S_z_new[symmetry_index];
	    S_z_added[symmetry_index] = S_z_new[symmetry_index] - S_z_old[symmetry_index];
	  }//else
	}//for symmetry_index
      }//for new_block_index

	/* changing one inner quantum number of the old state */

      for(int symmetry_index = 0; symmetry_index < symmetry_no; symmetry_index++){
	while(strcmp(symmetry_types[symmetry_index].get_name(),"SU(2)") && 
	      strcmp(symmetry_types[symmetry_index].get_name(),"charge_SU(2)") && 
	      symmetry_index < (symmetry_no - 1)) 

	  symmetry_index++;

	if(S_z_old[symmetry_index] > min_S_z_old[symmetry_index]){
	  S_z_old[symmetry_index] -= 2;
	  symmetry_index = symmetry_no;
	}//if
	else{
	  S_z_old[symmetry_index] = max_S_z_old[symmetry_index];
	}//else

      }//for symmetry_index
      
    }//for old_index

    return CG_coefficient_square_sum;
}//clebsch_gordan_coefficients_sum



/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void generate_spectral_function_peaks_nrg(int iteration_index,
					  
					  int symmetry_no, 
					  symmetry * symmetry_types,
					  
					  int new_block_state_no,
					  state *new_block_states_array,
					  
					  int new_block_sector_no,
					  sector_properties * new_block_sectors_properties_array, 
					  					  
					  double * new_block_energies_array, 
					  double * sorted_new_block_energies_array,
					  unsigned long *new_block_energies_index,
					  
					  quantum_operator & spectral_operator_A,
					  quantum_operator & spectral_operator_B
					  ){

/*******************************************************************************
input variables:  fiteen
     iteration_index             - the index of the current iteration. 
     symmetry_no                 - the number of symmetries in the problem.
     symmetry_types              - the types of the symmetries. 

     new_block_state_no                - the number of the new_block_states in the diagonal
                                    basis at the current iteration. 
     new_block_states_array            - array with the number of the new states in 
                                   the orthogonal basis. 

     new_block_sector_no               - the number of sectors in the orthogonal 
                                   basis. 
     new_block_sector_properties_array - array with the properties of the sectors. 

     new_block_energies_array          - array with the eigenvalues of the 
                                   Block_Hamiltonian
     sorte_new_block_energies_array    - sorted array with the eigenenergies of the 
                                   Block_Hamiltonian
     new_block_energies_index          - indexed array for the energies. 

     spectral_operator_A         - the first A+ spectral operator in the 
                                   combination of the two operator for which 
                                   the spectral function is computed. 
   
     spectral_operator_B         - the second B+ spectral operator in the 
                                   combination of the two operator for which 
                                   the spectral function is computed. 

output variables: one

return: void

What it does:
     This routine generates the spectral peaks in th course of the NRG 
     iterations. 
     It uses sum rule so no Clebsch-Gordan coefficients have to be computed. 
     In computes 
     the spectral function delta peaks for a combinations of two operators , A+, B+ 
     but the spectral function corresponds to the <A, B+> Green's function. 
     For each iteration the peaks are saved into a seperate file in the swap 
     folder before exiting the routine. 

How to call:
*******************************************************************************/

  if(spectral_operator_A.get_representation_index()!= spectral_operator_B.get_representation_index()){
    error_message("spectral operators have not the same representation_index.\nEXITING...\n");
  }//if 
  if (spectral_operator_A.get_sgn()!= spectral_operator_B.get_sgn()){
    error_message("spectral operators have not the same signs.\nEXITING...\n");
  }//if


    
  delta_peaks_positive_swap_file_name(delta_peaks_folder_name,
				      iteration_index,
				      delta_peaks_positive_file_name,
				      spectral_operator_A,
				      spectral_operator_B); 
  

  delta_peaks_negative_swap_file_name(delta_peaks_folder_name,
				      iteration_index,
				      delta_peaks_negative_file_name,
				      spectral_operator_A,
				      spectral_operator_B); 
    

  delta_peaks_positive_file_out.open(delta_peaks_positive_file_name, ios::out); 
  if (!delta_peaks_positive_file_out.good()){
    error_message("cannot open delta_peaks_positive_file_out");
  }//if
  
  delta_peaks_negative_file_out.open(delta_peaks_negative_file_name, ios::out);
  
  if (!delta_peaks_negative_file_out.good()){
    error_message("cannot open delta_peaks_negative_file_out");
  }//if
  
  double T; 

  double *peaks_plus;
  double *peaks_minus; 

  peaks_plus = new double [interval_no];
  for(int i = 0; i < interval_no; i++){
    peaks_plus[i] = 0.0;
  }//for i

  peaks_minus = new double [interval_no];
  for(int i = 0; i < interval_no; i++){
    peaks_minus[i] = 0.0;
  }//for i


  double energy_min = sorted_new_block_energies_array[0];
  double energy_max = sorted_new_block_energies_array[new_block_state_no - 1];
 

  // determination of the degeneracy of the ground_state 

  int ground_state_no = 0;

  for(int e_index = 0; e_index < new_block_state_no; e_index++){
    if(fabs(energy_min - new_block_energies_array[e_index]) <degeneracy_threshhold){
      ground_state_no++;
    }//if
  }//for energy_index
   // calculating the peak values in case of zero temperature 

  
  
  if(!(temperature > 0)){
    
    for(int degeneracy_index = 0; degeneracy_index < ground_state_no; degeneracy_index++){

    // We determine to which sector (ground_sect_ind) and in that sector
    //   to which state (ground_state_ind) does the ground state energy belong.

      int energy_min_index = new_block_energies_index[degeneracy_index] - 1;
      int ground_state_sector = new_block_states_array[energy_min_index].get_sector();
      int ground_state_sector_index = energy_min_index - new_block_sectors_properties_array[ground_state_sector].first_state;

      double ground_state_degeneracy= (double)new_block_sectors_properties_array[ground_state_sector].degeneracy;



      //calculating the dimension of the quantum_operator; 

      int operator_degeneracy =1;

      for (int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	if (symmetry_types[symmetry_index].get_abelian_tag()==0){
	  operator_degeneracy *= (spectral_operator_A.get_representation_index(symmetry_index) + 1);
      }//if
      }//symmetry_index
      

      // calculating the peak values for positive energies 
      
      for(int sector_row = 0; sector_row < new_block_sector_no; sector_row++){
	
	if((spectral_operator_A.get_sector_matrix()[sector_row][ground_state_sector].get_activation_state())
	   &&(spectral_operator_B.get_sector_matrix()[sector_row][ground_state_sector].get_activation_state())
	    ) {
	
	  
	  double degeneracy_coefficient =(double) new_block_sectors_properties_array[sector_row].degeneracy / ((double)(operator_degeneracy)*(ground_state_degeneracy)); 

	  
	  //cout << "CG_coef_sq_sum:\t" << CG_coef_sq_sum << endl;
	  
	  for(int sector_row_index = 0; sector_row_index < new_block_sectors_properties_array[sector_row].dimension; sector_row_index++){
	      
	      // calculating the index of the energy which belongs to the state
	      //in the sector: sect_rwind at the position: i between the states 

	    int E_index(0);
	    for(int sector = 0; sector < sector_row; sector++)
	      E_index += new_block_sectors_properties_array[sector].dimension;
	      
	    E_index += sector_row_index;
	      
	    // calculating the index of the interval into which E[E_ind] belongs 
	      
	    double matrix_element_A = spectral_operator_A.get_sector_matrix()[sector_row][ground_state_sector].get_matrix()[sector_row_index][ground_state_sector_index];
	    double matrix_element_B = spectral_operator_B.get_sector_matrix()[sector_row][ground_state_sector].get_matrix()[sector_row_index][ground_state_sector_index];
	    if(new_block_energies_array[E_index] > energy_min){
	      
	      int peak_index = (int)(ceil((float)((new_block_energies_array[E_index] - energy_min) * interval_no / (energy_max - energy_min))) - 1);


	      if(peak_index < interval_no)
		peaks_plus[peak_index] += degeneracy_coefficient * matrix_element_A * matrix_element_B; 
	    }//if energy
	    else {
	      peaks_plus[0] += degeneracy_coefficient * matrix_element_A * matrix_element_B; 
	    }//else
	  }//int sector_row              
	  

	}//if spectral_operator
	
      
      }//for sector_row  
     
            
      // calculating the peak values for negatives new_block_energies_array 
      
      for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++){

	if((spectral_operator_A.get_sector_matrix()[ground_state_sector][sector_column].get_activation_state())
	   && (spectral_operator_B.get_sector_matrix()[ground_state_sector][sector_column].get_activation_state())
	   ) {
	  
	  double degeneracy_coefficient =(double) new_block_sectors_properties_array[ground_state_sector].degeneracy / ((double)(operator_degeneracy)*(ground_state_degeneracy)); 
	  
	  
	    for(int sector_column_index = 0; sector_column_index < new_block_sectors_properties_array[sector_column].dimension; sector_column_index++){
	      
	      // calculating the index of the energy which belongs to the state
	      //in the sector: sect_rwind at the position: i between the states 

	      int E_index(0);
	      for(int sector = 0; sector < sector_column; sector++)
		E_index += new_block_sectors_properties_array[sector].dimension;
	      
	      E_index += sector_column_index;
	      
	      // calculating the index of the interval into which E[E_ind] belongs 
	      
	      double matrix_element_A = spectral_operator_A.get_sector_matrix()[ground_state_sector][sector_column].get_matrix()[ground_state_sector_index][sector_column_index];
	      double matrix_element_B = spectral_operator_B.get_sector_matrix()[ground_state_sector][sector_column].get_matrix()[ground_state_sector_index][sector_column_index];
	      
	      if(new_block_energies_array[E_index] > energy_min){

		int peak_index = (int)(ceil((float)((new_block_energies_array[E_index] - energy_min) * interval_no / (energy_max - energy_min))) - 1);

		if(peak_index < interval_no)
		  //we have a different sign for fermionic of bosonic operator. 
		  peaks_minus[peak_index] += -spectral_operator_A.get_sgn()*degeneracy_coefficient * matrix_element_A * matrix_element_B; 
	      }//if energy
	      else {		
		peaks_minus[0] += -spectral_operator_A.get_sgn()*degeneracy_coefficient * matrix_element_A * matrix_element_B; 
	      }//else
	    }//int sector_row              
	}//if one_operator
      }//for sector_row           
      
    }//for degenaracy_index  

    for(int i = 0; i < interval_no; i++) {      
      peaks_plus[i] = peaks_plus[i] / ground_state_no;
      peaks_minus[i] = peaks_minus[i] / ground_state_no;
    }//i


  }//if !temperature >0
  
  
  if(!(temperature > 0)){
    
    delta_peaks_negative_file_out.precision(double_precision);
    delta_peaks_positive_file_out.precision(double_precision);

        for(int i = 0; i < interval_no; i++) {
	  if((peaks_minus[i] > 0) || (peaks_minus[i] < 0)){
	    delta_peaks_negative_file_out<< (i + 0.5) * (energy_max - energy_min) / interval_no << "\t" << peaks_minus[i]  << endl;	
	  }//if
	  
	  if((peaks_plus[i] > 0) || (peaks_plus[i] < 0)){
	   delta_peaks_positive_file_out<< (i + 0.5) * (energy_max - energy_min) / interval_no << "\t" << peaks_plus[i]  << endl;	
	  }//if
	  

	}//for i
    //positive spectrum.
	   
  }//if !temperature


  //calculating the spectral peaks for positive temperature

  if(temperature  > 0){

    double rescaled_temperature = temperature*pow(lambda, 0.5*((double)(iteration_index)));

      for(int sector_row =0; sector_row< new_block_sector_no; sector_row++){
	for(int sector_column = 0; sector_column < new_block_sector_no; sector_column++){

      //calculating the dimension of the quantum_operator; 

	  int operator_degeneracy =1;
      
	  for (int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
	    if (symmetry_types[symmetry_index].get_abelian_tag()==0){
	      operator_degeneracy *= (spectral_operator_A.get_representation_index(symmetry_index) + 1);
	    }//if
	  }//symmetry_index
      


	if((spectral_operator_A.get_sector_matrix()[sector_row][sector_column].get_activation_state())
	   && (spectral_operator_B.get_sector_matrix()[sector_row][sector_column].get_activation_state())
	   ) {
	  
	  double degeneracy_coefficient =(double) new_block_sectors_properties_array[sector_row].degeneracy / ((double)(operator_degeneracy)); 
	  
	  for(int i = 0; i < new_block_sectors_properties_array[sector_row].dimension; i++){

	      // calculating the index of the energy which belongs to the state
	      //in the sector: sect_rwind at the position: i between the states 
		
	      int energy_row_index(0);

	      for(int sector= 0; sector < sector_row; sector++)
		energy_row_index += new_block_sectors_properties_array[sector].dimension;
	      energy_row_index += i;
		
	      double weight_row = exp(- (new_block_energies_array[energy_row_index] - energy_min) / rescaled_temperature);

	      for(int j = 0; j < new_block_sectors_properties_array[sector_column].dimension; j++){

		// calculating the index of the energy which belongs to the state
		//in the sector: sect_clind at the position: j between the states 

		int energy_column_index(0);
		for(int sector = 0; sector< sector_column; sector++)
		  energy_column_index += new_block_sectors_properties_array[sector].dimension;
		energy_column_index += j;

		double weight_column = exp(- (new_block_energies_array[energy_column_index] - energy_min) / rescaled_temperature);

		// calculating the index of the interval to which E_current = E[E_rwind] - E[E_clind] belongs 

		double delta_energy = new_block_energies_array[energy_row_index] - new_block_energies_array[energy_column_index];
		
		double matrix_element_A = spectral_operator_A.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j];
		double matrix_element_B = spectral_operator_B.get_sector_matrix()[sector_row][sector_column].get_matrix()[i][j];


		if(delta_energy > 0){
		  int peak_index = (int)(ceil((float)(delta_energy * interval_no / (energy_max - energy_min))) - 1);
		  if(peak_index < interval_no)
		    peaks_plus[peak_index] +=  degeneracy_coefficient*matrix_element_A * matrix_element_B * 
		      (weight_column - spectral_operator_A.get_sgn()* weight_row);
		  }//if delta_energy>0;		
	       
		  if(delta_energy<0){
		    int peak_index = (int)(ceil((float)(-delta_energy * interval_no / (energy_max - energy_min))) - 1);
		    if(peak_index < interval_no)
		      peaks_minus[peak_index] +=  degeneracy_coefficient*matrix_element_A * matrix_element_B * 
		  (weight_column - spectral_operator_A.get_sgn()* weight_row);
		  }//if		  

		  if(!(delta_energy > 0) && !(delta_energy < 0)){
		    peaks_plus[0] +=  degeneracy_coefficient*matrix_element_A * matrix_element_B * 
		  (weight_column - spectral_operator_A.get_sgn()* weight_row);
		    peaks_minus[0] +=  degeneracy_coefficient*matrix_element_A * matrix_element_B * 
		  (weight_column - spectral_operator_A.get_sgn()* weight_row);

		  }//if ! delta_energy
		}//for j
	      }// for i	    

	  }//if one_spectral_operator

	}//sector_column
      }//sector_row

      double Z(0);
      int state_sector; 
      double state_degeneracy; 

      for(int e_index = 0; e_index< new_block_state_no; e_index++){
	
	state_sector = new_block_states_array[e_index].get_sector();
	state_degeneracy = (double)new_block_sectors_properties_array[state_sector].degeneracy;
      
	Z += state_degeneracy*exp(- (new_block_energies_array[e_index] - energy_min) / rescaled_temperature);
      }//for e_index;

      for(int i = 0; i < interval_no; i++){
	peaks_plus[i] = peaks_plus[i] / Z;
	peaks_minus[i] = peaks_minus[i] /Z; 
      }//i
  }//if temperature>0


  if(temperature > 0){

    delta_peaks_negative_file_out.precision(double_precision);
    delta_peaks_positive_file_out.precision(double_precision);
        
    for(int i = 0; i < interval_no; i++) {
      
      if((peaks_minus[i] > 0) || (peaks_minus[i] < 0)){
	delta_peaks_negative_file_out<< (i + 0.5) * (energy_max - energy_min) / interval_no << "\t" << peaks_minus[i]  << endl;	
      }//if
      
      if((peaks_plus[i] > 0) || (peaks_plus[i] < 0)){
	delta_peaks_positive_file_out<< (i + 0.5) * (energy_max - energy_min) / interval_no << "\t" << peaks_plus[i]  << endl;	
      }//if
      
      
    }//for i
    //positive spectrum.
    
  }//if !temperature



  delete [] peaks_plus;
  peaks_plus = NULL;
  delete [] peaks_minus;
  peaks_minus = NULL;

  delta_peaks_positive_file_out.close();
  delta_peaks_negative_file_out.close();


  return; 
}//generate_spectral_function_peaks_nrg



void generate_spectral_function_peaks_dmnrg(int iteration_index,
					  
					    int symmetry_no, 
					    symmetry * symmetry_types,
					    
					    int new_block_state_no,
					    state *new_block_states_array,
					    
					    int new_block_sector_no,
					    sector_properties * new_block_sectors_properties_array, 
					  					  
					    double * new_block_energies_array, 
					    double * sorted_new_block_energies_array,
					    unsigned long *new_block_energies_index,
					    
					    quantum_operator & spectral_operator_A,
					    quantum_operator & spectral_operator_B,
					    quantum_operator & new_block_density_matrix
					    ){

/******************************************************************************
input variables:  two
    iteration_index             - the index of the current iteration. 
    symmetry_no                 - the number of symmetries in the problem.
    symmetry_types              - the types of the symmetries. 

    new_block_state_no                - the number of the new_block_states in the diagonal 
                                  basis at the current iteration. 
    new_block_states_array            - array with the number of the new states in 
                                  the orthogonal basis. 

    new_block_sector_no               - the number of sectors in the orthogonal basis. 
    new_block_sector_properties_array - array with the properties of the sectors. 
    spectral_operator_A         - the first A+ spectral operator in the 
                                  combination of the two operator for which the
                                  spectral function is computed. 
                                 
    spectral_operator_B         - the second B+ spectral operator in the 
                                  combination of the two operator for which the
                                  spectral function is computed. 
    
    new_block_density_matrix          - the reduced density matrix in the orthogonal
                                  basis at the given iteration_index.
    
output variables: none

return: void

What it does:
    This routine generates the spectral peaks in the course of the DMNRG in 
    the backward iteration. 
    It uses sum rules so no Clebsch-Gordan coefficients have to be computed. In computes 
    the spectral function delta peaks for a combinations of two operators , A+, B+ 
    but the spectral function corresponds to the <A, B+> Green's function. 
    For each iteration the peaks are saved into a seperate file in the swap folder
    before exiting the routine. 

How to call:
*******************************************************************************/

  if(spectral_operator_A.get_representation_index()!= spectral_operator_B.get_representation_index()){
    error_message("spectral operators have not the same representation_index.\nEXITING...\n");
  }//if 
  if (spectral_operator_A.get_sgn()!= spectral_operator_B.get_sgn()){
    error_message("spectral operators have not the same signs.\nEXITING...\n");
  }//if

  double sgn_A =(double)( spectral_operator_A.get_sgn());

  double energy_min = sorted_new_block_energies_array[0]; 
  double energy_max = sorted_new_block_energies_array[new_block_state_no-1]; 

  //cout<<energy_min-sorted_new_block_energies_array[0]<<endl;
  //cout<<energy_max-sorted_new_block_energies_array[0]<<endl;

  //calculating the dimension of the quantum_operator;

 double  degeneracy_spectral_operator_A =1.0;
  
  for (int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
    if (symmetry_types[symmetry_index].get_abelian_tag()==0){
      degeneracy_spectral_operator_A *= (double)(spectral_operator_A.get_representation_index(symmetry_index) + 1);
    }//if
  }//symmetry_index

  /* open files to dump delta peaks */ 
  
  delta_peaks_positive_dmnrg_swap_file_name(delta_peaks_folder_name,
					    iteration_index,
					    delta_peaks_positive_dmnrg_file_name,
					    spectral_operator_A,
					    spectral_operator_B);


  delta_peaks_positive_dmnrg_file_out.open(delta_peaks_positive_dmnrg_file_name, ios::out); 
  if (!delta_peaks_positive_dmnrg_file_out.good()){
    error_message("unable to open the delta_peaks_positive_dmnrg_file_out.");
  }//if

  delta_peaks_negative_dmnrg_swap_file_name(delta_peaks_folder_name,
					    iteration_index,
					    delta_peaks_negative_dmnrg_file_name,
					    spectral_operator_A,
					    spectral_operator_B);


  delta_peaks_negative_dmnrg_file_out.open(delta_peaks_negative_dmnrg_file_name, ios::out); 
  if (!delta_peaks_negative_dmnrg_file_out.good()){
    error_message("unable to open the delta_peaks_negative_dmnrg_file_out.");
  }//if

  /* create arrays for the delta peaks */ 

  double *peaks_plus; 
  double *peaks_minus;


  peaks_plus = new double [interval_no];
  for(int i = 0; i < interval_no; i++){
    peaks_plus[i] = 0.0;
  }//for i
  
  peaks_minus = new double [interval_no];
  for(int i = 0; i < interval_no; i++){
    peaks_minus[i] = 0.0;
  }//for i

  /* Compute the two terms of the expression of the spectral function separately */ 

  //this is the first term; 
  
  for (int sector_1=0; sector_1<new_block_sector_no; sector_1++){
    for(int sector_3=0; sector_3<new_block_sector_no; sector_3++){

	double degeneracy_sector_3 = (double)new_block_sectors_properties_array[sector_3].degeneracy;


	if(new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_activation_state() &&
	   spectral_operator_A.get_sector_matrix()[sector_3][sector_1].get_activation_state() &&
	   spectral_operator_B.get_sector_matrix()[sector_3][sector_1].get_activation_state()
	   ){

      
	  for (int state_1_index=0; state_1_index<new_block_sectors_properties_array[sector_1].dimension; state_1_index++){

	    int energy_1_index = 0; 
	    for(int sector= 0; sector < sector_1; sector++)
	      energy_1_index += new_block_sectors_properties_array[sector].dimension;
	    energy_1_index += state_1_index;
	    
 	    int state_1 = new_block_sectors_properties_array[sector_1].first_state + state_1_index; 
	    
	    for (int state_2_index=0; state_2_index<new_block_sectors_properties_array[sector_1].dimension; state_2_index++){

	      int energy_2_index = 0; 
	      for(int sector= 0; sector < sector_1; sector++)
		energy_2_index += new_block_sectors_properties_array[sector].dimension;
	      energy_2_index += state_2_index;

	      int state_2 = new_block_sectors_properties_array[sector_1].first_state + state_2_index; 

	      if (new_block_states_array[state_1].get_elimination_tag() != new_block_states_array[state_2].get_elimination_tag()) continue;

	      double matrix_element_density_matrix = new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_1_index][state_2_index]; 
	      if (matrix_element_density_matrix ==0.0) continue; 

	      for(int state_3_index=0; state_3_index<new_block_sectors_properties_array[sector_3].dimension; state_3_index++){

		int energy_3_index =0; 
		for (int sector =0; sector<sector_3; sector++)
		  energy_3_index += new_block_sectors_properties_array[sector].dimension;		  
		energy_3_index += state_3_index; 

		int state_3 = new_block_sectors_properties_array[sector_3].first_state + state_3_index; 
		
		if ((new_block_states_array[state_2].get_elimination_tag() == 'K') &&
		    (new_block_states_array[state_3].get_elimination_tag() == 'K')) continue; 
		
	
		double matrix_element_spectral_operator_A = spectral_operator_A.get_sector_matrix()[sector_3][sector_1].get_matrix()[state_3_index][state_2_index]; 
		double matrix_element_spectral_operator_B = spectral_operator_B.get_sector_matrix()[sector_3][sector_1].get_matrix()[state_3_index][state_1_index]; 

		double energy =  new_block_energies_array[energy_3_index]- 0.5*(new_block_energies_array[energy_1_index]+new_block_energies_array[energy_2_index]); 

		

		if(energy > 0) {
		  int peak_index = (int)(ceil((double)(energy * interval_no / (energy_max - energy_min))) - 1);
		  if(peak_index < interval_no)
		    peaks_plus[peak_index] +=  (degeneracy_sector_3/degeneracy_spectral_operator_A) * 
		      matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B; 
		}// if energy> 0; 

		
		if(energy < 0){
		    int peak_index = (int)(ceil((double)(energy * interval_no / (energy_min - energy_max))) - 1);
		    if(peak_index < interval_no)
		      peaks_minus[peak_index] += (degeneracy_sector_3/degeneracy_spectral_operator_A) * 
			matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B; 
		}//if energy<0;


		if (!(energy>0) && !(energy<0)){
		  peaks_plus[0] += (degeneracy_sector_3/degeneracy_spectral_operator_A) *
		    matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B;
		  peaks_minus[0] += (degeneracy_sector_3/degeneracy_spectral_operator_A) *
		    matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B;
		}//if not
		

	      }//for state_3_index	      
	    }//for state_2_index	    
	  }//state_1_index	  
	}//if new_block_density_matrix.get_sector....	
      }//for sector_3
  }//for sector_1

   
  //now comes the second term in the spectral function


  for (int sector_1=0; sector_1<new_block_sector_no; sector_1++){
      for(int sector_3=0; sector_3<new_block_sector_no; sector_3++){

	double degeneracy_sector_1 = (double)new_block_sectors_properties_array[sector_1].degeneracy;

	if(new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_activation_state() &&
	   spectral_operator_B.get_sector_matrix()[sector_1][sector_3].get_activation_state() &&
	   spectral_operator_A.get_sector_matrix()[sector_1][sector_3].get_activation_state()
	   ){
      
	  for (int state_1_index=0; state_1_index<new_block_sectors_properties_array[sector_1].dimension; state_1_index++){

	    int energy_1_index = 0; 
	    for(int sector= 0; sector < sector_1; sector++)
	      energy_1_index += new_block_sectors_properties_array[sector].dimension;
	    energy_1_index += state_1_index;
	    
 	    int state_1 = new_block_sectors_properties_array[sector_1].first_state + state_1_index; 

	    
	    for (int state_2_index=0; state_2_index<new_block_sectors_properties_array[sector_1].dimension; state_2_index++){

	      int energy_2_index = 0; 
	      for(int sector= 0; sector < sector_1; sector++)
		energy_2_index += new_block_sectors_properties_array[sector].dimension;
	      energy_2_index += state_2_index;

	      int state_2 = new_block_sectors_properties_array[sector_1].first_state + state_2_index; 

	      if (new_block_states_array[state_1].get_elimination_tag() != new_block_states_array[state_2].get_elimination_tag()) continue;
	      double matrix_element_density_matrix = new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_1_index][state_2_index]; 
	      if (matrix_element_density_matrix ==0) continue; 


	      for(int state_3_index=0; state_3_index<new_block_sectors_properties_array[sector_3].dimension; state_3_index++){

		int energy_3_index =0; 
		for (int sector =0; sector<sector_3; sector++)
		  energy_3_index += new_block_sectors_properties_array[sector].dimension;		  
		energy_3_index += state_3_index; 

		int state_3 = new_block_sectors_properties_array[sector_3].first_state + state_3_index; 


		if ((new_block_states_array[state_2].get_elimination_tag() == 'K') &&
		    (new_block_states_array[state_3].get_elimination_tag()== 'K')) continue; 
		
		

		double matrix_element_density_matrix = new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_1_index][state_2_index]; 
		double matrix_element_spectral_operator_B = spectral_operator_B.get_sector_matrix()[sector_1][sector_3].get_matrix()[state_2_index][state_3_index]; 
		double matrix_element_spectral_operator_A = spectral_operator_A.get_sector_matrix()[sector_1][sector_3].get_matrix()[state_1_index][state_3_index]; 

		double energy =  -(new_block_energies_array[energy_3_index]- 0.5*(new_block_energies_array[energy_1_index]+new_block_energies_array[energy_2_index])); 

		
		if(energy > 0) {
		  int peak_index = (int)(ceil((float)(energy * interval_no / (energy_max - energy_min))) - 1);
		  if(peak_index < interval_no)
		    peaks_plus[peak_index] += -sgn_A*(degeneracy_sector_1/degeneracy_spectral_operator_A) * 
		      matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B; 
		}// if energy> 0; 

		
		if(energy < 0){
		    int peak_index = (int)(ceil((float)(energy * interval_no / (energy_min - energy_max))) - 1);
		    if(peak_index < interval_no)
		      peaks_minus[peak_index] += -sgn_A*(degeneracy_sector_1/degeneracy_spectral_operator_A) * 
			matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B; 
		}//if energy<0;


		if (!(energy>0) && !(energy<0)){
		  peaks_plus[0] += -sgn_A*(degeneracy_sector_1/degeneracy_spectral_operator_A) *
		    matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B;

		  peaks_minus[0] += -sgn_A*(degeneracy_sector_1/degeneracy_spectral_operator_A) *
		  matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B;
		}//if not
		
	      }//for state_3_index	      
	    }//for state_2_index	    
	  }//state_1_index	  
	}//if new_block_density_matrix.get_sector....	
      }//for sector_3
  }//for sector_1
  
  

  //now we save the data into the files. 
  
  
  for(int i = 0; i < interval_no; i++) {
    //negative spectrum
    if((peaks_minus[i] > 0) || (peaks_minus[i] < 0)){
      delta_peaks_negative_dmnrg_file_out.precision(double_precision); 
      delta_peaks_negative_dmnrg_file_out <<(i + 0.5) * (energy_max - energy_min) / interval_no <<"\t" << peaks_minus[i]  << endl;
    }//if
    //positive spectrum.
  if((peaks_plus[i] > 0) || (peaks_plus[i] < 0)){
    //cout<<iteration_index<<"\t"<< (i + 1) * (energy_max - energy_min) / interval_no << "\t" << peaks_plus[i]  << endl;
    delta_peaks_positive_dmnrg_file_out.precision(double_precision); 
    delta_peaks_positive_dmnrg_file_out << (i + 0.5) * (energy_max - energy_min) / interval_no << "\t" << peaks_plus[i]  << endl;
    }//if
  
  }//for i

  delete [] peaks_plus;
  peaks_plus = NULL;
  delete [] peaks_minus; 
  peaks_minus = NULL;

  delta_peaks_positive_dmnrg_file_out.close();
  delta_peaks_negative_dmnrg_file_out.close();

  return; 
  
}//generate_spectral_function_peaks_dmnrg















void generate_static_average_dmnrg(int iteration_index,
				   
				   int symmetry_no, 
				   symmetry * symmetry_types,
				   
				   int new_block_state_no,
				   state *new_block_states_array,
				   
				   int new_block_sector_no,
				   sector_properties * new_block_sectors_properties_array, 
				   				   
				   quantum_operator & spectral_operator_A,
				   quantum_operator & spectral_operator_B,
				   quantum_operator & new_block_density_matrix
				   ){

/******************************************************************************
input variables:  two
    iteration_index             - the index of the current iteration. 
    symmetry_no                 - the number of symmetries in the problem.
    symmetry_types              - the types of the symmetries. 

    new_block_state_no                - the number of the new_block_states in the diagonal 
                                  basis at the current iteration. 
    new_block_states_array            - array with the number of the new states in 
                                  the orthogonal basis. 

    new_block_sector_no               - the number of sectors in the orthogonal basis. 
    new_block_sector_properties_array - array with the properties of the sectors. 
    spectral_operator_A         - the first A+ spectral operator in the 
                                  combination of the two operator for which the
                                  spectral function is computed. 
                                 
    spectral_operator_B         - the second B+ spectral operator in the 
                                  combination of the two operator for which the
                                  spectral function is computed. 
    
    new_block_density_matrix          - the reduced density matrix in the orthogonal
                                  basis at the given iteration_index.
    
output variables: none

return: void

What it does:

    This routine generates the static averages of a combinations of two operator
    in the course of DMNRG backward iteration. 
    It uses sum rules so no Clebsch-Gordan coefficients have to be computed. In computes 
    the static average for a combinations of two operators , A+, B+ 
    but the static average to the <A+, B> Green's function. 
    All the results are saved in a single file and the iteration_index is indexing the results. 

How to call:
*******************************************************************************/



  char static_average_file_name[400];

  if(spectral_operator_A.get_representation_index()!= spectral_operator_B.get_representation_index()){
    error_message("spectral operators have not the same representation_index.\nEXITING...\n");
  }//if 
  if (spectral_operator_A.get_sgn()!= spectral_operator_B.get_sgn()){
    error_message("spectral operators have not the same signs.\nEXITING...\n");
  }//if

  double sgn_A =(double)( spectral_operator_A.get_sgn());

  //calculating the dimension of the quantum_operator;

 double  degeneracy_spectral_operator_A =1.0;
  
  for (int symmetry_index=0; symmetry_index<symmetry_no; symmetry_index++){
    if (symmetry_types[symmetry_index].get_abelian_tag()==0){
      degeneracy_spectral_operator_A *= (double)(spectral_operator_A.get_representation_index(symmetry_index) + 1);
    }//if
  }//symmetry_index

  

  static_average_spectral_operators_swap_file_name(static_operators_folder_name,
						   static_average_file_name,
						   spectral_operator_A,
						   spectral_operator_B);

  ofstream static_average_file_out;
  
  if (iteration_index==iteration_no){
    static_average_file_out.open(static_average_file_name, ios::out); 
  }//if
  else{
    static_average_file_out.open(static_average_file_name, ios::app);
  }//else

  if (!static_average_file_out.good()){
    error_message("unable to open static_average_file_out.");
  }//if
  

  double static_average_quantum_operators_sum=0; 


  //this is actually the contribution of the current iteration. 
  
  for (int sector_1=0; sector_1<new_block_sector_no; sector_1++){
    double degeneracy_sector_1 = (double)new_block_sectors_properties_array[sector_1].degeneracy;
      for(int sector_3=0; sector_3<new_block_sector_no; sector_3++){



	if(new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_activation_state() &&
	   spectral_operator_A.get_sector_matrix()[sector_1][sector_3].get_activation_state() &&
	   spectral_operator_B.get_sector_matrix()[sector_1][sector_3].get_activation_state()
	   ){

      
	  for (int state_1_index=0; state_1_index<new_block_sectors_properties_array[sector_1].dimension; state_1_index++){

 	    int state_1 = new_block_sectors_properties_array[sector_1].first_state + state_1_index; 

	    
	    for (int state_2_index=0; state_2_index<new_block_sectors_properties_array[sector_1].dimension; state_2_index++){

	      int state_2 = new_block_sectors_properties_array[sector_1].first_state + state_2_index; 

	      if (new_block_states_array[state_1].get_elimination_tag() != new_block_states_array[state_2].get_elimination_tag()) continue;
	      
	      double matrix_element_density_matrix = new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_1_index][state_2_index]; 
	      if (matrix_element_density_matrix ==0.0) continue; 
	     
 
	      for(int state_3_index=0; state_3_index<new_block_sectors_properties_array[sector_3].dimension; state_3_index++){
				
		int state_3 = new_block_sectors_properties_array[sector_3].first_state + state_3_index; 
		
		if ((new_block_states_array[state_2].get_elimination_tag() == 'K') &&
		    (new_block_states_array[state_3].get_elimination_tag() == 'K')) continue; 
		
		double matrix_element_spectral_operator_A = spectral_operator_A.get_sector_matrix()[sector_1][sector_3].get_matrix()[state_2_index][state_3_index]; 
		double matrix_element_spectral_operator_B = spectral_operator_B.get_sector_matrix()[sector_1][sector_3].get_matrix()[state_1_index][state_3_index]; 


		static_average_quantum_operators_sum += (degeneracy_sector_1/degeneracy_spectral_operator_A) *
		  matrix_element_density_matrix*matrix_element_spectral_operator_A*matrix_element_spectral_operator_B; 
	       
	      }//for state_3_index	      
	    }//for state_2_index	    
	  }//state_1_index	  
	}//if new_block_density_matrix.get_sector....	
      }//for sector_3
  }//for sector_1

 
  //now we save the data into the files. 
  
  
  static_average_file_out.precision(double_precision); 

  static_average_file_out <<iteration_index<<"\t" << static_average_quantum_operators_sum << endl;


  static_average_file_out.close();
  static_average_file_out.clear();

  return; 
  
}//generate_static_average_dmnrg


void generate_static_average_dmnrg(int iteration_index,
				   
				   int symmetry_no, 
				   symmetry * symmetry_types,
				   
				   int new_block_state_no,
				   state *new_block_states_array,
				   
				   int new_block_sector_no,
				   sector_properties * new_block_sectors_properties_array, 
				   				   
				   quantum_operator & static_operator_A,
				   quantum_operator & new_block_density_matrix
				   ){

/******************************************************************************
input variables:  two
    iteration_index             - the index of the current iteration. 
    symmetry_no                 - the number of symmetries in the problem.
    symmetry_types              - the types of the symmetries. 

    new_block_state_no                - the number of the new_block_states in the diagonal 
                                  basis at the current iteration. 
    new_block_states_array            - array with the number of the new states in 
                                  the orthogonal basis. 

    new_block_sector_no               - the number of sectors in the orthogonal basis. 
    new_block_sector_properties_array - array with the properties of the sectors. 
    static_operator_A         - the static operator  for which the
                                  spectral function is computed. 
                                 
    spectral_operator_B         - the second B+ spectral operator in the 
                                  combination of the two operator for which the
                                  spectral function is computed. 
    
    new_block_density_matrix          - the reduced density matrix in the orthogonal
                                  basis at the given iteration_index.
    
output variables: none

return: void

What it does:

    This routine generates the static averages of a combinations of two operator
    in the course of DMNRG backward iteration. 
    It uses sum rules so no Clebsch-Gordan coefficients have to be computed. In computes 
    the static average for a combinations of two operators , A+, B+ 
    but the static average to the <A+, B> Green's function. 
    All the results are saved in a single file and the iteration_index is indexing the results. 

How to call:
*******************************************************************************/


  char static_average_file_name[400];
  
  if(static_operator_A.get_representation_index()!= new_block_density_matrix.get_representation_index()){
    error_message("static operator does not have  the  representation_index zero\nEXITING...\n");
  }//if 
  
  double sgn_A =(double)( static_operator_A.get_sgn());


  
  static_average_static_operator_swap_file_name(static_operators_folder_name,
						   static_average_file_name,
						   static_operator_A);

  ofstream static_average_file_out;
  
  if (iteration_index==iteration_no){
    static_average_file_out.open(static_average_file_name, ios::out); 
  }//if
  else{
    static_average_file_out.open(static_average_file_name, ios::app);
  }//else

  if (!static_average_file_out.good()){
    error_message("unable to open static_average_file_out.");
  }//if
  

  double static_average_quantum_operator_sum=0; 



  //this is actually the contribution of the current iteration. 

  
  for (int sector_1=0; sector_1<new_block_sector_no; sector_1++){

    double degeneracy_sector_1 = (double)new_block_sectors_properties_array[sector_1].degeneracy;

      if(new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_activation_state() &&
	 static_operator_A.get_sector_matrix()[sector_1][sector_1].get_activation_state() 
	 ){

	  for (int state_1_index=0; state_1_index<new_block_sectors_properties_array[sector_1].dimension; state_1_index++){
 	    int state_1 = new_block_sectors_properties_array[sector_1].first_state + state_1_index; 
	    
	    for (int state_2_index=0; state_2_index<new_block_sectors_properties_array[sector_1].dimension; state_2_index++){
	      int state_2 = new_block_sectors_properties_array[sector_1].first_state + state_2_index; 

	      if (new_block_states_array[state_1].get_elimination_tag() != new_block_states_array[state_2].get_elimination_tag()) continue;
	      		
	      if (new_block_states_array[state_1].get_elimination_tag() == 'K') continue;
	      if (new_block_states_array[state_2].get_elimination_tag() == 'K') continue; 
		
	      double matrix_element_density_matrix = new_block_density_matrix.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_1_index][state_2_index]; 
	      double matrix_element_static_operator_A = static_operator_A.get_sector_matrix()[sector_1][sector_1].get_matrix()[state_2_index][state_1_index]; 


		static_average_quantum_operator_sum += degeneracy_sector_1 * matrix_element_density_matrix*matrix_element_static_operator_A; 
	       
	    }//for state_2_index	    
	  }//state_1_index	  
	}//if new_block_density_matrix.get_sector....	
  }//for sector_1

  //now we save the data into the files. 
    
  static_average_file_out.precision(double_precision); 

  static_average_file_out <<iteration_index<<"\t" << static_average_quantum_operator_sum << endl;

  static_average_file_out.close();
  static_average_file_out.clear();

  return; 
  
}//generate_static_average_dmnrg



