/****************************************************************************

Last Update : 2007/09/25 Pascu Moca 

No documentation update necessary


*****************************************************************************/


#include "../include/state.h"

/*----------------------------------------------------------------------------*/
/*---------------CONSTRUCTORS AND DESTRUCTORS---------------------------------*/
/*----------------------------------------------------------------------------*/

state::state(){
/*********************************************************************
class states

    variables definition
        sector               - tells the sector the state belong to.
        sector_index         - index of the state within the sector, 
                               values between 0 and sector_size -1.
        representation_index - these are the quantum numbers describing the states.
        parent_state_index   - current state index of the parent state.
        local_state_index    - local state index of the parent state.
        sign                 - is the sign of the state. For bosons it is always 
                               1, while for fermions it can be +/- 1. On default
                               it is set to zero.
                               

Implicit constructor for the llvector class.

input variables = none

output variables  = none

return = none

What it does:
     set the 'sector', 'sector_index', 'parent_state_index', 
     'local_state_index', 'sgn' to zero and  
     'elimination_tag' to 'K' which labels kept states; 

how to call:
     state();
****************************************************************************/

  sector = sector_index = parent_state_index = local_state_index = sgn = 0;
  elimination_tag = 'K'; 
  //cout << "state constructor" << endl;
}; //state

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

state::~state()	{

/****************************************************************************
Implicit destructor for the class.

input variables = none

output variables = none

return = none

What it does:
     Destroys the class state.

how to call:
    ~state();
****************************************************************************/

}//~state
       
/*----------------------------------------------------------------------------*/
/*-------------------MEMBERS FUNCTIONS----------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::activate(int no_of_symmetries){ 
  if(!representation_index.get_ptr()){
    representation_index.allocate(no_of_symmetries);    
  } 	 
}//activate

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_parent_state_index() {
  return(parent_state_index);
}//get_parent_state_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_parent_state_index(int ind){
  parent_state_index = ind;
}//set_parent_state_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_sgn(void) {
/****************************************************************************
input variables = none

output variables = none

return = an integer that is the sign of the state.

What it does:
     returns the sign of the state (+/- 1)
how to call:
****************************************************************************/

  return(sgn);
}//sgn

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_sgn(int i) {
/****************************************************************************
input variables = one
     int i - sign of the state (+/- 1)

output variables = none

return =void

What it does:
     set the sign of the state to i
how to call:
****************************************************************************/

  sgn = i;
}//set_sgn

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_local_state_index(void) {
  return(local_state_index);
}//get_local_state_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_local_state_index(int ind) {
  local_state_index = ind;
}//set_local_state_index; 

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
 
int state::get_sector(void) {
  return(sector);
}//get_sector_number; 

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_sector(int i) { 
  sector = i;
}//set_sector

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_sector_index(void) {
  return(sector_index);
}//state_sector_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_sector_index(int i) {
  sector_index = i;
}//set_sector_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_representation_index(int sy_index) {
  return(representation_index[sy_index]);
}//get_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_representation_index(int sym_index,int rep_index){
  representation_index[sym_index] = rep_index;
}//set_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::get_representation_index_size(){
  return(representation_index.get_size());
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void state::set_representation_index(llvector & rep_index_master){ 
  int size = representation_index.get_size();
  if (size != rep_index_master.get_size())  {
    cerr << "Sizes do not match in set_repind\n exiting..."; 
    exit(1);
  }//if

  for(int ind = 0; ind < size; ind++) 
    representation_index[ind] = rep_index_master[ind];
}//set_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
 
llvector & state::get_representation_index() {
  return(representation_index);
}//get_representation_index



void state::set_elimination_tag(char new_tag){
  elimination_tag= new_tag; 
}//set_elimination_tag


char state::get_elimination_tag(){
  return elimination_tag; 
}//get_elimination_tag

/*----------------------------------------------------------------------------*/
/*------------------------------OPERATORS-------------------------------------*/
/*----------------------------------------------------------------------------*/


state & state::operator = (state & V){
/****************************************************************************
input variables = one
      state & V - a reference to a state class.

output variables = none

return = the reference to the state class.

What it does:
     '=' is similar with a copy constructor so the input state
      reference variable is copied in the *this variable

how to call:
     state & operator = (state & );
****************************************************************************/


  if(!representation_index.get_ptr()) {   
    
    /* if this is not yet allocated */
    /* I allocate it to be the same size as V */ 
    representation_index.allocate(V.representation_index.get_size());    
  } 	 
  representation_index = V.representation_index;
  sector               = V.sector;
  sector_index         = V.sector_index;
  parent_state_index   = V.parent_state_index;
  local_state_index    = V.local_state_index;
  sgn                  = V.sgn;
  elimination_tag      = V.elimination_tag; 

  return (* this);
}//operator = 

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::operator < (state & V2){

/****************************************************************************
input variables = one
      state & V  - a reference to a state class.

output variables = none

return = 1 if the representation index of the V state is larger than the 
         representation index of the present class. 
         0 otherwise.
What it does:
     '<' tests if representation index of state V is larger then the 
     representation index of the present class. 

how to call:
     int operator < (state & );
****************************************************************************/

  if(representation_index < V2.representation_index) return(1);
  else   return(0);
}//operator < 



/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int state::operator == (state & V){

/****************************************************************************
input variables = one
      state & V  - a reference to a llvector class.

output variables = none

return = 1 if the V is equal with the present state class variable
         0 if V does not match with the present class element.

What it does:
     '==' tests if two state variables are equal in all respects. If they are
     then returns 1 otherwise returns 0.

how to call:
     int operator == (state & );
****************************************************************************/

  if(representation_index != V.representation_index) return(0);
  if(sector               != V.sector)               return (0);
  if(sector_index         != V.sector_index)         return (0);
  if(parent_state_index   != V.parent_state_index)   return (0); 
  if(local_state_index    != V.local_state_index)    return (0);
  if (sgn                 != V.sgn)                  return (0);
  if (elimination_tag     != V.elimination_tag)      return (0); 
  return(1);
}//operator < 

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

ostream & state::print(ostream & os ){

/****************************************************************************
input variables = one
      ostream & os - stream where the state class is printed

output variables = none

return = the stream where the state class is printed

What it does:
     print the properties of the state into the os stream. This can be a file
     or any other output redirected.

how to call:
****************************************************************************/

  os<<endl;
  os <<"sector="<<  sector <<endl ;
  os <<"sector_index="<<  sector_index<<endl;
  os <<"representation_index="<< representation_index<<endl;
  os <<"parent_state_index="<<parent_state_index<<endl;
  os <<"local_state_index=" << local_state_index<<endl; 
  os <<"sgn="<< sgn<<endl;
  os <<"elimination_tag="<<elimination_tag<<endl;
  return os;
}//print

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

ostream & operator << (ostream & os, state& V){

/****************************************************************************
input variables = two
      ostream & os - stream where the state class is printed
      state V      - a reference to a state class

output variables = none

return = the stream where the state class is printed

What it does:
     '<<' print the V state into the os stream. This can be a file
     or any other output redirected.

how to call:
     friend  ostream & operator << (ostream & , state & );
****************************************************************************/

  return(V.print(os));
}//ostream 

/*----------------------------------------------------------------------------*/
/*-------------------------FRIEND FUNCTIONS-----------------------------------*/
/*----------------------------------------------------------------------------*/

int check_states_order (unsigned long n, state ra[], state rb[]){

/****************************************************************************
input variables = three
      unsigned long n - the length of vector ra[] and rb[]
      state ra[]      - array of length n
      state rb[]      - array of length n

output variables = one

return = 

What it does:
       Returns 1 if each element of vectors ra and rb is the same, otherwise,
       returns 0.
how to call:
****************************************************************************/

  int ordered = 1; 
  for (int i=0; i<n; i++){

    if (ra[i] == rb[i]){
      ordered *= 1; 
    }//if
    else ordered *=0; 
  }//i
  return ordered; 

}//check_states_ordering

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void sort_states(unsigned long n, state ra[]){

/****************************************************************************
input variables = two
     unsigned long n - length of array ra[]
     state ra[]      - array of length n

output variables = one
     state ra[]      - array of length n

return = void

What it does:
     returns the sorted form of state array ra[] according to ???? sorting 
     algorithm.
how to call:
****************************************************************************/

  unsigned long i,ir,j,l;
  state rra;

  
  if (n < 2) return;  /* Nothing to sort if n==1 */ 
  l=(n >> 1)+1;
  ir=n;
  for (;;) {
    if (l > 1) {
      rra=ra[--l];
    } else {
      rra=ra[ir];
      ra[ir]=ra[1];
      if (--ir == 1) {
	ra[1]=rra;
	break;
      }
    }
    i=l;
    j=l+l;
    while (j <= ir) {
      if (j < ir && ra[j] < ra[j+1]) j++;
      if (rra < ra[j]) {
	ra[i]=ra[j];
	i=j;
	j <<= 1;
      } else j=ir+1;
    }
    ra[i]=rra;
  }
}

