/****************************************************************************

Last Update : 2007/09/25 Pascu Moca 

No documentation update necessary


*****************************************************************************/

#include "../include/symmetry.h"

/*----------------------------------------------------------------------------*/
/*---------------CONSTRUCTORS AND DESTRUCTORS---------------------------------*/
/*----------------------------------------------------------------------------*/

symmetry::symmetry(void) { 

/****************************************************************************
symmetry class:
     this is a class of symmetry where possible symmetry types are defined
     in symmetry.h: 
     possible_symmetry_type_no = 6 
     possible_symmetry_type[possible_symmetry_type_no] = 
     {"U(1)","SU(2)","SU(3)","Z(2)","SL(2,C)","charge_SU(2)"} 

members:
     max_representation_index - maximum value of the representation index
     min_representation_index - minimum value of the representation index
     name                     - name of the symmetry
     abelian_tag              - zero for non-abelian symmetries and 
                                1 for abelian ones.

Implicit constructor for the symmetry class.

input variables = none

output variables  = none

return = none

What it does:
     set members to default values, i.e., max_representation_index and 
     min_representation_index to zero, name to NULL and abelian_tag to 0;

how to call:
     symmetry();
****************************************************************************/

  max_representation_index = min_representation_index = 0;
  name = NULL;
  abelian_tag = 0;  
}//symmetry


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

symmetry::~symmetry(){

/****************************************************************************
Implicit destructor for the class.

input variables = none

output variables = none

return = none

What it does:
     Destroys the class symmetry and set members to default values.

how to call:
    ~symmetry();
****************************************************************************/

  max_representation_index = min_representation_index = 0;
  if(name){
    delete name; 
    name = NULL;
  }//if
  abelian_tag = 0; 
	    //cout << "destructor of symmetry" << endl;
}//~symmetry

/*----------------------------------------------------------------------------*/
/*-------------------MEMBERS FUNCTIONS----------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::activate(){
/****************************************************************************

At first the symmetry must be activated. 

input variables = none

output variables = none

return = void

What it does:
     Allocates character array of length 256 characters.

how to call:
    activate();
****************************************************************************/

  if(!name)    /* if not active */
    name = new char[256];
  if(!name) {
    cerr << "Allocation failure: symmetry cannot be activated.\nError in routine symmetry::activate()\n EXITING...\n"; 
    exit(-1);
  }//if
}//activate


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::clear() {	

/****************************************************************************
input variables = none

output variables = none

return =void;

What it does:
     'clear' destroys the symmetry class if activated and set members to 
     default values. It is similar to a destructor.

how to call:
    void clear();
****************************************************************************/
 
  max_representation_index = min_representation_index = 0;
  /* remove the name */
  if(name) {    /* if active */
    delete [] name;
    name = NULL;
  }
  abelian_tag = 0; 
}//clear

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int symmetry::get_max_representation_index(){

/****************************************************************************
input variables = none

output variables = none

return = an integer that is the maximum value of the representation index;

What it does:
     returns the maximum value of representation index of the symmetry class.

how to call:
    int get_max_representation_index();
****************************************************************************/

  return(max_representation_index);
}//get_max_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int symmetry::get_min_representation_index(){

/****************************************************************************
input variables = none

output variables = none

return = an integer that is the minimum value of the representation index;

What it does:
     returns the minimum value of representation index of the symmetry class.

how to call:
     int get_max_representation_index();
****************************************************************************/

  return(min_representation_index);
}//get_min_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int symmetry::get_abelian_tag(){
/****************************************************************************
input variables = none

output variables = none

return = an integer that is the abelian_tag (0 for abelian and 1 for non-abelian
         symmetry)

What it does:
     returns the abelian_tag index of the symmetry class.

how to call:
     int get_max_representation_index();
****************************************************************************/

  return (abelian_tag); 
}//get_abelian_tag

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::set_max_representation_index (int max){

/****************************************************************************
input variables =one
     int max - maximum value of the representation index 

output variables = none

return = void

What it does:
     set max_represnation index to max 

how to call:
     void set_max_representation_index(int max);
****************************************************************************/

  max_representation_index = max;
}//set_max_representation_index

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::set_min_representation_index (int min){
/****************************************************************************
input variables =one
      int min - minimum value of the representation index 

output variables = none

return =void

What it does:
     set min_represnation index to min 
how to call:
    void set_min_representation_index(int min);
****************************************************************************/

  min_representation_index = min;
}//set_min_representation_index


/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/


void symmetry::set_name (char * word) {

/****************************************************************************
input variables =one
      char * word - name of the symmetry 

output variables = none

return =void

What it does:
     set name to word 
how to call:
    void set_name(char * word);
****************************************************************************/

  strcpy(name,word);
}//set_name

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::set_abelian_tag(int tag){

/****************************************************************************
input variables = one
     int tag - integer that can be 0 or 1, otherwise the code terminates.

output variables = none

return = void

What it does:
    set the abelian_tag to tag

how to call:
    void set_abelian_tag(int tag);
****************************************************************************/

  if ((tag >1)||(tag<0)){
    cout<<"Error in routine symmetry::set_abelian_tag(int)\nEXITING...\n";    
  }
  abelian_tag = tag; 
}//set_abelian_tag

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

void symmetry::set_abelian_tag(){
/****************************************************************************
input variables = none

output variables = none

return =void

What it does:
     set the abelian tag of the symmetry class based on name according to a 
     priory defined rules as given below. Cases must match with the list
     defined in symmetry.h. If symmetry class name is not activated yet 
     then the code terminates. 

how to call:
    void set_abelian_tag();
****************************************************************************/

  if(!name) {
    cerr << "Allocation failure: symmetry is not active.\nError in routine symmetry::set_abelian_tag()\n EXITING...\n"; 
    exit(-1);
  }//if

  for(int i = 0; i < possible_symmetry_type_no; i++){
    if(!strcmp(possible_symmetry_type[i], name)){
         
      switch(i)
	{
	case 0: // U(1) case 
	  set_abelian_tag(1); //abelian
	  return; 

	case 1: // SU(2) case 
	  set_abelian_tag(0); //non-abelian
	  return; 

	case 2: // SU(3) case 
	  set_abelian_tag(0); //non-abelian
	  return; 

	case 3:  // Z(2) case
	  set_abelian_tag(1); //abelian
	  return;

	case 4:  // SL(2,C) case 
	  set_abelian_tag(0); //non-abelian
	  return;
	  
	case 5: // charge_SU(2) case 
	  set_abelian_tag(0); //non-abelian
	  return; 

	default:
	  cout << "There are more possible symmetry types than cases in the switch statement.\nEXITING...\n" << endl;
	  exit(-1);
	}//switch	
    }
  }//i 
  return ;


}//set

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

char * symmetry::get_name(){

/****************************************************************************
input variables = none

output variables = none

return = char * pointer

What it does:
     returns the name of the symmetry if activated, otherwise, the code 
     terminates. 

how to call:
     char * get_name();
****************************************************************************/

  if(name) return name;
  else {
    cerr << "This symmetry is not yet allocated.\nError in routine symmetry::get_name()\nEXITING...\n"; 
    exit(-1);
  }//else
}//get_name

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

int symmetry::trial_state_no(){

/****************************************************************************
input variables = none

output variables = none

return = an integer

What it does:
     returns the trial state number of the symmetry(name) if activated according
     to a priory defined rules given below. These rules must match with the list
     of symmetry names defined in symmetry.h.
     If symmetry class name is not activated yet then the code terminates. 
     
how to call:
     int trial_state_no()
****************************************************************************/

  if(!name) {
    cerr << "This symmetry is not yet allocated.\nError in routine symmetry::trial_state_no()\nEXITING...\n"; 
    exit(-1);
  }//if  
  for(int i = 0; i < possible_symmetry_type_no; i++){
    if(!strcmp(possible_symmetry_type[i], name))
      switch(i)
	{
	case 0: /* U(1) case */
	  return (max_representation_index - min_representation_index + 1);
	case 1: /* SU(2) case */
	  return (max_representation_index - min_representation_index + 1);
	case 2: /* SU(3) case */
	  cout << "The code has no trial state number counter ready for the SU(3) case.\nEXITING...\n" << endl;
	  exit(-1);
	case 3:  /* Z(2) case */
	  if(max_representation_index == min_representation_index) return 1;
	  else return 2;
	case 4:  /* SL(2,C) case */
	  cout << "The code has no trial state number counter ready for the SL(2,C) case.\nEXITING...\n" << endl;
	  exit(-1);
	case 5: /* charge_SU(2) case */
	  return (max_representation_index - min_representation_index + 1);
	default:
	  cout << "There are more possible symmetry types than cases in the switch statement.\nEXITING...\n" << endl;
	  exit(-1);
	}//switch
  }//i 
  return 0;
}//trial_state_no

/*----------------------------------------------------------------------------*/
/*-------------------------FRIEND FUNCTIONS-----------------------------------*/
/*----------------------------------------------------------------------------*/


/* selection rules: */

bool selection_rule(symmetry & symmetry_type, int representation_index_1, int representation_index_2, int representation_index_new)
{
/****************************************************************************
input variables = none

output variables = none

return = a logical that is the ???

What it does:
     returns ???? 
how to call:
     ???
****************************************************************************/

    int representation_index_old_sum;
    for(int i = 0; i < possible_symmetry_type_no; i++)
    {
	if(!strcmp(possible_symmetry_type[i], symmetry_type.get_name()))
	    switch(i)
	    {
		case 0: /* U(1) case */
		    if( ((representation_index_1 + representation_index_2) == representation_index_new))
			return true;
		    else
			return false;
		    break;
		case 1: /* SU(2) case */
		    representation_index_old_sum = representation_index_1 + representation_index_2;
		    if (((int)abs(representation_index_1 - representation_index_2) <= representation_index_new) 
			&& ( representation_index_old_sum >= representation_index_new) 
			&& !((representation_index_old_sum - representation_index_new) % 2))
			return true;
		    else
			return false;
		    break;
		case 2: /* SU(3) case */
		    cout << "The code has no selection rule ready for the SU(3) case.\nEXITING...\n" << endl;
		    exit(-1);
		    break;
		case 3:  /* Z(2) case */
		    if((representation_index_1 || representation_index_2)-
		       (representation_index_1 && representation_index_2)== representation_index_new)
			return true;
		    else 
			return false;
		    break;
		case 4:  /* SL(2,C) case */
		    cout << "The code has no selection rule ready for the SL(2,C) case.\nEXITING...\n" << endl;
		    exit(-1);
		    break;
		case 5: /* charge_SU(2) case */
		    representation_index_old_sum = representation_index_1 + representation_index_2;
		    if (((int)abs(representation_index_1 - representation_index_2) <= representation_index_new) 
			&& ( representation_index_old_sum >= representation_index_new) 
			&& !((representation_index_old_sum - representation_index_new) % 2))
			return true;
		    else
			return false;
		    break;
		default:
		    cout << "There are more possible symmetry types than cases in the switch statement.\nEXITING...\n" << endl;
		    exit(-1);
	    }
    }
    return false;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/

bool grand_selection_rule(int symmetry_no, symmetry * symmetry_type, llvector & representation_index_old_1, llvector & representation_index_old_2, llvector & representation_index_new)
{
/****************************************************************************
input variables = none

output variables = none

return = a logical that is the ???

What it does:
     returns ???? 
how to call:
     ???
****************************************************************************/
    bool ans(true);
/*    if((state_old_1.length() != state_old_2.length()) || (state_old_1.length() != state_new.length()) || (state_old_1.length() != sym_no))
    {
	cerr<<"\nThe sizes of the llvectors do not match.\nEXITING...\n" << endl;
	exit(-1);
    }
*/
    for(int i = 0; i < symmetry_no; i++)
    {
	if(!selection_rule(symmetry_type[i], representation_index_old_1[i], representation_index_old_2[i], representation_index_new[i]))
	    ans = false;
    }
    return ans;
}





