(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     86355,       2182]
NotebookOptionsPosition[     83803,       2104]
NotebookOutlinePosition[     84158,       2120]
CellTagsIndexPosition[     84115,       2117]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[{
 "Anderson model  with ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", "charge"], TraditionalForm]]],
 "(1) \[Cross] ",
 Cell[BoxData[
  FormBox[
   StyleBox[
    SubscriptBox["SU", "spin"],
    FontSlant->"Italic"], TraditionalForm]]],
 "(2) symmetries"
}], "Title",
 CellChangeTimes->{{3.4224537204910755`*^9, 3.4224537226160755`*^9}, {
  3.422944985447185*^9, 3.422944989189542*^9}}],

Cell["Initialization ", "Subtitle"],

Cell[BoxData[{
 RowBox[{"ClearAll", "[", "\"\<Global`*\>\"", "]"}], "\[IndentingNewLine]", 
 RowBox[{"Off", "[", 
  RowBox[{
   RowBox[{"General", "::", "spell"}], ",", " ", 
   RowBox[{"General", "::", "spell1"}]}], "]"}]}], "Input"],

Cell[TextData[{
 "This notebook describes how to construct the input matrices for the case \
when two symmetries are considered U(1) for the charge and U(1) for the spin, \
corresponding to the total charge with respect to half-filling and the  \
z-component of the total spin, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"z", ",", "total"}]], ".", " "}], TraditionalForm]]],
 "\nWe shall represent the spin using fermions. The total charge is computed \
as \n\t",
 Cell[BoxData[
  FormBox[
   SubscriptBox["Q", "N"], TraditionalForm]]],
 " = ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     FormBox[
      SuperscriptBox[
       SubscriptBox["\[Sum]", 
        RowBox[{"n", "=", "0"}]], "N"],
      TraditionalForm], 
     RowBox[{
      SuperscriptBox[
       SubscriptBox["f", "n"], "+"], 
      SubscriptBox["f", "n"]}]}], " ", "+", 
    RowBox[{
     SuperscriptBox["d", "+"], "d"}], " ", "-", " ", 
    RowBox[{"(", 
     RowBox[{"N", "+", "1"}], ")"}]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "n"], TraditionalForm]]],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    SuperscriptBox[
     SubscriptBox["f", "n"], "+"]}], TraditionalForm]]],
 ") is the annihilation (creation) operator for the band electrons and ",
 StyleBox["d",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", " ", 
    SuperscriptBox["d", "+"], ")"}], TraditionalForm]]],
 " is the annihilation (creation) operator for the d-levels. The local spin  \
will be represented through the fermion ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["d", "+"], TraditionalForm]]],
 " . In the 0-th iteration,  we have thus two electronic (single particle) \
levels with two spins,  i.e.,  FOUR  single particle states:  \n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox[
     SubscriptBox["f", "\[UpArrow]"], "+"], "|", 
    RowBox[{"0", ">"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox[
      SubscriptBox["f", "\[DownArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", " ", 
    RowBox[{
     SuperscriptBox[
      SubscriptBox["d", "\[UpArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", 
    FormBox[
     RowBox[{
      SuperscriptBox[
       SubscriptBox["d", "\[DownArrow]"], "+"], "|", 
      RowBox[{"0", ">", "\[IndentingNewLine]"}]}],
     TraditionalForm]}], TraditionalForm]]],
 "\nwith  |0> the vacuum state than the single-electron states are. We shall \
therefore first construct a ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["2", "4"], "=", "16", " "}], TraditionalForm]]],
 "-dimensional  Hilbert spapce, generated by the operators ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[UpArrow]"], "+"], TraditionalForm]]],
 ",  ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[UpArrow]"], "+"], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", and define fermionic creation and annihilation operators there. These \
operators shall then be used to construct symbolically the many-body states \
and operators. "
}], "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
  3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
  3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
  3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
  3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
  3.422262329515971*^9, 3.422262518108431*^9}, {3.4224537548035755`*^9, 
  3.4224537566473255`*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSingleOccupiedStates", "=", "4"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"NumberOfTotalStates", " ", "=", " ", 
   SuperscriptBox["2", "NumberOfSingleOccupiedStates"]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42225970942581*^9, 3.422259753959048*^9}, {
  3.422260356857766*^9, 3.422260359126533*^9}}],

Cell["\<\
Next we construct effectively the NumberOfTotalStates * NumberOfTotalStates \
-dimensional creation operator matrices. This part you can skip. \
\>", "Text",
 CellChangeTimes->{{3.422260232752746*^9, 3.422260239500793*^9}, {
  3.422260278460007*^9, 3.422260287823447*^9}, {3.422260459641082*^9, 
  3.422260469284508*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Occupation", " ", "=", " ", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"IntegerDigits", "[", 
      RowBox[{"i", ",", "2", ",", " ", "NumberOfSingleOccupiedStates"}], 
      "]"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"i", ",", "0", ",", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Operator", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfSingleOccupiedStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorDagger", " ", "=", "Operator"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"For", "[", 
  RowBox[{
   RowBox[{"r", "=", "1"}], ",", " ", 
   RowBox[{"r", "\[LessEqual]", " ", "NumberOfSingleOccupiedStates"}], ",", 
   " ", 
   RowBox[{"r", "++"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"For", "[", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", " ", 
      RowBox[{"i", "\[LessEqual]", "NumberOfTotalStates"}], ",", " ", 
      RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", 
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", " ", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfTotalStates"}], ",", " ", 
        RowBox[{"j", "++"}], ",", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"For", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"state", "=", "1"}], ";", " ", 
            RowBox[{"StateSign", "=", "1"}]}], ",", " ", 
           RowBox[{"state", "<", "r"}], ",", " ", 
           RowBox[{"state", "++"}], ",", "\[IndentingNewLine]", 
           RowBox[{"StateSign", "*=", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"-", "1"}], ")"}], 
             RowBox[{"Occupation", "\[LeftDoubleBracket]", 
              RowBox[{"j", ",", " ", "state"}], 
              "\[RightDoubleBracket]"}]]}]}], "]"}], ";", 
         "\[IndentingNewLine]", "\[IndentingNewLine]", 
         RowBox[{"If", "[", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
                RowBox[{"BitXor", "[", 
                 RowBox[{
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "i", 
                   "\[RightDoubleBracket]"}], ",", " ", 
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "j", 
                   "\[RightDoubleBracket]"}]}], "]"}], ",", "2"}], "]"}], " ",
               "\[Equal]", " ", 
              SuperscriptBox["2", 
               RowBox[{"NumberOfSingleOccupiedStates", "-", "r"}]]}], ")"}], 
            "\[IndentingNewLine]", "&&", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "i", 
                "\[RightDoubleBracket]"}], "]"}], ">", 
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "j", 
                "\[RightDoubleBracket]"}], "]"}]}], ")"}]}], ",", 
           "\[IndentingNewLine]", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"OperatorDagger", "\[LeftDoubleBracket]", 
             RowBox[{"r", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
            "=", "StateSign"}]}], "\[IndentingNewLine]", 
          "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", "]"}]}], 
     "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
     "Operator", "\[LeftDoubleBracket]", "r", "\[RightDoubleBracket]"}], " ", 
     "=", 
     RowBox[{"Transpose", "[", 
      RowBox[{
      "OperatorDagger", "\[LeftDoubleBracket]", "r", 
       "\[RightDoubleBracket]"}], "]"}]}]}]}], "\[IndentingNewLine]", 
  "]"}]}], "Input",
 CellChangeTimes->{{3.42226025055771*^9, 3.422260260564144*^9}}],

Cell["\<\
We have now     NumberOfSingleOccupiedStates = 4 creation operators:  \
OperatorDagger\[LeftDoubleBracket]1\[RightDoubleBracket],..., OperatorDagger\
\[LeftDoubleBracket]4\[RightDoubleBracket]. These shall be used to construct \
the necessary operators: \
\>", "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
   3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
   3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
   3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
   3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
   3.422260330982351*^9, 3.422260334508321*^9}, {3.422260366221652*^9, 
   3.422260446806384*^9}, 3.422260485771486*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"fDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Sz", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"dDaggerUp", ".", "dUp"}], "-", 
      RowBox[{"dDaggerDown", ".", "dDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlus", " ", "=", 
   RowBox[{"dDaggerUp", ".", "dDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinus", " ", "=", 
   RowBox[{"dDaggerDown", ".", "dUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"UnitOperator", " ", "=", " ", 
   RowBox[{
    RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
    RowBox[{"dUp", ".", "dDaggerUp"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.422273235977622*^9, 3.422273314174119*^9}}],

Cell["Now we can construct all the states. ", "Text",
 CellChangeTimes->{{3.422260563311906*^9, 3.422260571250809*^9}}],

Cell["States", "Subtitle"],

Cell["\<\
In this section we construct the corresponding states and order them as in \
the representation index table. First we construct the VacuumState and then \
based on it, we construct all the other states that are allowed. We can \
construct based on these states triplet as well as singlet states.\
\>", "Text",
 CellChangeTimes->{{3.422260600765262*^9, 3.422260689774695*^9}, {
  3.4224683940067*^9, 3.4224683979442*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"VacuumState", " ", "=", "  ", 
   RowBox[{"Table", "[", 
    RowBox[{"IntegerDigits", "[", 
     RowBox[{
      SuperscriptBox["2", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]], ",", "2", ",", " ", 
      "NumberOfTotalStates"}], "]"}], "]"}]}], ";"}]], "Input"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfBlockStates", " ", "=", " ", "10"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BlockStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State1", "=", "VacuumState"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State2", " ", "=", 
    RowBox[{"dDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State3", " ", "=", 
    RowBox[{"fDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State4", " ", "=", 
    RowBox[{"dDaggerUp", ".", "dDaggerDown", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "5", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State5", " ", "=", 
    RowBox[{"fDaggerUp", ".", "fDaggerDown", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "6", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State6", " ", "=", 
    RowBox[{
     FractionBox["1", 
      SqrtBox["2"]], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"fDaggerDown", ".", "dDaggerUp"}], "-", 
        RowBox[{"fDaggerUp", ".", "dDaggerDown"}]}], ")"}], ".", 
      "VacuumState"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "7", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State7", " ", "=", 
    RowBox[{"fDaggerUp", ".", "dDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "8", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State8", " ", "=", " ", 
    RowBox[{
    "fDaggerUp", ".", "fDaggerDown", ".", "dDaggerUp", ".", 
     "VacuumState"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "9", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State9", " ", "=", 
    RowBox[{
    "fDaggerUp", ".", " ", "dDaggerUp", ".", "dDaggerDown", ".", 
     "VacuumState"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "10", "\[RightDoubleBracket]"}], 
   "=", 
   RowBox[{"State10", " ", "=", 
    RowBox[{
    "fDaggerUp", ".", " ", "fDaggerDown", ".", "dDaggerUp", ".", 
     "dDaggerDown", ".", "VacuumState"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
  3.422261115279298*^9, 3.422261172794231*^9}, {3.4224537966942*^9, 
  3.42245384672545*^9}, {3.4224539433660755`*^9, 3.4224540117410755`*^9}, {
  3.42245700397545*^9, 3.4224570229442*^9}, {3.4224570532567*^9, 
  3.4224572888035755`*^9}, {3.4224574096942*^9, 3.42245740978795*^9}, {
  3.42245954647545*^9, 3.4224595508348255`*^9}, {3.42269299930489*^9, 
  3.422693017643247*^9}, {3.422693068882953*^9, 3.422693088872416*^9}, {
  3.4229450337418*^9, 3.422945145976007*^9}, {3.422945202938709*^9, 
  3.422945210050571*^9}}],

Cell["\<\
Now we have to check that these states are indeed ordered.  If not, we have \
to reshuffle them.\
\>", "Text",
 CellChangeTimes->{{3.422260756543946*^9, 3.422260774656064*^9}, {
  3.422261998378347*^9, 3.42226200768368*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSymmetries", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SzTotal", " ", "=", " ", 
   RowBox[{"Sz", "+", 
    RowBox[{
     FractionBox["1", "2"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
       RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"QTotal", " ", "=", " ", 
    RowBox[{
     RowBox[{"fDaggerUp", ".", "fUp"}], "+", 
     RowBox[{"fDaggerDown", ".", "fDown"}], " ", "+", " ", 
     RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
     RowBox[{"dDaggerDown", ".", "dDown"}], "-", " ", 
     RowBox[{"2", " ", "UnitOperator"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "QTotal", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "SzTotal", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
  3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
  3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
  3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
  3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
  3.422273329630683*^9, 3.422273330834617*^9}, {3.422424163530083*^9, 
  3.422424182641464*^9}, {3.422424283395106*^9, 3.422424295014417*^9}, {
  3.422424504284488*^9, 3.422424529308189*^9}, {3.422425828934438*^9, 
  3.422425948432719*^9}, {3.422426022264904*^9, 3.422426031542518*^9}, {
  3.422426385786546*^9, 3.422426386364774*^9}, {3.4224583187723255`*^9, 
  3.4224583473192*^9}}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "2"}], "0"},
     {
      RowBox[{"-", "1"}], 
      FractionBox["1", "2"]},
     {
      RowBox[{"-", "1"}], 
      FractionBox["1", "2"]},
     {"0", "0"},
     {"0", "0"},
     {"0", "0"},
     {"0", "1"},
     {"1", 
      FractionBox["1", "2"]},
     {"1", 
      FractionBox["1", "2"]},
     {"2", "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{{3.42226107342192*^9, 3.422261105951935*^9}, 
   3.422261189857225*^9, {3.422261361886205*^9, 3.422261382372428*^9}, {
   3.422261557654303*^9, 3.422261582070163*^9}, 3.422261670042143*^9, 
   3.4222618584621*^9, {3.422261936692001*^9, 3.422261947147733*^9}, 
   3.422261982631925*^9, 3.422273333398469*^9, 3.422422421961272*^9, 
   3.422424185206617*^9, {3.422424284343575*^9, 3.422424308315455*^9}, {
   3.422424505813002*^9, 3.42242452992562*^9}, 3.422425952720056*^9, {
   3.422426001119572*^9, 3.422426038231274*^9}, 3.422426387262979*^9, 
   3.422427033012301*^9, 3.422427279541733*^9, 3.42243318197125*^9, 
   3.42245730310045*^9, 3.4224573352567*^9, 3.4224574130692*^9, 
   3.4224583480223255`*^9, 3.4224592651785755`*^9, 3.4224595608192*^9, 
   3.4224602309285755`*^9, 3.4224684033973255`*^9, 3.4224685338348255`*^9, 
   3.422519136118478*^9, 3.422676797870682*^9, 3.422692709940655*^9, 
   3.42269301989222*^9, 3.422693090950708*^9, 3.422942217076612*^9, 
   3.422945159637923*^9, 3.422945212912612*^9, 3.422945632174616*^9}]
}, Open  ]],

Cell["\<\
Now we can create the matrices for any of the spectral operators as well as \
the hopping operators.\
\>", "Text"],

Cell["Spectral and hopping operators.", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrderedfDaggerUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedfDaggerDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerUp", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerDown", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ReducedfDagger", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               FractionBox["1", "2"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               RowBox[{"-", 
                FractionBox["1", "2"]}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"ReducedfDagger", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427108128412*^9, 3.422427111736916*^9}, {
   3.422427153114619*^9, 3.422427212919719*^9}, 3.422427535099241*^9, {
   3.422433150378739*^9, 3.422433152654368*^9}, 3.422945248939676*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {-1.4142135623730951}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, \
{0.}}, {{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, 
  {0.}}, {{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, \
{0.}, {0.}, {0.}, {0.}, {0.7071067811865476}, 
  {-1.224744871391589}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {1.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-1.4142135623730951}, \
{0.}}}\
\>", "\<\
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

1.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         -1.41421   0.         0.         0.         0.         \
0.         0.         0.

0.         1.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         1.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.707107   -1.22474   \
0.         0.         0.

0.         0.         0.         1.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
0.         -1.41421   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422427009374603*^9, 3.422427041009018*^9, 3.422427112761862*^9, {
   3.422427155075844*^9, 3.422427213797051*^9}, 3.422427281483832*^9, 
   3.422433185296861*^9, 3.4224583676942*^9, 3.4224592675692*^9, 
   3.4224595650535755`*^9, 3.42246023291295*^9, 3.4224684071160755`*^9, 
   3.4224685357723255`*^9, 3.422519137973049*^9, 3.422676802855329*^9, 
   3.422692728287292*^9, 3.422693023593861*^9, 3.422693092447997*^9, 
   3.422942223455276*^9, {3.4229452223898*^9, 3.422945250650416*^9}, 
   3.422945636036459*^9}]
}, Open  ]],

Cell["Hamiltonian terms", "Subtitle"],

Cell[TextData[{
 "In this section we consider different terms appearing in the Hamiltonian \
for the Anderson model. The expression used for it : ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["H", "Anderson"], "=", 
    RowBox[{
     RowBox[{"E", 
      RowBox[{"\[Sum]", " ", 
       RowBox[{
        SuperscriptBox[
         SubscriptBox["d", "s"], "+"], " ", 
        SubscriptBox["d", "s"]}]}]}], "  ", "+", " ", 
     RowBox[{"U", " ", 
      SuperscriptBox[
       SubscriptBox["d", "\[UpArrow]"], "+"], " ", 
      SubscriptBox["d", "\[UpArrow]"], "  ", 
      SuperscriptBox[
       SubscriptBox["d", "\[DownArrow]"], "+"], " ", 
      SubscriptBox["d", "\[DownArrow]"]}], "  ", "+", "    ", 
     RowBox[{"t", 
      RowBox[{"\[Sum]", " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{
          SuperscriptBox[
           SubscriptBox["f", "s"], "+"], " ", 
          SubscriptBox["d", "s"]}], " ", "+", " ", 
         RowBox[{"h", ".", "c", "."}]}], ")"}]}]}]}]}], TraditionalForm]]]
}], "Text",
 CellChangeTimes->{{3.42245857303795*^9, 3.42245861603795*^9}, {
  3.4224586505067*^9, 3.4224586505067*^9}, {3.42245869453795*^9, 
  3.4224587475067*^9}, {3.4224588273348255`*^9, 3.4224589099285755`*^9}, {
  3.42246037591295*^9, 3.4224604233660755`*^9}, {3.4224684539598255`*^9, 
  3.4224684545067*^9}, {3.422942238606759*^9, 3.422942238607566*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfHamiltonianTerms", "=", "3"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HAnderson", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "HAnderson", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=", 
   RowBox[{
    RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
    RowBox[{"dDaggerDown", ".", "dDown"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "HAnderson", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=", 
   RowBox[{"dDaggerUp", ".", "dUp", ".", "dDaggerDown", ".", "dDown"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
    "HAnderson", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], "=", 
    
    RowBox[{
     RowBox[{"fDaggerUp", ".", "dUp"}], "+", 
     RowBox[{"dDaggerUp", ".", "fUp"}], "+", 
     RowBox[{"fDaggerDown", ".", "dDown"}], "+", 
     RowBox[{"dDaggerDown", ".", "fDown"}]}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedHamiltonian", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
         RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], " ", 
        "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", 
         RowBox[{
         "HAnderson", "\[LeftDoubleBracket]", "k", "\[RightDoubleBracket]"}], 
         ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"k", ",", "NumberOfHamiltonianTerms"}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427388370775*^9, 3.422427437070901*^9}, {
  3.42242749172612*^9, 3.422427526825983*^9}, {3.422427584982783*^9, 
  3.422427590687387*^9}, {3.422433136386855*^9, 3.422433139040832*^9}, {
  3.4224589290692*^9, 3.42245916878795*^9}, {3.4224593123817*^9, 
  3.4224593220692*^9}, {3.4224593526785755`*^9, 3.4224593903973255`*^9}, {
  3.4224600360692*^9, 3.4224601996317*^9}, {3.42246027835045*^9, 
  3.42246027891295*^9}, {3.42246032266295*^9, 3.4224603351473255`*^9}, {
  3.42246043103795*^9, 3.42246046197545*^9}, {3.422692743832292*^9, 
  3.422692749055595*^9}}],

Cell[OutputFormData["\<\
{{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {1.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {2.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {1.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {1.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {1.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {2.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {2.}}}, 
 {{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {1.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {1.}}}, 
 {{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {-1.4142135623730951}, {0.}, {0.}, {0.}, 
   {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {-1.4142135623730951}, {0.}, {0.}, \
{0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {-1.4142135623730951}, {-1.4142135623730951}, {0.}, \
{0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-1.}, {0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-1.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}}}\
\>", "\<\
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         1.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         2.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         1.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         1.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
1.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         2.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         2.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         1.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         1.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         1.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         1.         0.         0.         0.         0.         \
0.         0.         0.
0.         1.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         -1.41421   0.         \
0.         0.         0.
0.         0.         0.         0.         0.         -1.41421   0.         \
0.         0.         0.
0.         0.         0.         -1.41421   -1.41421   0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         -1.        0.
0.         0.         0.         0.         0.         0.         0.         \
-1.        0.         0.
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.\
\>"], "Output",
 CellChangeTimes->{
  3.4224601098192*^9, 3.4224601787723255`*^9, 3.4224602432098255`*^9, 
   3.4224602825067*^9, {3.4224603240692*^9, 3.4224603382410755`*^9}, 
   3.4224604646473255`*^9, 3.4224684186473255`*^9, 3.4224685387098255`*^9, 
   3.422519140865287*^9, 3.42267681005077*^9, {3.422692731318904*^9, 
   3.422692752363603*^9}, 3.422693153592689*^9, 3.422942252463408*^9, 
   3.422945264320159*^9, 3.422945638013961*^9}]
}, Open  ]],

Cell[TextData[{
 StyleBox["d",
  FontSlant->"Italic"],
 " level operator"
}], "Subtitle",
 CellChangeTimes->{{3.422257543929043*^9, 3.422257549617286*^9}, {
  3.422693209853698*^9, 3.422693211767615*^9}}],

Cell[TextData[{
 "In this section we consider the spectral operator for the ",
 StyleBox["d",
  FontSlant->"Italic"],
 " level"
}], "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.4224684857567*^9, 3.4224685036317*^9}, {3.422676818626302*^9, 
  3.422676826511941*^9}, {3.422676881332044*^9, 3.422676885867984*^9}, {
  3.422945319762778*^9, 3.422945321407039*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrdereddDaggerUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrdereddDaggerDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrdereddDaggerUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "dDaggerUp", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrdereddDaggerDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "dDaggerDown", ".", 
        RowBox[{
        "BlockStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ReduceddDagger", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrdereddDaggerUp", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReduceddDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrdereddDaggerUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               FractionBox["1", "2"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrdereddDaggerDown", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReduceddDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrdereddDaggerDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               RowBox[{"-", 
                FractionBox["1", "2"]}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"ReduceddDagger", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422257650153471*^9, 3.422257658110106*^9}, {
  3.422262270013571*^9, 3.422262273937019*^9}, {3.422427554252507*^9, 
  3.422427571961936*^9}, {3.42242760434566*^9, 3.422427621897156*^9}, {
  3.422433129489568*^9, 3.422433132897229*^9}, {3.422676835079676*^9, 
  3.422676918504581*^9}, {3.422945283092261*^9, 3.422945330386702*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{1.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, \
{-1.4142135623730951}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, 
  {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, \
{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {-1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.7071067811865476}, {1.224744871391589}, \
{0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-1.4142135623730951}, {0.}, \
{0.}}}\
\>", "\<\
0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

1.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         -1.41421   0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         1.         0.         0.         0.         0.         \
0.         0.         0.

0.         0.         -1.        0.         0.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         1.         0.         0.         \
0.         0.         0.

0.         0.         0.         0.         0.         0.707107   1.22474    \
0.         0.         0.

0.         0.         0.         0.         0.         0.         0.         \
-1.41421   0.         0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257525369031*^9, 3.422257659918318*^9, 3.422262275051522*^9, 
   3.422422433839974*^9, 3.422427623357605*^9, 3.422433194585268*^9, 
   3.4224594161473255`*^9, 3.4224595713660755`*^9, 3.4224604924910755`*^9, 
   3.42246854072545*^9, 3.42251914642439*^9, 3.422676919679311*^9, 
   3.42269284068618*^9, 3.422693223552578*^9, 3.422942263503772*^9, 
   3.422944370658689*^9, {3.422945323680138*^9, 3.422945331817388*^9}, 
   3.422945639597252*^9}]
}, Open  ]],

Cell[TextData[{
 StyleBox["d",
  FontSlant->"Italic"],
 " level occupation numbers"
}], "Subtitle",
 CellChangeTimes->{{3.422257543929043*^9, 3.422257549617286*^9}, {
  3.422693209853698*^9, 3.422693211767615*^9}, {3.422943448369536*^9, 
  3.422943457491814*^9}}],

Cell[TextData[{
 "In this section we consider the static operator for the ",
 StyleBox["d",
  FontSlant->"Italic"],
 " level occupation number"
}], "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.4224684857567*^9, 3.4224685036317*^9}, {3.422676818626302*^9, 
  3.422676826511941*^9}, {3.422676881332044*^9, 3.422676885867984*^9}, {
  3.422945499155751*^9, 3.422945512804971*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderednQD", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderednQD", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
           RowBox[{"dDaggerDown", ".", "dDown"}]}], ")"}], ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422257650153471*^9, 3.422257658110106*^9}, {
  3.422262270013571*^9, 3.422262273937019*^9}, {3.422427554252507*^9, 
  3.422427571961936*^9}, {3.42242760434566*^9, 3.422427621897156*^9}, {
  3.422433129489568*^9, 3.422433132897229*^9}, {3.422676835079676*^9, 
  3.422676918504581*^9}, {3.422943461077296*^9, 3.42294348314143*^9}, {
  3.422943518655405*^9, 3.422943528093079*^9}, {3.422945363158805*^9, 
  3.422945394769731*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {1.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {2.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {1.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {1.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {1.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {2.}, {0.}}, {{0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {2.}}}\
\>", "\<\
0.   0.   0.   0.   0.   0.   0.   0.   0.   0.

0.   1.   0.   0.   0.   0.   0.   0.   0.   0.

0.   0.   0.   0.   0.   0.   0.   0.   0.   0.

0.   0.   0.   2.   0.   0.   0.   0.   0.   0.

0.   0.   0.   0.   0.   0.   0.   0.   0.   0.

0.   0.   0.   0.   0.   1.   0.   0.   0.   0.

0.   0.   0.   0.   0.   0.   1.   0.   0.   0.

0.   0.   0.   0.   0.   0.   0.   1.   0.   0.

0.   0.   0.   0.   0.   0.   0.   0.   2.   0.

0.   0.   0.   0.   0.   0.   0.   0.   0.   2.\
\>"], "Output",
 CellChangeTimes->{
  3.422257525369031*^9, 3.422257659918318*^9, 3.422262275051522*^9, 
   3.422422433839974*^9, 3.422427623357605*^9, 3.422433194585268*^9, 
   3.4224594161473255`*^9, 3.4224595713660755`*^9, 3.4224604924910755`*^9, 
   3.42246854072545*^9, 3.42251914642439*^9, 3.422676919679311*^9, 
   3.42269284068618*^9, 3.422693223552578*^9, 3.422942263503772*^9, 
   3.422943530523324*^9, {3.422945384648363*^9, 3.4229453958803368`*^9}, 
   3.422945515570084*^9, 3.422945641206166*^9}]
}, Open  ]],

Cell["", "Text",
 CellChangeTimes->{3.42227230000648*^9}],

Cell["Local operators and local basis", "Subtitle",
 CellChangeTimes->{{3.422272304428917*^9, 3.422272311966455*^9}}],

Cell[TextData[{
 "Now let us continue with the construction of the local states. We shall \
represent the operators ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "N"], TraditionalForm]]],
 "by the same operators as before. Then we have four local states: "
}], "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfLocalStates", " ", "=", " ", "3"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"LocalStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=",
    " ", "VacuumState"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
    "LocalStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], 
    "=", 
    RowBox[{"fDaggerUp", ".", "VacuumState"}]}], ";"}], 
  " "}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"fDaggerUp", ".", "fDaggerDown", ".", "VacuumState"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
  3.422261115279298*^9, 3.422261172794231*^9}, {3.422272469310303*^9, 
  3.422272491843473*^9}, {3.422272529795952*^9, 3.422272615717838*^9}, {
  3.422272682500722*^9, 3.422272698956333*^9}, {3.42227335639425*^9, 
  3.422273365754123*^9}, {3.422273443687094*^9, 3.42227344684052*^9}, {
  3.422945404958965*^9, 3.422945412959639*^9}}],

Cell["We check if they are ordered properly:", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SzLocal", " ", "=", " ", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
      RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"QLocal", " ", "=", " ", 
    RowBox[{
     RowBox[{"fDaggerUp", ".", "fUp"}], "+", 
     RowBox[{"fDaggerDown", ".", "fDown"}], " ", "-", " ", "UnitOperator"}]}],
    ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "QLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "SzLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
   3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
   3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
   3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
   3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
   3.422272998493306*^9, 3.422273023589198*^9}, {3.422273060650893*^9, 
   3.42227306270624*^9}, {3.422273102646266*^9, 3.422273104921433*^9}, {
   3.422273156211252*^9, 3.422273179056414*^9}, {3.42227334592718*^9, 
   3.422273346665373*^9}, 3.422427646446775*^9, {3.422427701122459*^9, 
   3.422427780430807*^9}, {3.42242781943902*^9, 3.422427835662201*^9}}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "1"}], "0"},
     {"0", 
      FractionBox["1", "2"]},
     {"1", "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.422273025881706*^9, {3.422273077127199*^9, 3.422273105554888*^9}, 
   3.422273179924059*^9, {3.422273347766973*^9, 3.422273369596677*^9}, 
   3.422273452148074*^9, 3.42242243891415*^9, {3.422427783820186*^9, 
   3.422427793470396*^9}, 3.422427836351067*^9, 3.422433199232154*^9, 
   3.4224595916785755`*^9, 3.4224604974442*^9, 3.4224608112098255`*^9, {
   3.4224685188817*^9, 3.4224685450848255`*^9}, 3.422519151012997*^9, 
   3.422676929669047*^9, 3.422692867981985*^9, 3.422693285381231*^9, 
   3.422942269793095*^9, 3.422945416083504*^9, 3.422945645211685*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"OrderedLocalfDaggerUp", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalfDaggerDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerUp", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
        RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
       RowBox[{
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}],
         ".", "fDaggerDown", ".", 
        RowBox[{
        "LocalStates", "\[LeftDoubleBracket]", "j", 
         "\[RightDoubleBracket]"}]}]}]}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"ReducedLocalfDagger", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedLocalfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               FractionBox["1", "2"]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "\[NotEqual]", 
          "0"}], ",", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{"ReducedLocalfDagger", "\[LeftDoubleBracket]", 
           RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", " ", 
          
          RowBox[{
           RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
            RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "/", 
           RowBox[{"ClebschGordan", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "j"}], "\[RightDoubleBracket]"}]}], "}"}], 
             ",", 
             RowBox[{"{", 
              RowBox[{
               FractionBox["1", "2"], ",", 
               RowBox[{"-", 
                FractionBox["1", "2"]}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], ",", 
               RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
                RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}],
             "]"}]}]}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"ReducedLocalfDagger", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], "}"}]}], ",", 
      " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422262166469301*^9, 3.422262185711958*^9}, {
   3.422262232176988*^9, 3.422262232892694*^9}, {3.42227287163859*^9, 
   3.422272880981803*^9}, {3.4224229432499*^9, 3.422422953983789*^9}, {
   3.422427878701063*^9, 3.422427932721799*^9}, {3.422433118922209*^9, 
   3.42243312210934*^9}, 3.422945435464439*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}}, {{1.}, {0.}, {0.}}, {{0.}, {-1.4142135623730951}, {0.}}}\
\
\>", "\<\
0.         0.         0.

1.         0.         0.

0.         -1.41421   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, 
   3.422272883289382*^9, 3.422273385171968*^9, 3.422273462344101*^9, 
   3.422422442190021*^9, {3.422422946203317*^9, 3.422422954776595*^9}, 
   3.422427934223118*^9, 3.422433201722852*^9, 3.4224595941942*^9, 
   3.4224604988348255`*^9, 3.4224608128192*^9, {3.4224685200692*^9, 
   3.4224685463660755`*^9}, 3.422519153690887*^9, 3.422676932308325*^9, 
   3.422692868989181*^9, 3.422693286351669*^9, 3.422942271740467*^9, 
   3.422945436826891*^9, 3.422945646283938*^9}]
}, Open  ]],

Cell["\<\
We write the quantum numbers to file \"QuantumNumbers\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./QuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422428031091177*^9, 3.422428228324012*^9}, 
   3.422428274771525*^9, {3.422428334035563*^9, 3.422428414302918*^9}, {
   3.422433090180519*^9, 3.422433103257273*^9}}],

Cell["\<\
Next, the hopping, spectral and Hamiltonian operators are saved in \
\"HoppingOperators\", \"SpectralOperators\", and \"HamiltonianBlocks\", \
respectively:\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428580988006*^9}, {3.422428618117171*^9, 3.422428703986899*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<fDaggerUp\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"ReducedfDagger", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422945539137039*^9, 3.422945561796952*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./SpectralOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<dDagger\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"ReduceddDagger", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422676942669582*^9, 3.422676974471557*^9}, {
  3.422945454061184*^9, 3.422945471666721*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./StaticOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<nQD\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderednQD", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422943369781743*^9, 3.422943408838973*^9}, {
   3.422945477129712*^9, 3.422945484562461*^9}, 3.422945578497573*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HamiltonianBlocks\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"For", "[", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", 
    RowBox[{"k", "\[LessEqual]", " ", "NumberOfHamiltonianTerms"}], ",", " ", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Write", "[", 
      RowBox[{
      "strm", ",", "\"\<HamiltonianBlock \>\"", ",", "k", ",", 
       "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", 
       RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"For", "[", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", 
          RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", 
             RowBox[{"N", "[", 
              RowBox[{"CForm", "[", 
               RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
                RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
               " ", "]"}], "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
             RowBox[{"WriteString", "[", 
              RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"WriteString", "[", 
         RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"i", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Write", "[", 
      RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]",
      
     RowBox[{"k", "++"}]}]}], "]"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422422461984407*^9, 3.422422590368324*^9}, {
  3.422422657679912*^9, 3.422422675770448*^9}, {3.422422769910713*^9, 
  3.422422774310689*^9}, {3.422422907355884*^9, 3.422422907479271*^9}, {
  3.422423014025691*^9, 3.422423024399214*^9}, {3.422423065573881*^9, 
  3.422423075136177*^9}, {3.422428735589752*^9, 3.422428797092516*^9}, {
  3.42242951569936*^9, 3.422429705585544*^9}, {3.422429793487009*^9, 
  3.4224298166685*^9}, {3.422429927802331*^9, 3.422429930748695*^9}, {
  3.422432038815338*^9, 3.422432247503146*^9}, {3.422432279060799*^9, 
  3.422432423000301*^9}, {3.422432470625659*^9, 3.422432486448996*^9}, {
  3.422513762990068*^9, 3.422513804257083*^9}}],

Cell["\<\
Finally, we write the local quantum numbers to and local operators to file \
\"LocalQuantumNumbers\" and \"LocalHoppingOperators\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}, {3.422432507395505*^9, 3.422432527004284*^9}, {
  3.422432571670566*^9, 3.422432576538502*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalQuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfLocalStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Close", "[", "strm", "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalHoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<LocalfDagger\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"ReducedLocalfDagger", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfLocalStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422432891090526*^9, 3.422433072089265*^9}, 
   3.42243322189554*^9, {3.422945590061502*^9, 3.42294560853277*^9}}]
},
WindowSize->{1093, 645},
WindowMargins->{{68, Automatic}, {Automatic, 32}},
Magnification->1,
FrontEndVersion->"6.0 for Linux x86 (32-bit) (April 20, 2007)",
StyleDefinitions->"Classroom.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[568, 21, 404, 14, 59, "Title"],
Cell[975, 37, 35, 0, 40, "Subtitle"],
Cell[1013, 39, 234, 5, 63, "Input"],
Cell[1250, 46, 3823, 118, 244, "Text"],
Cell[5076, 166, 361, 8, 64, "Input"],
Cell[5440, 176, 331, 6, 24, "Text"],
Cell[5774, 184, 4167, 102, 429, "Input"],
Cell[9944, 288, 758, 12, 24, "Text"],
Cell[10705, 302, 2032, 61, 252, "Input"],
Cell[12740, 365, 119, 1, 24, "Text"],
Cell[12862, 368, 26, 0, 40, "Subtitle"],
Cell[12891, 370, 430, 7, 41, "Text"],
Cell[13324, 379, 303, 8, 69, "Input"],
Cell[13630, 389, 3674, 107, 277, "Input"],
Cell[17307, 498, 234, 5, 24, "Text"],
Cell[CellGroupData[{
Cell[17566, 507, 3220, 75, 253, "Input"],
Cell[20789, 584, 1940, 47, 249, "Output"]
}, Open  ]],
Cell[22744, 634, 124, 3, 24, "Text"],
Cell[22871, 639, 51, 0, 40, "Subtitle"],
Cell[CellGroupData[{
Cell[22947, 643, 5934, 138, 443, "Input"],
Cell[28884, 783, 2350, 52, 226, "Output"]
}, Open  ]],
Cell[31249, 838, 37, 0, 40, "Subtitle"],
Cell[31289, 840, 1379, 35, 26, "Text"],
Cell[CellGroupData[{
Cell[32693, 879, 3250, 82, 234, "Input"],
Cell[35946, 963, 5613, 103, 341, "Output"]
}, Open  ]],
Cell[41574, 1069, 204, 6, 40, "Subtitle"],
Cell[41781, 1077, 394, 9, 24, "Text"],
Cell[CellGroupData[{
Cell[42200, 1090, 6077, 140, 443, "Input"],
Cell[48280, 1232, 2281, 51, 226, "Output"]
}, Open  ]],
Cell[50576, 1286, 263, 7, 40, "Subtitle"],
Cell[50842, 1295, 410, 9, 24, "Text"],
Cell[CellGroupData[{
Cell[51277, 1308, 1662, 40, 115, "Input"],
Cell[52942, 1350, 1674, 39, 226, "Output"]
}, Open  ]],
Cell[54631, 1392, 57, 1, 24, "Text"],
Cell[54691, 1395, 117, 1, 40, "Subtitle"],
Cell[54811, 1398, 387, 9, 24, "Text"],
Cell[55201, 1409, 1352, 36, 132, "Input"],
Cell[56556, 1447, 263, 3, 24, "Text"],
Cell[CellGroupData[{
Cell[56844, 1454, 2982, 68, 236, "Input"],
Cell[59829, 1524, 1235, 29, 107, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[61101, 1558, 6154, 140, 497, "Input"],
Cell[67258, 1700, 768, 18, 103, "Output"]
}, Open  ]],
Cell[68041, 1721, 432, 8, 24, "Text"],
Cell[68476, 1731, 1484, 34, 149, "Input"],
Cell[69963, 1767, 533, 9, 24, "Text"],
Cell[70499, 1778, 1994, 50, 268, "Input"],
Cell[72496, 1830, 2065, 51, 285, "Input"],
Cell[74564, 1883, 2055, 51, 268, "Input"],
Cell[76622, 1936, 3047, 67, 302, "Input"],
Cell[79672, 2005, 602, 10, 24, "Text"],
Cell[80277, 2017, 3522, 85, 455, "Input"]
}
]
*)

(* End of internal cache information *)
