(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 6.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[     66764,       1784]
NotebookOptionsPosition[     64232,       1705]
NotebookOutlinePosition[     64585,       1721]
CellTagsIndexPosition[     64542,       1718]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[{
 "Kondo model  with ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", "charge"], TraditionalForm]]],
 "(1) \[Cross] ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["U", "spin"], TraditionalForm]]],
 "(1) symmetries"
}], "Title"],

Cell["Initialization ", "Subtitle"],

Cell[BoxData[{
 RowBox[{"ClearAll", "[", "\"\<Global`*\>\"", "]"}], "\[IndentingNewLine]", 
 RowBox[{"Off", "[", 
  RowBox[{
   RowBox[{"General", "::", "spell"}], ",", " ", 
   RowBox[{"General", "::", "spell1"}]}], "]"}]}], "Input"],

Cell[TextData[{
 "This notebook describes how to construct the input matrices for the case \
when two symmetries are considered U(1) for the charge and U(1) for the spin, \
corresponding to the total charge with respect to half-filling and the  \
z-component of the total spin, ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["S", 
     RowBox[{"z", ",", "total"}]], ".", " "}], TraditionalForm]]],
 "\nWe shall represent the spin using fermions. The total charge is computed \
as \n\t",
 Cell[BoxData[
  FormBox[
   SubscriptBox["Q", "N"], TraditionalForm]]],
 " = ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     FormBox[
      SuperscriptBox[
       SubscriptBox["\[Sum]", 
        RowBox[{"n", "=", "0"}]], "N"],
      TraditionalForm], 
     RowBox[{
      SuperscriptBox[
       SubscriptBox["f", "n"], "+"], 
      SubscriptBox["f", "n"]}]}], " ", "+", 
    RowBox[{
     SuperscriptBox["d", "+"], "d"}], " ", "-", " ", 
    RowBox[{"(", 
     RowBox[{"N", "+", "1"}], ")"}]}], TraditionalForm]]],
 "\nwhere ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "n"], TraditionalForm]]],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", 
    SuperscriptBox[
     SubscriptBox["f", "n"], "+"]}], TraditionalForm]]],
 ") is the annihilation (creation) operator for the band electrons and ",
 StyleBox["d",
  FontSlant->"Italic"],
 " ",
 Cell[BoxData[
  FormBox[
   RowBox[{"(", " ", 
    SuperscriptBox["d", "+"], ")"}], TraditionalForm]]],
 " is the annihilation (creation) operator for the d-levels. The local spin  \
will be represented through the fermion ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["d", "+"], TraditionalForm]]],
 " . In the 0-th iteration,  we have thus two electronic (single particle) \
levels with two spins,  i.e.,  FOUR  single particle states:  \n\t",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox[
     SubscriptBox["f", "\[UpArrow]"], "+"], "|", 
    RowBox[{"0", ">"}]}], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    RowBox[{
     SuperscriptBox[
      SubscriptBox["f", "\[DownArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", " ", 
    RowBox[{
     SuperscriptBox[
      SubscriptBox["d", "\[UpArrow]"], "+"], "|", 
     RowBox[{"0", ">"}]}], ",", 
    FormBox[
     RowBox[{
      SuperscriptBox[
       SubscriptBox["d", "\[DownArrow]"], "+"], "|", 
      RowBox[{"0", ">", "\[IndentingNewLine]"}]}],
     TraditionalForm]}], TraditionalForm]]],
 "\nwith  |0> the vacuum state than the single-electron states are. We shall \
therefore first construct a ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SuperscriptBox["2", "4"], "=", "16", " "}], TraditionalForm]]],
 "-dimensional  Hilbert spapce, generated by the operators ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[UpArrow]"], "+"], TraditionalForm]]],
 ",  ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["f", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[UpArrow]"], "+"], TraditionalForm]]],
 ", and ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox[
    SubscriptBox["d", "\[DownArrow]"], "+"], TraditionalForm]]],
 ", and define fermionic creation and annihilation operators there. These \
operators shall then be used to construct symbolically the many-body states \
and operators. \nIn the Kondo model, the constraint is that the d-level is \
allways  singly-occupied, so after constructing the states we will keep only \
states with occupation one for the d-level. "
}], "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
  3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
  3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
  3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
  3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
  3.422262329515971*^9, 3.422262518108431*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSingleOccupiedStates", "=", "4"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"NumberOfTotalStates", " ", "=", " ", 
   SuperscriptBox["2", "NumberOfSingleOccupiedStates"]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42225970942581*^9, 3.422259753959048*^9}, {
  3.422260356857766*^9, 3.422260359126533*^9}}],

Cell["\<\
Next we construct effectively the NumberOfTotalStates * NumberOfTotalStates \
-dimensional creation operator matrices. This part you can skip. \
\>", "Text",
 CellChangeTimes->{{3.422260232752746*^9, 3.422260239500793*^9}, {
  3.422260278460007*^9, 3.422260287823447*^9}, {3.422260459641082*^9, 
  3.422260469284508*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"Occupation", " ", "=", " ", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"IntegerDigits", "[", 
      RowBox[{"i", ",", "2", ",", " ", "NumberOfSingleOccupiedStates"}], 
      "]"}], ",", " ", 
     RowBox[{"{", 
      RowBox[{"i", ",", "0", ",", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]}], "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Operator", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfSingleOccupiedStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], ";"}], "\n", 
 RowBox[{
  RowBox[{
   RowBox[{"OperatorDagger", " ", "=", "Operator"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{"For", "[", 
  RowBox[{
   RowBox[{"r", "=", "1"}], ",", " ", 
   RowBox[{"r", "\[LessEqual]", " ", "NumberOfSingleOccupiedStates"}], ",", 
   " ", 
   RowBox[{"r", "++"}], ",", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"For", "[", 
     RowBox[{
      RowBox[{"i", "=", "1"}], ",", " ", 
      RowBox[{"i", "\[LessEqual]", "NumberOfTotalStates"}], ",", " ", 
      RowBox[{"i", "++"}], ",", "\[IndentingNewLine]", 
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", " ", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfTotalStates"}], ",", " ", 
        RowBox[{"j", "++"}], ",", "\[IndentingNewLine]", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"For", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"state", "=", "1"}], ";", " ", 
            RowBox[{"StateSign", "=", "1"}]}], ",", " ", 
           RowBox[{"state", "<", "r"}], ",", " ", 
           RowBox[{"state", "++"}], ",", "\[IndentingNewLine]", 
           RowBox[{"StateSign", "*=", " ", 
            SuperscriptBox[
             RowBox[{"(", 
              RowBox[{"-", "1"}], ")"}], 
             RowBox[{"Occupation", "\[LeftDoubleBracket]", 
              RowBox[{"j", ",", " ", "state"}], 
              "\[RightDoubleBracket]"}]]}]}], "]"}], ";", 
         "\[IndentingNewLine]", "\[IndentingNewLine]", 
         RowBox[{"If", "[", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
                RowBox[{"BitXor", "[", 
                 RowBox[{
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "i", 
                   "\[RightDoubleBracket]"}], ",", " ", 
                  RowBox[{
                  "Occupation", "\[LeftDoubleBracket]", "j", 
                   "\[RightDoubleBracket]"}]}], "]"}], ",", "2"}], "]"}], " ",
               "\[Equal]", " ", 
              SuperscriptBox["2", 
               RowBox[{"NumberOfSingleOccupiedStates", "-", "r"}]]}], ")"}], 
            "\[IndentingNewLine]", "&&", " ", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "i", 
                "\[RightDoubleBracket]"}], "]"}], ">", 
              RowBox[{"FromDigits", "[", 
               RowBox[{
               "Occupation", "\[LeftDoubleBracket]", "j", 
                "\[RightDoubleBracket]"}], "]"}]}], ")"}]}], ",", 
           "\[IndentingNewLine]", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{"OperatorDagger", "\[LeftDoubleBracket]", 
             RowBox[{"r", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
            "=", "StateSign"}]}], "\[IndentingNewLine]", 
          "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", "]"}]}], 
     "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
     "Operator", "\[LeftDoubleBracket]", "r", "\[RightDoubleBracket]"}], " ", 
     "=", 
     RowBox[{"Transpose", "[", 
      RowBox[{
      "OperatorDagger", "\[LeftDoubleBracket]", "r", 
       "\[RightDoubleBracket]"}], "]"}]}]}]}], "\[IndentingNewLine]", 
  "]"}]}], "Input",
 CellChangeTimes->{{3.42226025055771*^9, 3.422260260564144*^9}}],

Cell["\<\
We have now     NumberOfSingleOccupiedStates = 4 creation operators:  \
OperatorDagger\[LeftDoubleBracket]1\[RightDoubleBracket],..., OperatorDagger\
\[LeftDoubleBracket]4\[RightDoubleBracket]. These shall be used to construct \
the necessary operators: \
\>", "Text",
 CellChangeTimes->{{3.422257807746289*^9, 3.422258008894191*^9}, {
   3.422258058754923*^9, 3.422258072175811*^9}, {3.422258129598818*^9, 
   3.422258130575411*^9}, {3.422259264756965*^9, 3.422259265264323*^9}, {
   3.422259320850274*^9, 3.422259651493224*^9}, {3.422259689880664*^9, 
   3.422259703369674*^9}, {3.422259772897277*^9, 3.422259784098445*^9}, {
   3.422260330982351*^9, 3.422260334508321*^9}, {3.422260366221652*^9, 
   3.422260446806384*^9}, 3.422260485771486*^9}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"fDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerUp", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDaggerDown", " ", "=", 
   RowBox[{
   "OperatorDagger", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}],
   ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"fDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dUp", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"dDown", " ", "=", 
   RowBox[{
   "Operator", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Szd", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"dDaggerUp", ".", "dUp"}], "-", 
      RowBox[{"dDaggerDown", ".", "dDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlusd", " ", "=", 
   RowBox[{"dDaggerUp", ".", "dDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinusd", " ", "=", 
   RowBox[{"dDaggerDown", ".", "dUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Szf", "=", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
      RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlusf", " ", "=", 
   RowBox[{"fDaggerUp", ".", "fDown"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinusf", " ", "=", 
   RowBox[{"fDaggerDown", ".", "fUp"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Sz", "=", 
   RowBox[{"Szd", "+", "Szf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SPlus", "=", 
   RowBox[{"SPlusd", "+", "SPlusf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SMinus", "=", 
   RowBox[{"SMinusd", "+", "SMinusf"}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"UnitOperator", " ", "=", " ", 
   RowBox[{
    RowBox[{"dDaggerUp", ".", "dUp"}], "+", 
    RowBox[{"dUp", ".", "dDaggerUp"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.422273235977622*^9, 3.422273314174119*^9}, {
  3.422513221796632*^9, 3.422513250741941*^9}}],

Cell["Now we can construct all the states. ", "Text",
 CellChangeTimes->{{3.422260563311906*^9, 3.422260571250809*^9}}],

Cell["States", "Subtitle"],

Cell["\<\
In this section we construct the corresponding states and order them as in \
the representation index table. First we construct the VacuumState and then \
based on it, we construct all the other states that are allowed. We can \
construct based on these states triplet as well as singlet states. These \
results can be extended to the case when SU(2) symmetries for charge or spin \
are considered. We could also use other basis states and mix State4 and \
State5 \
\>", "Text",
 CellChangeTimes->{{3.422260600765262*^9, 3.422260689774695*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"VacuumState", " ", "=", "  ", 
   RowBox[{"Table", "[", 
    RowBox[{"IntegerDigits", "[", 
     RowBox[{
      SuperscriptBox["2", 
       RowBox[{"NumberOfTotalStates", "-", "1"}]], ",", "2", ",", " ", 
      "NumberOfTotalStates"}], "]"}], "]"}]}], ";"}]], "Input"],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfBlockStates", " ", "=", " ", "8"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"BlockStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfBlockStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State1", "=", 
    RowBox[{"dDaggerDown", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State2", " ", "=", 
    RowBox[{"dDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State3", " ", "=", 
    RowBox[{"fDaggerDown", ".", "dDaggerDown", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State4", " ", "=", " ", 
    RowBox[{
     FractionBox["1", 
      SqrtBox["2"]], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"fDaggerDown", ".", "dDaggerUp"}], "-", 
        RowBox[{"fDaggerUp", ".", "dDaggerDown"}]}], ")"}], ".", 
      "VacuumState"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "5", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State5", " ", "=", " ", 
    RowBox[{
     FractionBox["1", 
      SqrtBox["2"]], " ", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"fDaggerDown", ".", "dDaggerUp"}], "+", 
        RowBox[{"fDaggerUp", ".", "dDaggerDown"}]}], ")"}], ".", 
      "VacuumState"}]}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "6", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State6", "=", " ", 
    RowBox[{"fDaggerUp", ".", "dDaggerUp", ".", "VacuumState"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "7", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State7", "=", " ", 
    RowBox[{
    "fDaggerUp", ".", "fDaggerDown", ".", "dDaggerDown", ".", 
     "VacuumState"}]}]}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "BlockStates", "\[LeftDoubleBracket]", "8", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"State8", " ", "=", " ", 
    RowBox[{
    "fDaggerUp", ".", "fDaggerDown", ".", "dDaggerUp", ".", 
     "VacuumState"}]}]}], ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
  3.422261115279298*^9, 3.422261172794231*^9}, {3.422639152725886*^9, 
  3.422639187965861*^9}, {3.422863134028229*^9, 3.422863179689223*^9}}],

Cell["\<\
Now we have to check that these states are indeed ordered.  If not, we have \
to reshuffle them.\
\>", "Text",
 CellChangeTimes->{{3.422260756543946*^9, 3.422260774656064*^9}, {
  3.422261998378347*^9, 3.42226200768368*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfSymmetries", "=", "2"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"QTotal", " ", "=", " ", 
    RowBox[{
     RowBox[{"fDaggerUp", ".", "fUp"}], "+", 
     RowBox[{"fDaggerDown", ".", "fDown"}], " ", "-", " ", "UnitOperator"}]}],
    ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "QTotal", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "Sz", ".", 
       RowBox[{
       "BlockStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}], "\[IndentingNewLine]"}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
  3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
  3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
  3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
  3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
  3.422273329630683*^9, 3.422273330834617*^9}, {3.422424163530083*^9, 
  3.422424182641464*^9}, {3.422424283395106*^9, 3.422424295014417*^9}, {
  3.422424504284488*^9, 3.422424529308189*^9}, {3.422425828934438*^9, 
  3.422425948432719*^9}, {3.422426022264904*^9, 3.422426031542518*^9}, {
  3.422426385786546*^9, 3.422426386364774*^9}, {3.42251326110266*^9, 
  3.422513266106889*^9}}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "1"}], 
      RowBox[{"-", 
       FractionBox["1", "2"]}]},
     {
      RowBox[{"-", "1"}], 
      FractionBox["1", "2"]},
     {"0", 
      RowBox[{"-", "1"}]},
     {"0", "0"},
     {"0", "0"},
     {"0", "1"},
     {"1", 
      RowBox[{"-", 
       FractionBox["1", "2"]}]},
     {"1", 
      FractionBox["1", "2"]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{{3.42226107342192*^9, 3.422261105951935*^9}, 
   3.422261189857225*^9, {3.422261361886205*^9, 3.422261382372428*^9}, {
   3.422261557654303*^9, 3.422261582070163*^9}, 3.422261670042143*^9, 
   3.4222618584621*^9, {3.422261936692001*^9, 3.422261947147733*^9}, 
   3.422261982631925*^9, 3.422273333398469*^9, 3.422422421961272*^9, 
   3.422424185206617*^9, {3.422424284343575*^9, 3.422424308315455*^9}, {
   3.422424505813002*^9, 3.42242452992562*^9}, 3.422425952720056*^9, {
   3.422426001119572*^9, 3.422426038231274*^9}, 3.422426387262979*^9, 
   3.422427033012301*^9, 3.422427279541733*^9, 3.42243318197125*^9, 
   3.42251304385896*^9, 3.422513278159466*^9, 3.422513834119571*^9, 
   3.4225932132783*^9, 3.422639200427157*^9, 3.42269236806703*^9, 
   3.422863182853609*^9}]
}, Open  ]],

Cell["\<\
Now we can create the matrices for any of the spectral operators as well as \
the hopping operators.\
\>", "Text"],

Cell["Spectral and hopping operators.", "Subtitle"],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedfDaggerUp", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "fDaggerUp", ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427108128412*^9, 3.422427111736916*^9}, {
   3.422427153114619*^9, 3.422427212919719*^9}, 3.422427535099241*^9, {
   3.422433150378739*^9, 3.422433152654368*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{-0.7071067811865476}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, \
{{0.7071067811865476}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {1.}, {0.}, \
{0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.7071067811865476}, {0.7071067811865476}, {0.}, {0.}, \
{0.}}}\
\>", "\<\
0.          0.          0.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.          0.          0.          0.    \
      0.

-0.707107   0.          0.          0.          0.          0.          0.    \
      0.

0.707107    0.          0.          0.          0.          0.          0.    \
      0.

0.          1.          0.          0.          0.          0.          0.    \
      0.

0.          0.          1.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.707107    0.707107    0.          0.    \
      0.\
\>"], "Output",
 CellChangeTimes->{
  3.422427009374603*^9, 3.422427041009018*^9, 3.422427112761862*^9, {
   3.422427155075844*^9, 3.422427213797051*^9}, 3.422427281483832*^9, 
   3.422433185296861*^9, 3.422513049729296*^9, 3.422513283827814*^9, 
   3.422513836491431*^9, 3.422593215638943*^9, 3.422639214317539*^9, 
   3.42269249421169*^9, 3.422863189516204*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedfDaggerDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "fDaggerDown", ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427306798278*^9, 3.42242733085097*^9}, 
   3.422427531987565*^9, {3.422433144007586*^9, 3.422433146373886*^9}}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}}, {{1.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.7071067811865476}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, \
{0.7071067811865476}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, \
{0.7071067811865476}, {-0.7071067811865476}, {0.}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {-1.}, {0.}, {0.}}}\
\>", "\<\
0.          0.          0.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.          0.          0.          0.    \
      0.

1.          0.          0.          0.          0.          0.          0.    \
      0.

0.          0.707107    0.          0.          0.          0.          0.    \
      0.

0.          0.707107    0.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.          0.          0.          0.    \
      0.

0.          0.          0.          0.707107    -0.707107   0.          0.    \
      0.

0.          0.          0.          0.          0.          -1.         0.    \
      0.\
\>"], "Output",
 CellChangeTimes->{3.422427332136108*^9, 3.422433187895875*^9, 
  3.422513053322246*^9, 3.422513286257581*^9, 3.42251383791203*^9, 
  3.422593217436718*^9, 3.422639219827187*^9, 3.422692496811594*^9, 
  3.422863200872016*^9}]
}, Open  ]],

Cell["Hamiltonian terms", "Subtitle"],

Cell[TextData[{
 "In this section we consider different terms appearing in the Hamiltonian \
for the Kondo Hamiltonian. The expression used for it : ",
 Cell[BoxData[
  FormBox[
   RowBox[{
    SubscriptBox["H", "Kondo"], "=", 
    RowBox[{
     FractionBox["1", "2"], "J", 
     RowBox[{"\[Sum]", " ", 
      RowBox[{
       OverscriptBox["S", 
        RowBox[{"\[Rule]", " "}]], " ", 
       SuperscriptBox["f", "+"], "\[Sigma]", " ", "f"}]}]}]}], 
   TraditionalForm]]],
 "   +   ",
 Cell[BoxData[
  FormBox[
   SuperscriptBox["S", "z"], TraditionalForm]]],
 ", there are two terms:"
}], "Text",
 CellChangeTimes->{{3.42251306960253*^9, 3.422513070535576*^9}, {
  3.42251312212709*^9, 3.422513128759267*^9}, {3.42251316560735*^9, 
  3.422513179294315*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfHamiltonianTerms", "=", "2"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HKondo", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"HKondo", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
   "=", 
   RowBox[{
    RowBox[{"Szd", ".", "Szf"}], " ", "+", " ", 
    RowBox[{
     FractionBox["1", "2"], 
     RowBox[{"(", 
      RowBox[{
       RowBox[{"SPlusd", ".", "SMinusf"}], " ", "+", " ", 
       RowBox[{"SMinusd", ".", "SPlusf"}]}], ")"}]}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"HKondo", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], 
    "=", "Szd"}], ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"OrderedHamiltonian", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfHamiltonianTerms", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
         RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], " ", 
        "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", 
         RowBox[{
         "HKondo", "\[LeftDoubleBracket]", "k", "\[RightDoubleBracket]"}], 
         ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"k", ",", "NumberOfHamiltonianTerms"}], "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422427388370775*^9, 3.422427437070901*^9}, {
  3.42242749172612*^9, 3.422427526825983*^9}, {3.422427584982783*^9, 
  3.422427590687387*^9}, {3.422433136386855*^9, 3.422433139040832*^9}, {
  3.422513104824159*^9, 3.422513107084334*^9}, {3.422513182350112*^9, 
  3.422513182441258*^9}, {3.422513298533192*^9, 3.422513434248784*^9}}],

Cell[OutputFormData["\<\
{{{{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.25}, {0.}, {0.}, {0.}, {0.}, {0.}}, 
  {{0.}, {0.}, {0.}, {-0.75}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, \
{0.}, {0.25}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {0.25}, {0.}, \
{0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}}}, 
 {{{-0.5}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.5}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {-0.5}, {0.}, {0.}, {0.}, {0.}, \
{0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.5}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, \
{0.5}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {0.5}, {0.}, \
{0.}}, 
  {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-0.5}, {0.}}, {{0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.5}}}}\
\>", "\<\
0.      0.      0.      0.      0.      0.      0.      0.
0.      0.      0.      0.      0.      0.      0.      0.
0.      0.      0.25    0.      0.      0.      0.      0.
0.      0.      0.      -0.75   0.      0.      0.      0.
0.      0.      0.      0.      0.25    0.      0.      0.
0.      0.      0.      0.      0.      0.25    0.      0.
0.      0.      0.      0.      0.      0.      0.      0.
0.      0.      0.      0.      0.      0.      0.      0.

-0.5    0.      0.      0.      0.      0.      0.      0.
0.      0.5     0.      0.      0.      0.      0.      0.
0.      0.      -0.5    0.      0.      0.      0.      0.
0.      0.      0.      0.      0.5     0.      0.      0.
0.      0.      0.      0.5     0.      0.      0.      0.
0.      0.      0.      0.      0.      0.5     0.      0.
0.      0.      0.      0.      0.      0.      -0.5    0.
0.      0.      0.      0.      0.      0.      0.      0.5\
\>"], "Output",
 CellChangeTimes->{
  3.422257525369031*^9, 3.422262250568518*^9, 3.42242243161654*^9, {
   3.422427510858111*^9, 3.422427528394162*^9}, 3.422427592710337*^9, 
   3.422433191045149*^9, 3.422513434909226*^9, 3.422513839720494*^9, 
   3.422593227115342*^9, 3.422639236780431*^9, 3.422692506123911*^9, 
   3.422863239385734*^9}]
}, Open  ]],

Cell["Spin operator", "Subtitle",
 CellChangeTimes->{{3.422257543929043*^9, 3.422257549617286*^9}}],

Cell["In this section we consider the spin operator", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422513148969978*^9, 3.422513151758013*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedSz", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfBlockStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfBlockStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedSz", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "Szd", ".", 
         RowBox[{
         "BlockStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfBlockStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfBlockStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422257650153471*^9, 3.422257658110106*^9}, {
   3.422262270013571*^9, 3.422262273937019*^9}, {3.422427554252507*^9, 
   3.422427571961936*^9}, {3.42242760434566*^9, 3.422427621897156*^9}, {
   3.422433129489568*^9, 3.422433132897229*^9}, 3.422513455801693*^9}],

Cell[OutputFormData["\<\
{{{-0.5}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.5}, {0.}, \
{0.}, {0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {-0.5}, {0.}, {0.}, {0.}, {0.}, \
{0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.5}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.5}, \
{0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}, {0.}, {0.5}, {0.}, {0.}}, 
 {{0.}, {0.}, {0.}, {0.}, {0.}, {0.}, {-0.5}, {0.}}, {{0.}, {0.}, {0.}, {0.}, \
{0.}, {0.}, {0.}, {0.5}}}\
\>", "\<\
-0.5   0.     0.     0.     0.     0.     0.     0.

0.     0.5    0.     0.     0.     0.     0.     0.

0.     0.     -0.5   0.     0.     0.     0.     0.

0.     0.     0.     0.     0.5    0.     0.     0.

0.     0.     0.     0.5    0.     0.     0.     0.

0.     0.     0.     0.     0.     0.5    0.     0.

0.     0.     0.     0.     0.     0.     -0.5   0.

0.     0.     0.     0.     0.     0.     0.     0.5\
\>"], "Output",
 CellChangeTimes->{3.422257525369031*^9, 3.422257659918318*^9, 
  3.422262275051522*^9, 3.422422433839974*^9, 3.422427623357605*^9, 
  3.422433194585268*^9, 3.422513456698519*^9, 3.422513842732745*^9, 
  3.422593230374349*^9, 3.422639263411425*^9, 3.422692644223613*^9, 
  3.422863256984717*^9, 3.422863300323248*^9}]
}, Open  ]],

Cell["", "Text",
 CellChangeTimes->{3.42227230000648*^9}],

Cell["Local operators and local basis", "Subtitle",
 CellChangeTimes->{{3.422272304428917*^9, 3.422272311966455*^9}}],

Cell[TextData[{
 "Now let us continue with the construction of the local states. We shall \
represent the operators ",
 Cell[BoxData[
  FormBox[
   SubscriptBox["f", "N"], TraditionalForm]]],
 "by the same operators as before. Then we have four local states: "
}], "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{"NumberOfLocalStates", " ", "=", " ", "4"}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"LocalStates", "=", 
   RowBox[{"Table", "[", 
    RowBox[{"0", ",", 
     RowBox[{"{", "NumberOfLocalStates", "}"}], ",", " ", 
     RowBox[{"{", "NumberOfTotalStates", "}"}]}], "]"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], "=",
    " ", "VacuumState"}], ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"fDaggerDown", ".", "VacuumState"}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{
    "LocalStates", "\[LeftDoubleBracket]", "3", "\[RightDoubleBracket]"}], 
    "=", 
    RowBox[{"fDaggerUp", ".", "VacuumState"}]}], ";"}], 
  " "}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
   "LocalStates", "\[LeftDoubleBracket]", "4", "\[RightDoubleBracket]"}], "=",
    
   RowBox[{"fDaggerUp", ".", "fDaggerDown", ".", "VacuumState"}]}], 
  ";"}]}], "Input",
 CellChangeTimes->{{3.42226084656336*^9, 3.422260942028706*^9}, {
  3.422261115279298*^9, 3.422261172794231*^9}, {3.422272469310303*^9, 
  3.422272491843473*^9}, {3.422272529795952*^9, 3.422272615717838*^9}, {
  3.422272682500722*^9, 3.422272698956333*^9}, {3.42227335639425*^9, 
  3.422273365754123*^9}, {3.422273443687094*^9, 3.42227344684052*^9}}],

Cell["We check if they are ordered properly:", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}}],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalQuantumNumbers", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfSymmetries", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"SzLocal", " ", "=", " ", 
   RowBox[{
    FractionBox["1", "2"], 
    RowBox[{"(", 
     RowBox[{
      RowBox[{"fDaggerUp", ".", "fUp"}], "-", 
      RowBox[{"fDaggerDown", ".", "fDown"}]}], ")"}]}]}], 
  ";"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"QLocal", " ", "=", " ", 
    RowBox[{
     RowBox[{"fDaggerUp", ".", "fUp"}], "+", 
     RowBox[{"fDaggerDown", ".", "fDown"}], " ", "-", " ", "UnitOperator"}]}],
    ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Table", "[", 
   RowBox[{
    RowBox[{
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "QLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
       RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}], "=", 
      RowBox[{
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}], ".", "SzLocal", ".", 
       RowBox[{
       "LocalStates", " ", "\[LeftDoubleBracket]", "i", 
        "\[RightDoubleBracket]"}]}]}], ";", "\[IndentingNewLine]", 
     "\[IndentingNewLine]", 
     RowBox[{"{", " ", 
      RowBox[{
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], ",", " ", 
       RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
        RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "}"}]}], ",", 
    RowBox[{"{", 
     RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], " ", "//",
   " ", "MatrixForm"}]}], "Input",
 CellChangeTimes->{{3.422261008583634*^9, 3.422261105170266*^9}, {
   3.422261185877381*^9, 3.422261257246138*^9}, {3.422261290925316*^9, 
   3.42226138106269*^9}, {3.422261556601073*^9, 3.422261621932931*^9}, {
   3.422261667725704*^9, 3.422261669214305*^9}, {3.422261754271065*^9, 
   3.422261761177845*^9}, {3.422261827807973*^9, 3.422261982154367*^9}, {
   3.422272998493306*^9, 3.422273023589198*^9}, {3.422273060650893*^9, 
   3.42227306270624*^9}, {3.422273102646266*^9, 3.422273104921433*^9}, {
   3.422273156211252*^9, 3.422273179056414*^9}, {3.42227334592718*^9, 
   3.422273346665373*^9}, 3.422427646446775*^9, {3.422427701122459*^9, 
   3.422427780430807*^9}, {3.42242781943902*^9, 3.422427835662201*^9}, 
   3.422593238762993*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"-", "1"}], "0"},
     {"0", 
      RowBox[{"-", 
       FractionBox["1", "2"]}]},
     {"0", 
      FractionBox["1", "2"]},
     {"1", "0"}
    },
    GridBoxAlignment->{
     "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{
  3.422273025881706*^9, {3.422273077127199*^9, 3.422273105554888*^9}, 
   3.422273179924059*^9, {3.422273347766973*^9, 3.422273369596677*^9}, 
   3.422273452148074*^9, 3.42242243891415*^9, {3.422427783820186*^9, 
   3.422427793470396*^9}, 3.422427836351067*^9, 3.422433199232154*^9, 
   3.422513473454277*^9, 3.422513846525845*^9, {3.422593237016459*^9, 
   3.422593241662812*^9}, 3.422639277320322*^9, 3.422692647412816*^9, 
   3.422863303468141*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalfDaggerUp", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "LocalStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "fDaggerUp", ".", 
         RowBox[{
         "LocalStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422262166469301*^9, 3.422262185711958*^9}, {
   3.422262232176988*^9, 3.422262232892694*^9}, {3.42227287163859*^9, 
   3.422272880981803*^9}, {3.4224229432499*^9, 3.422422953983789*^9}, {
   3.422427878701063*^9, 3.422427932721799*^9}, {3.422433118922209*^9, 
   3.42243312210934*^9}, 3.422593243403996*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, {0.}}, {{1.}, {0.}, {0.}, \
{0.}}, {{0.}, {1.}, {0.}, {0.}}}\
\>", "\<\
0.   0.   0.   0.

0.   0.   0.   0.

1.   0.   0.   0.

0.   1.   0.   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, 
   3.422272883289382*^9, 3.422273385171968*^9, 3.422273462344101*^9, 
   3.422422442190021*^9, {3.422422946203317*^9, 3.422422954776595*^9}, 
   3.422427934223118*^9, 3.422433201722852*^9, 3.422513499654047*^9, 
   3.422513847815914*^9, 3.422593244149783*^9, 3.422639312490154*^9, 
   3.422692649356723*^9, 3.422863304590594*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"OrderedLocalfDaggerDown", "=", 
    RowBox[{"Table", "[", 
     RowBox[{"0", ",", 
      RowBox[{"{", "NumberOfLocalStates", " ", "}"}], ",", " ", 
      RowBox[{"{", "NumberOfLocalStates", "}"}]}], "]"}]}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
         RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "=", 
        RowBox[{
         RowBox[{
         "LocalStates", "\[LeftDoubleBracket]", "i", 
          "\[RightDoubleBracket]"}], ".", "fDaggerDown", ".", 
         RowBox[{
         "LocalStates", "\[LeftDoubleBracket]", "j", 
          "\[RightDoubleBracket]"}]}]}], "}"}], ",", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"i", ",", "NumberOfLocalStates"}], " ", "}"}], ",", " ", 
      RowBox[{"{", 
       RowBox[{"j", ",", "NumberOfLocalStates"}], " ", "}"}]}], "]"}], "//", 
    "MatrixForm"}], "//", "OutputForm"}], "//", "N"}]}], "Input",
 CellChangeTimes->{{3.422262166469301*^9, 3.422262185711958*^9}, {
   3.422262232176988*^9, 3.422262232892694*^9}, {3.42227287163859*^9, 
   3.422272910227324*^9}, {3.422427943371719*^9, 3.42242795422947*^9}, {
   3.422433110442019*^9, 3.422433114362987*^9}, 3.422593247366937*^9}],

Cell[OutputFormData["\<\
{{{0.}, {0.}, {0.}, {0.}}, {{1.}, {0.}, {0.}, {0.}}, {{0.}, {0.}, {0.}, \
{0.}}, {{0.}, {0.}, {-1.}, {0.}}}\
\>", "\<\
0.    0.    0.    0.

1.    0.    0.    0.

0.    0.    0.    0.

0.    0.    -1.   0.\
\>"], "Output",
 CellChangeTimes->{
  3.422257511474477*^9, 3.422262186465329*^9, 3.422262234138565*^9, {
   3.422272883289382*^9, 3.422272910698491*^9}, 3.42227341602693*^9, 
   3.422273467127413*^9, 3.422422443315091*^9, 3.422427955645153*^9, 
   3.422433204022178*^9, 3.422513504970295*^9, 3.422513849216269*^9, 
   3.422593248351383*^9, 3.422639313647962*^9, 3.422692650593469*^9, 
   3.4228633056748*^9}]
}, Open  ]],

Cell["\<\
We write the quantum numbers to file \"QuantumNumbers\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./QuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfBlockStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422428031091177*^9, 3.422428228324012*^9}, 
   3.422428274771525*^9, {3.422428334035563*^9, 3.422428414302918*^9}, {
   3.422433090180519*^9, 3.422433103257273*^9}}],

Cell["\<\
Next, the hopping, spectral and Hamiltonian operators are saved in \
\"HoppingOperators\", \"SpectralOperators\", and \"HamiltonianBlocks\", \
respectively:\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428580988006*^9}, {3.422428618117171*^9, 3.422428703986899*^9}, {
  3.422513558824415*^9, 3.422513564019709*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<fDaggerUp\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderedfDaggerUp", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<\\nfDaggerDown\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderedfDaggerDown", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{
  3.42251352727455*^9, {3.422513570627052*^9, 3.422513657546017*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./SpectralOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<Sz\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderedSz", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422513664909368*^9, 3.422513717313067*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./HamiltonianBlocks\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"For", "[", 
   RowBox[{
    RowBox[{"k", "=", "1"}], ",", 
    RowBox[{"k", "\[LessEqual]", " ", "NumberOfHamiltonianTerms"}], ",", " ", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"Write", "[", 
      RowBox[{
      "strm", ",", "\"\<HamiltonianBlock \>\"", ",", "k", ",", 
       "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"For", "[", 
      RowBox[{
       RowBox[{"i", "=", "1"}], ",", 
       RowBox[{"i", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", " ", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"For", "[", 
         RowBox[{
          RowBox[{"j", "=", "1"}], ",", 
          RowBox[{"j", "\[LessEqual]", " ", "NumberOfBlockStates"}], ",", 
          "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", 
             RowBox[{"N", "[", 
              RowBox[{"CForm", "[", 
               RowBox[{"OrderedHamiltonian", "\[LeftDoubleBracket]", 
                RowBox[{"k", ",", "i", ",", "j"}], "\[RightDoubleBracket]"}], 
               " ", "]"}], "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{"j", "<", "NumberOfBlockStates"}], ",", 
             RowBox[{"WriteString", "[", 
              RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
           "\[IndentingNewLine]", 
           RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
        RowBox[{"WriteString", "[", 
         RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"i", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Write", "[", 
      RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", "\[IndentingNewLine]",
      
     RowBox[{"k", "++"}]}]}], "]"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422422461984407*^9, 3.422422590368324*^9}, {
  3.422422657679912*^9, 3.422422675770448*^9}, {3.422422769910713*^9, 
  3.422422774310689*^9}, {3.422422907355884*^9, 3.422422907479271*^9}, {
  3.422423014025691*^9, 3.422423024399214*^9}, {3.422423065573881*^9, 
  3.422423075136177*^9}, {3.422428735589752*^9, 3.422428797092516*^9}, {
  3.42242951569936*^9, 3.422429705585544*^9}, {3.422429793487009*^9, 
  3.4224298166685*^9}, {3.422429927802331*^9, 3.422429930748695*^9}, {
  3.422432038815338*^9, 3.422432247503146*^9}, {3.422432279060799*^9, 
  3.422432423000301*^9}, {3.422432470625659*^9, 3.422432486448996*^9}, {
  3.422513762990068*^9, 3.422513804257083*^9}}],

Cell["\<\
Finally, we write the local quantum numbers to and local operators to file \
\"LocalQuantumNumbers\" and \"LocalHoppingOperators\" (spin quantum numbers \
are multiplied by 2):\
\>", "Text",
 CellChangeTimes->{{3.422257565286391*^9, 3.422257625769576*^9}, {
  3.422272336181767*^9, 3.422272442871899*^9}, {3.422272715828128*^9, 
  3.422272724895062*^9}, {3.42227298331989*^9, 3.422272993852855*^9}, {
  3.422427977332516*^9, 3.422427994966664*^9}, {3.422428458510858*^9, 
  3.422428522554788*^9}, {3.422432507395505*^9, 3.422432527004284*^9}, {
  3.422432571670566*^9, 3.422432576538502*^9}}],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalQuantumNumbers\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Table", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"WriteString", "[", " ", 
      RowBox[{"strm", ",", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
          RowBox[{"1", ",", "i"}], "\[RightDoubleBracket]"}], " ", "]"}], 
        "]"}], ",", " ", "\"\<\\t\>\"", ",", " ", 
       RowBox[{"N", "[", 
        RowBox[{"CForm", "[", " ", 
         RowBox[{"2", " ", 
          RowBox[{"OrderedLocalQuantumNumbers", "\[LeftDoubleBracket]", 
           RowBox[{"2", ",", "i"}], "\[RightDoubleBracket]"}]}], "]"}], "]"}],
        ",", " ", "\"\<\\n\>\""}], "]"}], ",", "\[IndentingNewLine]", 
     RowBox[{"{", 
      RowBox[{"i", ",", "NumberOfLocalStates"}], "}"}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Close", "[", "strm", "]"}], ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"strm", "=", 
    RowBox[{"OpenWrite", "[", 
     RowBox[{"\"\<./LocalHoppingOperators\>\"", ",", 
      RowBox[{"FormatType", "\[Rule]", "OutputForm"}], ",", 
      RowBox[{"DOSTextFormat", "\[Rule]", "False"}], ",", 
      RowBox[{"CharacterEncoding", "\[Rule]", "\"\<Unicode\>\""}]}], "]"}]}], 
   ";"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<LocalfDaggerUp\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderedLocalfDaggerUp", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfLocalStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"Write", "[", 
    RowBox[{"strm", ",", "\"\<\\nLocalfDaggerDown\\n\>\""}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{
   RowBox[{"For", "[", 
    RowBox[{
     RowBox[{"i", "=", "1"}], ",", 
     RowBox[{"i", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", " ", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"For", "[", 
       RowBox[{
        RowBox[{"j", "=", "1"}], ",", 
        RowBox[{"j", "\[LessEqual]", " ", "NumberOfLocalStates"}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"WriteString", "[", 
          RowBox[{"strm", ",", 
           RowBox[{"N", "[", 
            RowBox[{"CForm", "[", 
             RowBox[{"OrderedLocalfDaggerDown", "\[LeftDoubleBracket]", 
              RowBox[{"i", ",", "j"}], "\[RightDoubleBracket]"}], " ", "]"}], 
            "]"}]}], " ", "]"}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"j", "<", "NumberOfLocalStates"}], ",", 
           RowBox[{"WriteString", "[", 
            RowBox[{"strm", ",", "\"\<\\t\>\""}], "]"}]}], "]"}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"j", "++"}]}]}], "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"WriteString", "[", 
       RowBox[{"strm", ",", "\"\<\\n\>\""}], "]"}], ";", 
      "\[IndentingNewLine]", 
      RowBox[{"i", "++"}]}]}], "]"}], ";"}], 
  "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
 RowBox[{
  RowBox[{"Close", "[", "strm", "]"}], ";"}]}], "Input",
 CellChangeTimes->{{3.422432891090526*^9, 3.422433072089265*^9}, 
   3.42243322189554*^9}]
},
WindowSize->{1280, 738},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Magnification->1,
FrontEndVersion->"6.0 for Linux x86 (32-bit) (April 20, 2007)",
StyleDefinitions->"Classroom.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[568, 21, 241, 10, 56, "Title"],
Cell[812, 33, 35, 0, 40, "Subtitle"],
Cell[850, 35, 234, 5, 63, "Input"],
Cell[1087, 42, 3955, 119, 273, "Text"],
Cell[5045, 163, 361, 8, 64, "Input"],
Cell[5409, 173, 331, 6, 24, "Text"],
Cell[5743, 181, 4167, 102, 429, "Input"],
Cell[9913, 285, 758, 12, 24, "Text"],
Cell[10674, 299, 2869, 86, 373, "Input"],
Cell[13546, 387, 119, 1, 24, "Text"],
Cell[13668, 390, 26, 0, 40, "Subtitle"],
Cell[13697, 392, 554, 9, 58, "Text"],
Cell[14254, 403, 303, 8, 69, "Input"],
Cell[14560, 413, 2991, 93, 270, "Input"],
Cell[17554, 508, 234, 5, 24, "Text"],
Cell[CellGroupData[{
Cell[17813, 517, 2799, 62, 217, "Input"],
Cell[20615, 581, 1692, 44, 215, "Output"]
}, Open  ]],
Cell[22322, 628, 124, 3, 24, "Text"],
Cell[22449, 633, 51, 0, 40, "Subtitle"],
Cell[CellGroupData[{
Cell[22525, 637, 1294, 32, 115, "Input"],
Cell[23822, 671, 1599, 39, 191, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25458, 715, 1249, 31, 115, "Input"],
Cell[26710, 748, 1476, 36, 191, "Output"]
}, Open  ]],
Cell[28201, 787, 37, 0, 40, "Subtitle"],
Cell[28241, 789, 759, 23, 42, "Text"],
Cell[CellGroupData[{
Cell[29025, 816, 2737, 71, 236, "Input"],
Cell[31765, 889, 2181, 40, 209, "Output"]
}, Open  ]],
Cell[33961, 932, 99, 1, 40, "Subtitle"],
Cell[34063, 935, 176, 2, 24, "Text"],
Cell[CellGroupData[{
Cell[34264, 941, 1346, 33, 115, "Input"],
Cell[35613, 976, 1218, 29, 191, "Output"]
}, Open  ]],
Cell[36846, 1008, 57, 1, 24, "Text"],
Cell[36906, 1011, 117, 1, 40, "Subtitle"],
Cell[37026, 1014, 387, 9, 24, "Text"],
Cell[37416, 1025, 1504, 42, 149, "Input"],
Cell[38923, 1069, 263, 3, 24, "Text"],
Cell[CellGroupData[{
Cell[39211, 1076, 2985, 69, 219, "Input"],
Cell[42199, 1147, 1184, 31, 131, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[43420, 1183, 1446, 34, 115, "Input"],
Cell[44869, 1219, 653, 18, 121, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[45559, 1242, 1404, 33, 115, "Input"],
Cell[46966, 1277, 641, 18, 121, "Output"]
}, Open  ]],
Cell[47622, 1298, 432, 8, 24, "Text"],
Cell[48057, 1308, 1484, 34, 149, "Input"],
Cell[49544, 1344, 582, 10, 24, "Text"],
Cell[50129, 1356, 3515, 89, 455, "Input"],
Cell[53647, 1447, 1983, 50, 268, "Input"],
Cell[55633, 1499, 3047, 67, 302, "Input"],
Cell[58683, 1568, 602, 10, 24, "Text"],
Cell[59288, 1580, 4940, 123, 608, "Input"]
}
]
*)

(* End of internal cache information *)
